/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
 The MIT License

 Copyright (c) 2013 - 2014
 1. SEECS, National University of Sciences and Technology, Pakistan (2013 - 2014)
 2. Bibrak Qamar  (2013 - 2014)

 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the
 "Software"), to deal in the Software without restriction, including
 without limitation the rights to use, copy, modify, merge, publish,
 distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to
 the following conditions:

 The above copyright notice and this permission notice shall be included
 in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


 /*
 * File         : mpi_Group.c
 * Author       : Bibrak Qamar
 * Created      : Tue Sept  17 11:15:11 PKT 2013
 * Revision     : $Revision: 1.1 $
 * Updated      : $Date: 2014/03/11 11:30:00 $
 *
 */

#include "mpi_Group.h"
#include "mpi.h"
#include "mpi_Commons.h"
#include <inttypes.h>
#include <stdlib.h> // for malloc

jfieldID GrouphandleID;

/* class mpi_Group */

/*
 * Class:     mpi_Group
 * Method:    init
 * Signature: ()V
 *
 * This function is called once in the static region of mpjdev.Group class
 * it only gets the field ID for handle in the global jfieldID GrouphandleID
 * so that we don't have to
 * do a lookup each time its used.
 */
JNIEXPORT void JNICALL Java_mpi_Group_init
    (JNIEnv *env, jclass thisClass)
{
    GrouphandleID = (*env)->GetFieldID(env, thisClass, "handleOfGroup", "J");
}

/*
 * Class:     mpi_Group
 * Method:    GetGroup
 * Signature: (I)V
 *
 * This function set the group to either the default group i.e. the group of
 * universal COMM_WORLD (typeOfWorld == 2) or it sets to MPI_GROUP_EMPTY
 * (typeOfWorld == 3)
 */
JNIEXPORT void JNICALL Java_mpi_Group_GetGroup(JNIEnv  *env,
                                               jobject thisObject,
                                               jint typeOfWorld)
{
    MPI_Group group;

    switch (typeOfWorld) {
        case 0: //Group for MPI_COMM_WORLD
            MPI_Comm_group(MPI_COMM_WORLD, &group);
            (*env)->SetLongField(env,
                                 thisObject,
                                 GrouphandleID,
                                 (jlong)group);
            break;
        case 1: //Group for MPI_GROUP_EMPTY
            (*env)->SetLongField(env,
                                 thisObject,
                                 GrouphandleID,
                                 (jlong)MPI_GROUP_EMPTY);
            break;
        case 2: //Group for MPI_GROUP_NULL
            (*env)->SetLongField(env,
                                 thisObject,
                                 GrouphandleID,
                                 (jlong)MPI_GROUP_NULL);
            break;
        default:
            break;
    }
}

/*
 * Class:     mpi_Group
 * Method:    nativeFree
 * Signature: ()V
 *
 * Free the group and set the handle to MPI_GROUP_NULL
 */
JNIEXPORT void JNICALL Java_mpi_Group_nativeFree(JNIEnv *env,
                                                 jobject thisObject)
{
    MPI_Group group = (MPI_Group)((*env)->GetLongField(env, thisObject,
            GrouphandleID));
    MPI_Group_free(&group);
    (*env)->SetLongField(env, thisObject, GrouphandleID,
            (jlong)MPI_GROUP_NULL);
}

/*
 * Class:     mpi_Group
 * Method:    nativeSize
 * Signature: ()I
 *
 * Size of the group
 */
JNIEXPORT jint JNICALL Java_mpi_Group_nativeSize(JNIEnv *env,
                                                 jobject thisObject)
{
    int size;

    MPI_Group group = (MPI_Group)(
        (*env)->GetLongField(env, thisObject, GrouphandleID));

    int errCode = MPI_Group_size(group, &size);

    if (MPI_SUCCESS != errCode) {
        println("ERROR: nativeSize for Group failed.");
        return -1;
    } else {
        return size;
    }
}

/*
 * Class:     mpi_Group
 * Method:    nativeRank
 * Signature: ()I
 *
 * Rank of this process with in the group
 */
JNIEXPORT jint JNICALL Java_mpi_Group_nativeRank(JNIEnv *env,
                                                 jobject thisObject)
{
    int rank;

    MPI_Group group = (MPI_Group)(
        (*env)->GetLongField(env, thisObject, GrouphandleID));

    int errCode = MPI_Group_rank(group, &rank);

    if (MPI_ERR_GROUP == errCode || MPI_ERR_ARG == errCode) {
        println("ERROR: nativeRank for Group failed.");
        fflush (stdout);
        rank = -1;
    }

    if (0 > rank) {
        return -1;
    }

    return rank;
}

/*
 * Class:     mpi_Group
 * Method:    nativeIncl
 * Signature: ([I)J
 *
 * Incl for group
 */
JNIEXPORT jlong JNICALL Java_mpi_Group_nativeIncl(JNIEnv *env,
						  jobject thisObject,
						  jintArray ranks)
{
    int n;
    jint *rks;
    jboolean isCopy = JNI_TRUE;
    MPI_Group newGroup;

    n = (*env)->GetArrayLength(env, ranks);
    rks = (*env)->GetIntArrayElements(env, ranks, &isCopy);

    MPI_Group existingGroup = (MPI_Group)(
        (*env)->GetLongField(env, thisObject, GrouphandleID));
    
    MPI_Group_incl(existingGroup, n, (int*) rks, &newGroup);
   
    (*env)->ReleaseIntArrayElements(env, ranks, rks, 0);
    
    return (jlong) newGroup;
}

/*
 * Class:     mpi_Group
 * Method:    nativeExcl
 * Signature: ([I)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Group_nativeExcl(JNIEnv *env,
    						  jobject thisObject,
						  jintArray ranks)
{
    int n;
    jint *rks;
    jboolean isCopy = JNI_TRUE;
    MPI_Group newGroup;

    n = (*env)->GetArrayLength(env, ranks);
    rks = (*env)->GetIntArrayElements(env, ranks, &isCopy);
    
    MPI_Group existingGroup = (MPI_Group)(
        (*env)->GetLongField(env, thisObject, GrouphandleID));

    int errCode = MPI_Group_excl(existingGroup, n, (int*) rks, &newGroup);
  
    (*env)->ReleaseIntArrayElements(env, ranks, rks, 0);

    if (MPI_SUCCESS != errCode) {
        println("ERROR: nativeExcl Failed.");
        return -1;
    } else {
        return (jlong) newGroup;
    }
}

/*
 * Class:     mpi_Group
 * Method:    nativeRangeIncl
 * Signature: (I[I)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Group_nativeRangeIncl(JNIEnv *env,
    						       jobject thisObject,
					               jint n,
						       jintArray oneDimRanges)
{
    int oneDimRangesLen;
    jint *nativeOneDimRanges;
    jboolean isCopy = JNI_TRUE;
    MPI_Group newGroup;

    oneDimRangesLen = (*env)->GetArrayLength(env, oneDimRanges);
    nativeOneDimRanges = (*env)->GetIntArrayElements(env, oneDimRanges, &isCopy);
   
    MPI_Group existingGroup = (MPI_Group)(
        (*env)->GetLongField(env, thisObject, GrouphandleID));

    int errCode = MPI_Group_range_incl(existingGroup, n,
        (int (*)[3]) nativeOneDimRanges, &newGroup);

    (*env)->ReleaseIntArrayElements(env, oneDimRanges, nativeOneDimRanges, 0);

    if (MPI_SUCCESS != errCode) {
        println("ERROR: nativeRangeIncl Failed.");
        return -1;
    } else {
        return (jlong) newGroup;
    }
}

/*
 * Class:     mpi_Group
 * Method:    nativeRangeExcl
 * Signature: (I[I)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Group_nativeRangeExcl(JNIEnv    *env,
						       jobject   thisObject,
						       jint      n,
						       jintArray oneDimRanges)
{
    int oneDimRangesLen;
    jint *nativeOneDimRanges;
    jboolean isCopy = JNI_TRUE;
    MPI_Group newGroup;

    oneDimRangesLen = (*env)->GetArrayLength(env, oneDimRanges);
    nativeOneDimRanges =
        (*env)->GetIntArrayElements(env, oneDimRanges, &isCopy);
 
    MPI_Group existingGroup = (MPI_Group)(
        (*env)->GetLongField(env, thisObject, GrouphandleID));

    int errCode = MPI_Group_range_excl(existingGroup, n,
        (int (*)[3]) nativeOneDimRanges, &newGroup);

    (*env)->ReleaseIntArrayElements(env, oneDimRanges, nativeOneDimRanges, 0);

    if (MPI_SUCCESS != errCode) {
        println("ERROR: nativeRangeExcl Failed.");
        return -1;
    } else {
        return (jlong) newGroup;
    }
}
/*
 * Class:     mpi_Group
 * Method:    nativetransRanks
 * Signature: (JJ[I)[I
 */
JNIEXPORT jintArray JNICALL Java_mpi_Group_nativetransRanks(JNIEnv    *env,
							    jclass    thisClass,
							    jlong     group1_,
							    jlong     group2_,
    							    jintArray ranks)
{
    int errCode;
    int n;
    jintArray result;
    jint *rks, *rks2;
    jboolean isCopy = JNI_TRUE;
  
    MPI_Group group1 = (MPI_Group) (intptr_t) group1_;
    MPI_Group group2 = (MPI_Group) (intptr_t) group2_;

    n = (*env)->GetArrayLength(env, ranks);
    rks = (*env)->GetIntArrayElements(env, ranks, &isCopy);
    rks2 = malloc(sizeof(int) * n);
    result = (*env)->NewIntArray(env, n);

    if (result == NULL) {
      return NULL; /* out of memory error thrown */
    }

    errCode = MPI_Group_translate_ranks(group1, n, (int*) rks, group2,
        (int*) rks2);

    if (MPI_SUCCESS != errCode) {
        println("ERROR: nativetransRanks Failed.");
        free(rks2);
        (*env)->ReleaseIntArrayElements(env, ranks, rks, 0);
        return NULL;
    }

    (*env)->ReleaseIntArrayElements(env, ranks, rks, 0);
    // move from rks2 to the java jintArray result
    (*env)->SetIntArrayRegion(env, result, 0, n, rks2);
    free(rks2);

    return (jintArray) result;
}

/*
 * Class:     mpi_Group
 * Method:    nativeCompare
 * Signature: (JJ)I
 */
JNIEXPORT jint JNICALL Java_mpi_Group_nativeCompare(JNIEnv *env,
    						    jobject thisObject,
						    jlong _group1,
						    jlong _group2)
{
    int result;
    int errCode;

    MPI_Group group1 = (MPI_Group) (intptr_t) _group1;
    MPI_Group group2 = (MPI_Group) (intptr_t) _group2;

    errCode = MPI_Group_compare(group1, group2, &result);
    
    if (MPI_SUCCESS != errCode) {
        println("ERROR: nativeCompare failed.");
        return -1;
    } else {
        return result;
    }
}

/*
 * Class:     mpi_Group
 * Method:    nativeUnion
 * Signature: (JJ)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Group_nativeUnion(JNIEnv  *env,
						   jobject thisObject,
						   jlong   group1_,
						   jlong   group2_)
{
    int result;
    int errCode;

    MPI_Group group1 = (MPI_Group) (intptr_t) group1_;
    MPI_Group group2 = (MPI_Group) (intptr_t) group2_;
    MPI_Group newGroup;
    errCode = MPI_Group_union(group1, group2, &newGroup);

    if (MPI_SUCCESS != errCode) {
        printf("ERROR: nativeUnion failed.");
        return -1;
    }

    return (jlong) newGroup;
}

/*
 * Class:     mpi_Group
 * Method:    nativeIntersection
 * Signature: (JJ)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Group_nativeIntersection(JNIEnv  *env,
							  jobject thisObject,
							  jlong   group1_,
							  jlong   group2_)
{
    int result;
    int errCode;

    MPI_Group group1 = (MPI_Group) (intptr_t) group1_;
    MPI_Group group2 = (MPI_Group) (intptr_t) group2_;
    MPI_Group newGroup;
    errCode = MPI_Group_intersection(group1, group2, &newGroup);

    if (MPI_SUCCESS != errCode) {
        println("ERROR: nativeIntersection failed.");
        return -1;
    }

    return (jlong) newGroup;
}

/*
 * Class:     mpi_Group
 * Method:    nativeDifference
 * Signature: (JJ)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Group_nativeDifference(JNIEnv *env,
							jobject thisObject,
							jlong group1_,
							jlong group2_)
{
    int result;
    int errCode;

    MPI_Group group1 = (MPI_Group) (intptr_t) group1_;
    MPI_Group group2 = (MPI_Group) (intptr_t) group2_;
    MPI_Group newGroup;
    errCode = MPI_Group_difference(group1, group2, &newGroup);

    if (MPI_SUCCESS != errCode) {
        println("ERROR: nativeDifference failed.");
        return -1;
    }

    return (jlong) newGroup;
}
