/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

#include "mpi.h"
#include "mpi_Intracomm.h"
#include <inttypes.h>
#include "mpi_Commons.h"
#include <stdlib.h>
#include <string.h>

/*
 * Class:     mpi_Intracomm
 * Method:    nativeBarrier
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeBarrier(JNIEnv *env,
                                                        jobject thisClass,
                                                        jlong comm)
{
    MPI_Barrier((MPI_Comm) (intptr_t) comm);
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeBcast
 * Signature: (JLjava/nio/ByteBuffer;II)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeBcast(JNIEnv *env,
                                                      jobject thisClass,
                                                      jlong comm,
                                                      jobject buf,
                                                      jint count,
                                                      jint root)
{
    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;

    char *buffer_address = NULL;

    buffer_address = (char *)(*env)->GetDirectBufferAddress(env,(jobject)buf);

    MPI_Bcast(buffer_address, count, MPI_BYTE, root, mpi_comm);
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeReduce
 * Signature: (JLjava/nio/Buffer;Ljava/nio/Buffer;IIII)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeReduce(JNIEnv *env,
                                                       jobject thisClass,
                                                       jlong comm,
                                                       jobject sendbuf,
                                                       jobject recvbuf,
                                                       jint op,
                                                       jint count,
                                                       jint datatype,
                                                       jint root)
{
    int id;
    void *sbuf, *rbuf;
    
    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;
    MPI_Comm_rank(mpi_comm, &id);

    sbuf = (*env)->GetDirectBufferAddress(env,(jobject)sendbuf);

    if (id == root) {
        rbuf = (*env)->GetDirectBufferAddress(env,(jobject)recvbuf);
    }

    MPI_Reduce(sbuf, rbuf, count, datatype, op, root, mpi_comm);
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeAllReduce
 * Signature: (JLjava/nio/Buffer;Ljava/nio/Buffer;III)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeAllReduce(JNIEnv *env,
                                                         jobject thisClass,
                                                         jlong comm,
                                                         jobject sendbuf,
                                                         jobject recvbuf,
                                                         jint op,
                                                         jint datatype,
                                                         jint count)
{
    void *sbuf = (*env)->GetDirectBufferAddress(env,(jobject)sendbuf);
    void *rbuf = (*env)->GetDirectBufferAddress(env,(jobject)recvbuf);

    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;

    MPI_Allreduce(sbuf, rbuf, count, datatype, op, mpi_comm);
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeGather
 * Signature: (JLjava/nio/Buffer;IIILjava/nio/Buffer;IIII)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeGather(JNIEnv *env,
                                                       jobject thisClass,
                                                       jlong comm,
                                                       jobject sendbuf,
                                                       jint sendcount,
                                                       jobject recvbuf,
                                                       jint recvcount,
                                                       jint root)
{
    int rank = -1;
    void *sbuf = NULL;
    void *rbuf = NULL;

    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;
    MPI_Comm_rank(mpi_comm, &rank);

    sbuf = (*env)->GetDirectBufferAddress(env, (jobject)sendbuf);

    if (rank == root) {
      rbuf = (*env)->GetDirectBufferAddress(env, (jobject)recvbuf);
    }

    MPI_Gather(sbuf, sendcount, MPI_BYTE, rbuf, recvcount,
            MPI_BYTE, root, mpi_comm);
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeGatherv
 * Signature: (JLjava/nio/Buffer;IIILjava/nio/Buffer;I[I[III)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeGatherv(JNIEnv *env,
                                                        jobject thisClass,
                                                        jlong comm,
                                                        jobject sendbuf,
                                                        jint sendcount,
                                                        jint sendtype,
                                                        jobject recvbuf,
                                                        jintArray recvcount,
                                                        jint recvtype,
                                                        jintArray displs,
                                                        jint root)
{
    int rank = -1;
    int *rcount;
    int *disp;

    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;
    MPI_Comm_rank(mpi_comm, &rank);

    jboolean isCopy;

    if (rank == root) {
      rcount = (*env)->GetIntArrayElements(env, recvcount, &isCopy);
      disp = (*env)->GetIntArrayElements(env, displs, &isCopy);
    }

    void *sbuf = (*env)->GetDirectBufferAddress(env,(jobject)sendbuf);
    void *rbuf;

    if (rank == root) {
        rbuf = (*env)->GetDirectBufferAddress(env,(jobject)recvbuf);
    }

    MPI_Gatherv(sbuf, sendcount, MPI_BYTE, rbuf, rcount, disp, MPI_BYTE,
          root, mpi_comm);

    if (rank == root) {
      (*env)->ReleaseIntArrayElements(env, recvcount, rcount, JNI_ABORT);
      (*env)->ReleaseIntArrayElements(env, displs, disp, JNI_ABORT);
    }
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeAllgatherv
 * Signature: (JLjava/nio/Buffer;IIILjava/nio/Buffer;I[I[II)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeAllgatherv(JNIEnv *env,
                                                           jobject thisClass,
                                                           jlong comm,
                                                           jobject sendbuf,
                                                           jint sendcount,
                                                           jint sendtype,
                                                           jobject recvbuf,
                                                           jintArray recvcount,
                                                           jint recvtype,
                                                           jintArray displs)
{
    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;

    int *rcount;
    int *disp;

    jboolean isCopy;
    rcount = (*env)->GetIntArrayElements(env, recvcount, &isCopy);
    disp = (*env)->GetIntArrayElements(env, displs, &isCopy);

    void *sbuf = (*env)->GetDirectBufferAddress(env,(jobject)sendbuf);
    void *rbuf = (*env)->GetDirectBufferAddress(env,(jobject)recvbuf);

    MPI_Allgatherv(sbuf, sendcount, MPI_BYTE, rbuf, rcount, disp, MPI_BYTE, mpi_comm);

    (*env)->ReleaseIntArrayElements(env, recvcount, rcount, JNI_ABORT);
    (*env)->ReleaseIntArrayElements(env, displs, disp, JNI_ABORT);
}


/*
 * Class:     mpi_Intracomm
 * Method:    nativeAllgather
 * Signature: (JLjava/nio/Buffer;IIILjava/nio/Buffer;III)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeAllgather(JNIEnv *env,
                                                          jobject thisClass,
                                                          jlong comm,
                                                          jobject sendbuf,
                                                          jint sendcount,
                                                          jobject recvbuf,
                                                          jint recvcount)
{
    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;

    void *sbuf = (*env)->GetDirectBufferAddress(env,(jobject)sendbuf);
    void *rbuf = (*env)->GetDirectBufferAddress(env,(jobject)recvbuf);

    MPI_Allgather(sbuf, sendcount, MPI_BYTE, rbuf, recvcount,
            MPI_BYTE, mpi_comm);
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeScatter
 * Signature: (JLjava/nio/Buffer;IIILjava/nio/Buffer;IIII)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeScatter(JNIEnv *env,
                                                        jobject thisClass,
                                                        jlong comm,
                                                        jobject sendbuf,
                                                        jint sendcount,
                                                        jobject recvbuf,
                                                        jint recvcount,
                                                        jint root)
{
    int rank = -1;
    void *sbuf;

    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;
    MPI_Comm_rank(mpi_comm, &rank);

    if (rank == root) {
        sbuf = (*env)->GetDirectBufferAddress(env,(jobject)sendbuf);
    }

    void *rbuf = (*env)->GetDirectBufferAddress(env,(jobject)recvbuf);

    MPI_Scatter(sbuf, sendcount, MPI_BYTE, rbuf, recvcount,
            MPI_BYTE, root, mpi_comm);
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeScatterv
 * Signature: (JLjava/nio/Buffer;I[I[IILjava/nio/Buffer;IIII)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeScatterv(JNIEnv *env,
                                                         jobject thisClass,
                                                         jlong comm,
                                                         jobject sendbuf,
                                                         jintArray sendcount,
                                                         jint sendtype,
                                                         jintArray displs,
                                                         jobject recvbuf,
                                                         jint recvcount,
                                                         jint recvtype,
                                                         jint root)
{
    int rank = -1;
    int *scount;
    int *disp;

    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;
    MPI_Comm_rank(mpi_comm, &rank);

    jboolean isCopy;

    if (rank == root) {
      scount = (*env)->GetIntArrayElements(env, sendcount, &isCopy);
      disp = (*env)->GetIntArrayElements(env, displs, &isCopy);
    }

    void *sbuf;
    void *rbuf = (*env)->GetDirectBufferAddress(env,(jobject)recvbuf);

    if (rank == root) {
        sbuf = (*env)->GetDirectBufferAddress(env,(jobject)sendbuf);
    }

    MPI_Scatterv(sbuf, scount, disp, MPI_BYTE, rbuf, recvcount, MPI_BYTE,
          root, mpi_comm);

    if (rank == root) {
      (*env)->ReleaseIntArrayElements(env, sendcount, scount, JNI_ABORT);
      (*env)->ReleaseIntArrayElements(env, displs, disp, JNI_ABORT);
    }
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeAlltoall
 * Signature: (JLjava/nio/Buffer;IIILjava/nio/Buffer;III)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeAlltoall(JNIEnv *env,
                                                         jobject thisClass,
                                                         jlong comm,
                                                         jobject sendbuf,
                                                         jint sendcount,
                                                         jobject recvbuf,
                                                         jint recvcount)
{
    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;

    void *sbuf = (*env)->GetDirectBufferAddress(env,(jobject)sendbuf);
    void *rbuf = (*env)->GetDirectBufferAddress(env,(jobject)recvbuf);

    MPI_Alltoall(sbuf, sendcount, MPI_BYTE, rbuf, recvcount, MPI_BYTE, mpi_comm);
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeAlltoallv
 * Signature: (JLjava/nio/Buffer;I[I[IILjava/nio/Buffer;I[I[II)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeAlltoallv(JNIEnv *env,
                                                          jobject thisClass,
                                                          jlong comm,
                                                          jobject sendbuf,
                                                          jintArray sendcounts,
                                                          jint sendtype,
                                                          jintArray sdispls,
                                                          jobject recvbuf,
                                                          jintArray recvcounts,
                                                          jint recvtype,
                                                          jintArray rdispls)
{
    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;

    jboolean isCopy;
    int *scounts, *rcounts, *sdisps, *rdisps;

    scounts =(*env)->GetIntArrayElements(env, sendcounts, &isCopy);
    rcounts =(*env)->GetIntArrayElements(env, recvcounts, &isCopy);

    sdisps  =(*env)->GetIntArrayElements(env, sdispls, &isCopy);
    rdisps  =(*env)->GetIntArrayElements(env, rdispls, &isCopy);

    void *sbuf = (*env)->GetDirectBufferAddress(env,(jobject)sendbuf);
    void *rbuf = (*env)->GetDirectBufferAddress(env,(jobject)recvbuf);

    MPI_Alltoallv(sbuf, scounts, sdisps, MPI_BYTE, rbuf, rcounts, rdisps,
            MPI_BYTE, mpi_comm);

    (*env)->ReleaseIntArrayElements(env, sendcounts, scounts, JNI_ABORT);
    (*env)->ReleaseIntArrayElements(env, recvcounts, rcounts, JNI_ABORT);
    (*env)->ReleaseIntArrayElements(env, sdispls, sdisps, JNI_ABORT);
    (*env)->ReleaseIntArrayElements(env, rdispls, rdisps, JNI_ABORT);
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeReduceScatter
 * Signature: (JLjava/nio/Buffer;ILjava/nio/Buffer;I[III)V
 */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeReduceScatter(JNIEnv *env,
                                                              jobject thisClass,
                                                              jlong comm,
                                                              jobject sendbuf,
                                                              jobject recvbuf,
                                                              jintArray recvcounts,
                                                              jint datatype,
                                                              jint op)
{
    jint *rcount;
    jboolean isCopy;

    rcount = (*env)->GetIntArrayElements(env, recvcounts, &isCopy);
    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;

    void *sbuf = (*env)->GetDirectBufferAddress(env,(jobject)sendbuf);
    void *rbuf = (*env)->GetDirectBufferAddress(env,(jobject)recvbuf);

    MPI_Reduce_scatter(sbuf, rbuf, rcount, datatype, op, mpi_comm);

    (*env)->ReleaseIntArrayElements(env, recvcounts, rcount, JNI_ABORT);
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeSplit
 * Signature: (JII)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Intracomm_nativeSplit(JNIEnv  *env,
                                                       jobject thisObject,
                                                       jlong   comm,
                                                       jint    color,
                                                       jint    key)
{
    int errCode;
    MPI_Comm newComm;
    MPI_Comm existingComm = (MPI_Comm) (intptr_t) comm;

    errCode = MPI_Comm_split(comm, color, key, &newComm);


    if (MPI_SUCCESS != errCode) {
        println("ERROR: Could not split communicator.");
    }

    return (jlong) newComm;
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeCreate
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Intracomm_nativeCreate(JNIEnv  *env,
                                                        jobject thisObject,
                                                        jlong   group)
{
    int errCode;
    MPI_Comm newComm;

    MPI_Group existingGroup = (MPI_Group) (intptr_t) group;
    MPI_Comm existingComm = (MPI_Comm)(*env)->GetLongField(env, thisObject,
        FID_mpi_Comm_CommhandleID);

    errCode = MPI_Comm_create(existingComm, existingGroup, &newComm);

    if (MPI_SUCCESS != errCode) {
        println("ERROR: Could not create communicator.");
    }

    return (jlong) newComm;
}

/*
 * Class:     mpi_Intracomm
 * Method:    nativeCreateGroup
 * Signature: (JJI)J
 */
JNIEXPORT jlong JNICALL Java_mpi_Intracomm_nativeCreateGroup(JNIEnv *env,
                                                             jobject thisObject,
                                                             jlong comm,
                                                             jlong group,
                                                             jint tag)
{
    MPI_Comm new_comm;

    int errCode = MPI_Comm_create_group((MPI_Comm)comm, (MPI_Group)group, (int)tag, &new_comm);
    
    if (MPI_SUCCESS != errCode) {
        println("ERROR: Could not create Group.");
    }
    
    return (jlong)new_comm;
}

/*
 *  * Class:     mpi_Intracomm
 *   * Method:    nativeScan
 *    * Signature: (JLjava/nio/Buffer;Ljava/nio/Buffer;III)V
 *     */
JNIEXPORT void JNICALL Java_mpi_Intracomm_nativeScan(JNIEnv *env,
                                                     jobject thisObj,
                                                     jlong comm,
                                                     jobject sendbuf,
                                                     jobject recvbuf,
                                                     jint count,
                                                     jint type,
                                                     jint op) 

{
    void *sbuf = (*env)->GetDirectBufferAddress(env,(jobject)sendbuf);
    void *rbuf = (*env)->GetDirectBufferAddress(env,(jobject)recvbuf);

    MPI_Comm mpi_comm = (MPI_Comm)(intptr_t)comm;

    MPI_Scan(sbuf, rbuf, count, type, op, mpi_comm); 
}
