/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

#include "mpi.h"
#include "mpi_MPI.h"
#include <inttypes.h>
#include "mpi_Commons.h"
#include <stdlib.h>
#include <string.h>

/*
 * Class:     mpi_MPI
 * Method:    nativeWtime
 * Signature: ()D
 */
JNIEXPORT jdouble JNICALL Java_mpi_MPI_nativeWtime(JNIEnv *env,
                                                   jclass thisObj)
{
    return MPI_Wtime();
}

/*
 * Class:     mpi_MPI
 * Method:    SetConstant
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_MPI_SetConstant(JNIEnv *env,
                                                jclass jthis)
{
   jfieldID anysourceID =
       (*env)->GetStaticFieldID(env, jthis, "ANY_SOURCE", "I");
   jfieldID anytagID = (*env)->GetStaticFieldID(env, jthis, "ANY_TAG", "I");
   jfieldID procnullID =
       (*env)->GetStaticFieldID(env, jthis, "PROC_NULL", "I");
   jfieldID graphID = (*env)->GetStaticFieldID(env, jthis, "GRAPH", "I");
   jfieldID cartID = (*env)->GetStaticFieldID(env, jthis, "CART", "I");
   jfieldID bsendoverID =
       (*env)->GetStaticFieldID(env, jthis, "BSEND_OVERHEAD", "I");
   jfieldID undefinedID =
       (*env)->GetStaticFieldID(env, jthis, "UNDEFINED", "I");
   jfieldID identID = (*env)->GetStaticFieldID(env, jthis, "IDENT", "I");
   jfieldID congruentID =
       (*env)->GetStaticFieldID(env, jthis, "CONGRUENT", "I");
   jfieldID similarID = (*env)->GetStaticFieldID(env, jthis, "SIMILAR", "I");
   jfieldID unequalID = (*env)->GetStaticFieldID(env, jthis, "UNEQUAL", "I");
   jfieldID tagubID = (*env)->GetStaticFieldID(env, jthis, "TAG_UB", "I");
   jfieldID hostID = (*env)->GetStaticFieldID(env, jthis, "HOST", "I");
   jfieldID ioID = (*env)->GetStaticFieldID(env, jthis, "IO", "I");

   (*env)->SetStaticIntField(env, jthis, anysourceID, MPI_ANY_SOURCE);
   (*env)->SetStaticIntField(env, jthis, anytagID, MPI_ANY_TAG);
   (*env)->SetStaticIntField(env, jthis, procnullID, MPI_PROC_NULL);
   (*env)->SetStaticIntField(env, jthis, graphID, MPI_GRAPH);
   (*env)->SetStaticIntField(env, jthis, cartID, MPI_CART);
   (*env)->SetStaticIntField(env, jthis, bsendoverID, MPI_BSEND_OVERHEAD);
   (*env)->SetStaticIntField(env, jthis, undefinedID, MPI_UNDEFINED);
   (*env)->SetStaticIntField(env, jthis, identID, MPI_IDENT);
   (*env)->SetStaticIntField(env, jthis, congruentID, MPI_CONGRUENT);
   (*env)->SetStaticIntField(env, jthis, similarID, MPI_SIMILAR);
   (*env)->SetStaticIntField(env, jthis, unequalID, MPI_UNEQUAL);
   (*env)->SetStaticIntField(env, jthis, tagubID, MPI_TAG_UB);
   (*env)->SetStaticIntField(env, jthis, hostID, MPI_HOST);
   (*env)->SetStaticIntField(env, jthis, ioID, MPI_IO);
}

/*
 * Class:     mpi_MPI
 * Method:    nativeInit
 * Signature: ([Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_mpi_MPI_nativeInit(JNIEnv *env,
                                               jclass thisClass,
                                               jobjectArray args)
{
    int len = (*env)->GetArrayLength(env, args);
    char **sargs = (char **) calloc(len + 1, sizeof(char *));
    int i;
    int flag;

    for (i = 0; i < len; i++) {
      jstring jc = (jstring)(*env)->GetObjectArrayElement(env, args, i);
      const char* t = (*env)->GetStringUTFChars(env, jc, 0);
      int slen = strlen(t) + 1;
      sargs[i] = (char *) calloc(slen, sizeof(char));
      strcpy(sargs[i], t);
      (*env)->ReleaseStringUTFChars(env, jc, t);
      (*env)->DeleteLocalRef(env, jc);
    }

    if (MPI_SUCCESS != MPI_Init(&len, &sargs)) {
      fprintf(stderr, "Error in MPI_Init() \n");
    }

    for (i = 0; i < len; i++) free(sargs[i]);
    free(sargs);

    //debugging block
    int rank, len_of_proc_name;
    char name[MPI_MAX_PROCESSOR_NAME];
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);
    MPI_Get_processor_name(name, &len_of_proc_name);
    fflush(stdout);
}

/*
 * Class:     mpi_Comm
 * Method:    nativeFinish
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_MPI_nativeFinish(JNIEnv *env,
                                                 jclass thisClass)
{
    if (MPI_SUCCESS != MPI_Finalize()) {
        fprintf(stderr, "Error in MPI_Finalize() \n");
    }
}

/*
 * Class:     mpi_MPI
 * Method:    nativeInitThread
 * Signature: ([Ljava/lang/String;I)I
 */
JNIEXPORT jint JNICALL Java_mpi_MPI_nativeInitThread(JNIEnv *env,
                                                     jclass thisClass,
                                                     jobjectArray args,
                                                     jint required)
{
    int len = (*env)->GetArrayLength(env, args);
    char **sargs = (char **) calloc(len+1, sizeof(char *));
    int i;
    int flag;

    for (i = 0; i < len; i++) {
        jstring jc = (jstring)(*env)->GetObjectArrayElement(env, args, i);
        const char *t = (*env)->GetStringUTFChars(env, jc, 0);
        int slen = strlen(t) + 1;
        sargs[i] = (char *) calloc(slen, sizeof(char));
        strcpy(sargs[i], t);
        (*env)->ReleaseStringUTFChars(env, jc, t);
        (*env)->DeleteLocalRef(env, jc);
    }

    int provided;
    int rc = MPI_Init_thread(&len, &sargs, required, &provided);
    if (MPI_SUCCESS != rc) {
        fprintf(stderr, "Error in MPI_Init() \n");
    }

    for (i = 0; i < len; i++) free(sargs[i]);
    free(sargs);

    int rank, len_of_proc_name;
    char name[MPI_MAX_PROCESSOR_NAME];
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);
    MPI_Get_processor_name(name, &len_of_proc_name);
    fflush(stdout);

    return provided;
}
