/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
 The MIT License

 Copyright (c) 2013 - 2014
 1. SEECS, National University of Sciences and Technology, Pakistan (2013 - 2014)
 2. Bibrak Qamar  (2013 - 2014)

 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the
 "Software"), to deal in the Software without restriction, including
 without limitation the rights to use, copy, modify, merge, publish,
 distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to
 the following conditions:

 The above copyright notice and this permission notice shall be included
 in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * File         : mpi_Request.c
 * Author       : Bibrak Qamar
 * Created      : Tue Sept  17 11:15:11 PKT 2013
 * Revision     : $Revision: 1.1 $
 * Updated      : $Date: 2014/03/11 11:30:00 $
 *
 */

#include "mpi.h"
#include "mpi_Request.h"
#include "mpi_Commons.h"
#include "stdlib.h"

/*
 * Class:     mpi_Request
 * Method:    Wait
 * Signature: (Lmpjdev/Status;)Lmpjdev/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_Wait(JNIEnv *env,
                                                jobject thisObject,
                                                jobject stat)
{
    int elements;

    MPI_Request request = (MPI_Request)((*env)->GetLongField(env,
        thisObject, FID_mpi_Request_handle));

    MPI_Status *mpi_status =
        (MPI_Status *)((*env)->GetLongField(env, stat, FID_mpi_Status_handle));
    
    MPI_Wait(&request, mpi_status);
    
    MPI_Get_count(mpi_status, MPI_BYTE, &elements);

    (*env)->SetIntField(env, stat, FID_mpi_Status_source,
            mpi_status->MPI_SOURCE);
    (*env)->SetIntField(env, stat, FID_mpi_Status_tag, mpi_status->MPI_TAG);
    (*env)->SetIntField(env, stat, FID_mpi_Status_numEls, elements);

    return stat;
}

/*
 * Class:     mpi_Request
 * Method:    Test
 * Signature: (Lmpjdev/Status;)Lmpjdev/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_Test(JNIEnv *env,
                                                jobject thisObject,
                                                jobject stat)
{
    int flag;

    MPI_Request request = (MPI_Request)((*env)->GetLongField(env,
        thisObject, FID_mpi_Request_handle));
    MPI_Status *mpi_status =
        (MPI_Status *)((*env)->GetLongField(env, stat, FID_mpi_Status_handle));
    MPI_Test(&request, &flag, mpi_status);

    if (flag) {
        int elements;
        MPI_Get_count(mpi_status, MPI_BYTE, &elements);
        (*env)->SetIntField(env, stat, FID_mpi_Status_source,
                mpi_status->MPI_SOURCE);
        (*env)->SetIntField(env, stat, FID_mpi_Status_tag, mpi_status->MPI_TAG);
        (*env)->SetIntField(env, stat, FID_mpi_Status_numEls, elements);

        return stat;
    } else {
        return NULL;
    }
}

/*
 * Class:     mpi_Request
 * Method:    GetReq
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_mpi_Request_GetReq(JNIEnv  *env,
					       jobject jthis,
					       jint    type)
{
    switch (type) {
        case 0:
            (*env)->SetLongField(env, jthis,
                    FID_mpi_Request_handle,
                    (jlong)MPI_REQUEST_NULL);
            break;
        default:
            break;
    }
}

/*
 * Class:     mpi_NativeRequest
 * Method:    nativeWaitAll
 * Signature: ([Lmpi/Request;)[Lmpi/Status;
 */
JNIEXPORT jobjectArray JNICALL Java_mpi_Request_nativeWaitAll(JNIEnv   *env,
							  jclass       jthis,
						          jobjectArray requests)
{
    int i;
    int count = (*env)->GetArrayLength(env, requests);

    MPI_Request *reqs = (MPI_Request*)malloc(count * sizeof(MPI_Request));
    MPI_Status *stas = (MPI_Status*)malloc(count * sizeof(MPI_Status));
    jobjectArray array_of_status =
        (*env)->NewObjectArray(env, count, CL_mpi_Status, NULL);

    /* Copy initial native requests in Java `requests' to `reqs'. */
    for (i = 0; i < count; i++) {
        reqs[i] = (MPI_Request)((*env)->GetLongField(env,
                    (*env)->GetObjectArrayElement(env, requests, i),
          	    FID_mpi_Request_handle));
    }

    MPI_Waitall(count, reqs, stas);

    int mpiCount;
    jboolean isBufferComms;

    for (i = 0; i < count; i++) {
        isBufferComms = (MPI_Request)((*env)->GetBooleanField(env,
            (*env)->GetObjectArrayElement(env, requests, i),
            FID_mpi_Request_isBufferComms));

        jobject req  = (*env)->GetObjectArrayElement(env, requests, i);
        jobject jstas = (*env)->NewObject(env, CL_mpi_Status,
            MID_mpi_Status_constructor, isBufferComms);
        MPI_Status *status =
            (MPI_Status *)((*env)->GetLongField(env, jstas, FID_mpi_Status_handle));

        /* Copy final native request to `array_of_request'. */
        (*env)->SetLongField(env, req,
            FID_mpi_Request_handle, (jlong) reqs[i]);

        /* Copy final native status to Java `array_of_status'... */
        *status = stas[i];
        (*env)->SetIntField(env, jstas, FID_mpi_Status_source,
            stas[i].MPI_SOURCE);
        (*env)->SetIntField(env, jstas, FID_mpi_Status_tag, stas[i].MPI_TAG);
        MPI_Get_count(&(stas[i]), MPI_BYTE, &mpiCount);
        (*env)->SetIntField(env, jstas, FID_mpi_Status_numEls, mpiCount);
        (*env)->SetIntField(env, jstas, FID_mpi_Status_count, mpiCount);
        (*env)->SetObjectArrayElement(env, array_of_status, i, jstas);
        (*env)->CallVoidMethod(env, (*env)->GetObjectArrayElement(env, requests, i),
            MID_mpi_Request_complete, jstas);
        (*env)->DeleteLocalRef(env, req);
        (*env)->DeleteLocalRef(env, jstas);
    }

    free(reqs);
    free(stas);

    return array_of_status;
}

/*
 * Class:     mpi_NativeRequest
 * Method:    nativeTestAll
 * Signature: ([Lmpi/Request;)[Lmpi/Status;
 */
JNIEXPORT jobjectArray JNICALL Java_mpi_Request_nativeTestAll(JNIEnv   *env,
							  jclass       jthis,
						          jobjectArray requests)
{
    int flag, i;
    int count = (*env)->GetArrayLength(env, requests);

    MPI_Request *reqs = (MPI_Request*)malloc(count * sizeof(MPI_Request));
    MPI_Status *stas = (MPI_Status*)malloc(count * sizeof(MPI_Status));
    jobjectArray array_of_status =
        (*env)->NewObjectArray(env, count, CL_mpi_Status, NULL);

    /* Copy initial native requests in Java `requests' to `reqs'. */
    for (i = 0; i < count; i++) {
        reqs[i] = (MPI_Request)((*env)->GetLongField(env,
                    (*env)->GetObjectArrayElement(env, requests, i),
          	    FID_mpi_Request_handle));
    }

    MPI_Testall(count, reqs, &flag, stas);

    if (flag) {
        int mpiCount;
        jboolean isBufferComms;

        for (i = 0; i < count; i++) {
            isBufferComms = (MPI_Request)((*env)->GetBooleanField(env,
                (*env)->GetObjectArrayElement(env, requests, i),
                FID_mpi_Request_isBufferComms));
            jobject req  = (*env)->GetObjectArrayElement(env, requests, i);
            jobject jstas = (*env)->NewObject(env, CL_mpi_Status,
                MID_mpi_Status_constructor, isBufferComms);
            MPI_Status *status =
                (MPI_Status *)((*env)->GetLongField(env,
                                                    jstas,
                                                    FID_mpi_Status_handle));

            /* Copy final native request to `array_of_request'. */
            (*env)->SetLongField(env, req,
                FID_mpi_Request_handle, (jlong) reqs[i]);

            /* Copy final native status to Java `array_of_status'... */
            *status = stas[i];
            (*env)->SetIntField(env, jstas, FID_mpi_Status_source,
                stas[i].MPI_SOURCE);
            (*env)->SetIntField(env, jstas, FID_mpi_Status_tag,
                    stas[i].MPI_TAG);
            MPI_Get_count(&(stas[i]), MPI_BYTE, &mpiCount);
            (*env)->SetIntField(env, jstas, FID_mpi_Status_numEls, mpiCount);
            (*env)->SetIntField(env, jstas, FID_mpi_Status_count, mpiCount);
            (*env)->SetObjectArrayElement(env, array_of_status, i, jstas);
            (*env)->CallVoidMethod(env,
                    (*env)->GetObjectArrayElement(env, requests, i),
                    MID_mpi_Request_complete, jstas);
            (*env)->DeleteLocalRef(env, req);
            (*env)->DeleteLocalRef(env, jstas);
        }
    }

    free(reqs);
    free(stas);

    if (flag) {
        return array_of_status;
    } else {
        return NULL;
    }
}

/*
 * Class:     mpi_Request
 * Method:    nativeWaitAny
 * Signature: ([Lmpi/Request;)Lmpi/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_nativeWaitAny(JNIEnv	      *env,
							 jclass       jthis,
						         jobjectArray requests)
{
    int i, index;
    int count = (*env)->GetArrayLength(env, requests);

    MPI_Request *reqs = (MPI_Request*)malloc(count * sizeof(MPI_Request));
    MPI_Status *stas = (MPI_Status*)malloc(1 * sizeof(MPI_Status));

    /* Copy initial native requests in Java `requests' to `reqs'. */
    for (i = 0; i < count; i++) {
        reqs[i] = (MPI_Request)((*env)->GetLongField(env,
                    (*env)->GetObjectArrayElement(env, requests, i),
          	    FID_mpi_Request_handle));
    }

    MPI_Waitany(count, reqs, &index, stas);

    for (i = 0; i < count; i++) {
        (*env)->SetLongField(env,
                (*env)->GetObjectArrayElement(env, requests, i),
                FID_mpi_Request_handle, (jlong) reqs[i]);
    }

    jobject jstas = NULL;
    int mpiCount;
    int mpiNumEls;

    int elements;
    jobject req = (*env)->GetObjectArrayElement(env, requests, index);
    jboolean isBufferComms = (MPI_Request)((*env)->GetBooleanField(env, req,
        FID_mpi_Request_isBufferComms));
    jstas = (*env)->NewObject(env, CL_mpi_Status,
        MID_mpi_Status_constructor, isBufferComms);
    MPI_Status *status =
        (MPI_Status *)((*env)->GetLongField(env,
                                            jstas,
                                            FID_mpi_Status_handle));

    /* Copy final native status to Java `array_of_status'... */
    *status = *stas;

    (*env)->SetIntField(env, jstas, FID_mpi_Status_source,
        stas[0].MPI_SOURCE);
    (*env)->SetIntField(env, jstas, FID_mpi_Status_tag, stas[0].MPI_TAG);
    MPI_Get_count(&(stas[0]), MPI_BYTE, &mpiCount);
    MPI_Get_elements(&(stas[0]), MPI_BYTE, &mpiNumEls);
    (*env)->SetIntField(env, jstas, FID_mpi_Status_numEls, mpiCount);
    (*env)->SetIntField(env, jstas, FID_mpi_Status_count, mpiNumEls);
    (*env)->SetIntField(env, jstas, FID_mpi_Status_index, index);
    (*env)->CallVoidMethod(env, req, MID_mpi_Request_complete, jstas);
    (*env)->DeleteLocalRef(env, req);

    free(reqs);
    free(stas);

    return jstas;
}

/*
 * Class:     mpi_Request
 * Method:    nativeTestAny
 * Signature: ([Lmpi/Request;)Lmpi/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_nativeTestAny(JNIEnv	      *env,
							 jclass       jthis,
						         jobjectArray requests)
{
    int flag, i, index;
    int count = (*env)->GetArrayLength(env, requests);

    MPI_Request *reqs = (MPI_Request*)malloc(count * sizeof(MPI_Request));
    MPI_Status *stas = (MPI_Status*)malloc(1 * sizeof(MPI_Status));

    /* Copy initial native requests in Java `requests' to `reqs'. */
    for (i = 0; i < count; i++) {
        reqs[i] = (MPI_Request)((*env)->GetLongField(env,
                    (*env)->GetObjectArrayElement(env, requests, i),
          	    FID_mpi_Request_handle));
    }

    MPI_Testany(count, reqs, &index, &flag, stas);

    for (i = 0; i < count; i++) {
        (*env)->SetLongField(env,
            (*env)->GetObjectArrayElement(env, requests, i),
            FID_mpi_Request_handle, (jlong) reqs[i]);
    }

    jobject jstas = NULL;
    int mpiCount;
    int mpiNumEls;

    if (flag) {
        int elements;
        jobject req = (*env)->GetObjectArrayElement(env, requests, index);
        jboolean isBufferComms = (MPI_Request)((*env)->GetBooleanField(env,
                    req, FID_mpi_Request_isBufferComms));
        jstas = (*env)->NewObject(env, CL_mpi_Status,
            MID_mpi_Status_constructor, isBufferComms);
        MPI_Status *status =
            (MPI_Status *)((*env)->GetLongField(env,
                                                jstas,
                                                FID_mpi_Status_handle));

        /* Copy final native status to Java `array_of_status'... */
        *status = *stas;
        (*env)->SetIntField(env, jstas, FID_mpi_Status_source,
            stas[0].MPI_SOURCE);
        (*env)->SetIntField(env, jstas, FID_mpi_Status_tag, stas[0].MPI_TAG);
        MPI_Get_count(&(stas[0]), MPI_BYTE, &mpiCount);
        MPI_Get_elements(&(stas[0]), MPI_BYTE, &mpiNumEls);
        (*env)->SetIntField(env, jstas, FID_mpi_Status_count, mpiCount);
        (*env)->SetIntField(env, jstas, FID_mpi_Status_numEls, mpiNumEls);
        (*env)->SetIntField(env, jstas, FID_mpi_Status_index, index);
        (*env)->CallVoidMethod(env, req, MID_mpi_Request_complete, jstas);
        (*env)->DeleteLocalRef(env, req);
    }

    free(reqs);
    free(stas);

    if (flag) {
        return jstas;
    } else {
        return NULL;
    }
}
