/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
 The MIT License

 Copyright (c) 2005 - 2008
   1. Distributed Systems Group, University of Portsmouth (2005)
   2. Community Grids Laboratory, Indiana University (2005)
   3. Aamir Shafi (2005 - 2008)
   4. Bryan Carpenter (2005 - 2008)
   5. Mark Baker (2005 - 2008)

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
 */
/*
 * File         : Cartcomm.java
 * Author       : Aamir Shafi, Bryan Carpenter
 * Created      : Fri Sep 10 12:22:15 BST 2004
 * Revision     : $Revision: 1.11 $
 * Updated      : $Date: 2005/09/28 13:33:59 $
 */

package mpi;

/**
 * This class represents a communicator with a cartesian structure. 
 */
public class Cartcomm extends Intracomm {

  CartParms cartParms = null;

  // so the index of this cartParms will be the rank of that process

  Cartcomm(int[] dims, boolean[] periods, boolean reorder, mpi.Comm mpjdevComm, Group group)
      throws MPIException {

    super(mpjdevComm, group);

    int rank = group.getRank();

    cartParms = new CartParms();

    /*
     * Bug: Using 'copying by reference' technique to store dims and
     * periods array in cartParms object. This is detrimental because if
     * the value of dims or periods array is changed later, it also
     * changes the value of stored dims or periods array.
     * Identified by: Andre Vehreschild
     * Fixed: Wed Sep 28 12:30:16 GMT 2005
     */
    cartParms.dims = new int[dims.length];
    System.arraycopy(dims, 0, cartParms.dims, 0, dims.length);

    cartParms.periods = new boolean[periods.length];
    System.arraycopy(periods, 0, cartParms.periods, 0, cartParms.periods.length);

    try {
      cartParms.coords = Coords(rank);
    } catch (Exception e) {
      throw new MPIException(e);
    }
  }

  /**
   * Select a balanced distribution of processes per coordinate direction.
   *
   * @param nnodes number of nodes in a grid.
   * @param dims array specifying number of nodes in each dimension.
   *
   * Java binding of the MPI operation {@code MPI_DIMS_CREATE}.
   *
   * Note that {@code dims} is an inout parameter.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static void createDims(int nnodes, int[] dims) throws MPIException {

    int rank = dims.length;

    while (rank > 0) {

      int ext = root(nnodes, rank);
      rank--;
      nnodes /= ext;
      dims[rank] = ext;
    }
  }

  static int root(int n, int d) {

    switch (d) {
      case 1:
        return n;
      case 2:
        return (int) Math.sqrt(n);
      default:
        int ceiling = 1;
        int power = 1;
        while (power < n) {
          ceiling++;
          power = ceiling;
          for (int i = 1; i < d; i++) power *= ceiling;
        }
        if (power == n) return ceiling;
        else return ceiling - 1;
    }
  }

  /**
   * Clones this communicator. 
   * 
   * <p>Java binding of MPI operation {@code MPI_COMM_DUP}.
   * 
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Cartcomm clone() throws MPIException {
    return null;
  }

  /**
   * Translate logical process coordinates to process rank.
   *
   * @param coords Cartesian coordinates of a process.
   * @return rank of the specified process.
   *
   * <p>Java binding of the MPI operation {@code MPI_CART_RANK}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int getRank(int[] coords) throws MPIException {
    int factor = 1;
    int rank = 0;
    int i = cartParms.dims.length - 1;
    int d = i; // d is basically dims index ....
    int c = i; // c is coords index ...
    int ord = 0;
    int dim = 0;

    for (; i >= 0; --i, --c, --d) {
      dim = (cartParms.dims[i] > 0) ? cartParms.dims[i] : -(cartParms.dims[i]);
      ord = coords[c];

      if ((ord < 0) || (ord >= dim)) {
        ord %= dim;

        if (ord < 0) {
          ord += dim;
        }
      }

      rank += factor * ord;
      factor *= dim;
    }

    return rank;
  }

  /**
   * Translate process rank to logical process coordinates.
   *
   * @param rank rank of a process.
   * @return Cartesian coordinates of the specified process.
   *
   * <p>Java binding of the MPI operation {@code MPI_CART_COORDS}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int[] Coords(int rank) throws MPIException {
    int dim = 0;
    int nprocs = group.getSize();
    int remprocs = nprocs;
    int orank = rank; // just for debugging
    int[] coords = new int[cartParms.dims.length];

    for (int i = 0; i < cartParms.dims.length; ++i) {
      dim = cartParms.dims[i];
      remprocs /= dim;
      coords[i] = rank / remprocs;
      rank %= remprocs;
    }

    return coords;
  }

  /**
   * Compute source and destination ranks for ``shift'' communication.
   *
   * @param direction coordinate dimension of shift.
   * @param disp displacement.
   * @return object containing ranks of source and destination processes.
   *
   * <p>Java binding of the MPI operation {@code MPI_CART_SHIFT}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public ShiftParms shift(int direction, int disp) throws MPIException {

    ShiftParms sParms = new ShiftParms();
    int factor = 0;
    int thisdim = 0;
    int thisperiod = 0;
    int ord = 0;
    int srcord = 0;
    int destord = 0;
    int i = 0;
    int p = 0;
    int rank_src = 0;
    int rank_dst = 0;

    ord = this.group.getRank();

    /*
     * Naive case ...
     */
    if (disp == 0) {
      rank_src = ord;
      rank_dst = ord;
      sParms.rank_source = rank_src;
      sParms.rank_dest = rank_dst;
      return sParms;
    }

    /* this is MPICH cloning */
    boolean periodic = cartParms.periods[direction];
    // save position ..dont understand this ...
    int saved_position = cartParms.coords[direction];
    int dest_position, source_position;
    dest_position = source_position = saved_position;

    // Shift for the destination ...
    dest_position += disp;

    if (dest_position >= cartParms.dims[direction]) {
      if (periodic) dest_position %= cartParms.dims[direction];
      else dest_position = -1;
    } else if (dest_position < 0) {
      if (periodic) dest_position += cartParms.dims[direction];
      else dest_position = -1;
    }

    cartParms.coords[direction] = dest_position;

    if (dest_position != -1) {
      rank_dst = getRank(cartParms.coords);
    } else {
      rank_dst = -1;
    }

    // Shift for the source ..
    source_position -= disp;

    if (source_position >= cartParms.dims[direction]) {
      if (periodic) {
        source_position %= cartParms.dims[direction];
      } else source_position = -1;
    } else if (source_position < 0) {
      if (periodic) {
        source_position += cartParms.dims[direction];
      } else source_position = -1;
    }

    cartParms.coords[direction] = source_position;

    if (source_position != -1) {
      rank_src = getRank(cartParms.coords);
    } else {
      rank_src = -1;
    }

    // Restore my position ..
    cartParms.coords[direction] = saved_position;

    sParms.rank_source = rank_src;
    sParms.rank_dest = rank_dst;

    return sParms;
  }

  /**
   * Partition Cartesian communicator into subgroups of lower dimension.
   *
   * @param remain_dims by dimension, {@code true} if dimension is to be kept,
   *                    {@code false} otherwise.
   * @return communicator containing subgrid including this process.
   *
   * <p>Java binding of the MPI operation {@code MPI_CART_SUB}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Cartcomm sub(boolean[] remain_dims) throws MPIException {
    int color = 0;
    int key = 0;
    int keyfactor = 1;
    int colfactor = 1;

    int ndim = 0;
    int dim = 0;
    int allfalse = 0;
    int i = cartParms.dims.length - 1;

    for (; i >= 0; --i) {
      dim = cartParms.dims[i];

      if (remain_dims[i] == false) {
        color += colfactor * cartParms.coords[i];
        colfactor *= dim;
      } else {
        ++ndim;
        key += keyfactor * (cartParms.coords[i]);
        keyfactor *= dim;
      }
    }

    /* special case ...comm with just yourself 'this' process init */
    if (ndim == 0) {
      color = this.group.getRank();
      ndim = 1;
      allfalse = 1;
      // MPICH makes a dup of MPI_COMM_SELF and returns it.
    }

    /* split the communicator */
    Intracomm newcomm = this.split(color, key);

    /* fill the comm with topology information */
    /* and compute the callers coordinate ... */
    if (newcomm != null) {
      // get total number of processes in this comm ...
      int nprocs = newcomm.getSize();
      int[] ndimsArray = new int[ndim];
      boolean[] nperiods = new boolean[ndim];

      int index = 0;

      if (allfalse == 1) {
        ndimsArray[0] = 1;
      } else {
        for (int l = 0; l < cartParms.dims.length; l++) {
          if (remain_dims[l] == true) {
            ndimsArray[index] = cartParms.dims[l];
            nperiods[index] = cartParms.periods[l];
            index++;
          }
        }
      }

      return newcomm.createCart(ndimsArray, nperiods, false);
    } else {
      return null;
    }
  }

  /**
   * Compute an optimal placement.
   *
   * @param dims the number of processes in each dimnesion.
   * @param periods {@code true} if grid is periodic in each dimension,
   *                  otherwise {@code false}.
   * @return reordered rank of calling process.
   *
   * <p>Java binding of the MPI operation {@code MPI_CART_MAP}.
   *
   * <p>The number of dimensions is taken to be size of the {@code dims} argument.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int map(int[] dims, boolean[] periods) throws MPIException {
    int procs = 1;

    for (int i = 0; i < dims.length; i++) {
      if (dims[i] < 0) {
        throw new MPIException(" Error in Cartcomm.Map: dims[" + i + "] is " + "less than zero");
      }
      procs *= dims[i];
    }

    int size = this.group.getSize();
    int rank = this.group.getRank();

    if (procs > size) {
      throw new MPIException(
          " Error in Cartcomm.Map: procs <" + procs + "> is " + "greater than size <" + size + ">");
    }

    return ((rank < 0) || (rank >= procs)) ? -1 : rank; // -1 should UNDEF ...
  }

  /**
   * Returns Cartesian topology information.
   *
   * @return object containing dimensions,
   *           periods and local coordinates.
   *
   * <p>Java binding of the MPI operations {@code MPI_CARTDIM_GET} and {@code MPI_CART_GET}.
   *
   * <p>The number of dimensions can be obtained from the size of (eg) {@code dims} field of the
   * returned object.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public CartParms getTopo() throws MPIException {
    return cartParms;
  }

  /**
   * Returns the type of topology associated with the communicator.
   *
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * @return topology type of communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_TOPO_TEST}.
   *
   * <p>The return value will be one of {@code MPI.GRAPH}, {@code MPI.CART} or
   * {@code MPI.UNDEFINED}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int getTopology() throws MPIException {
    return MPI.CART; // this means no topology ....
  }
}
