/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
 The MIT License

 Copyright (c) 2005 - 2008
   1. Distributed Systems Group, University of Portsmouth (2005)
   2. Community Grids Laboratory, Indiana University (2005)
   3. Aamir Shafi (2005 - 2008)
   4. Bryan Carpenter (2005 - 2008)
   5. Mark Baker (2005 - 2008)

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * File         : Comm.java
 * Author       : Aamir Shafi, Bryan Carpenter
 * Created      : Fri Sep 10 12:22:15 BST 2004
 * Revision     : $Revision: 1.60 $
 * Updated      : $Date: 2014/03/11 13:36:40 $
 */

package mpi;

import mpjbuf.*;
import java.lang.reflect.Method;
import java.nio.ByteBuffer;
import java.util.Hashtable;

/**
 * This class represents communicators.
 */
public class Comm {

  public static final long MPI_COMM_NULL_VAL = 67108864;
  static final int NULL = 0;
  static final int SELF = 1;
  static final int WORLD = 2;
  public long handle = -1;
  Group group = null;

  Group localgroup = null;

  Comm mpjdevComm = null;
  private Hashtable attrTable = new Hashtable();

  Comm() {}

  // This constructor is used to create the following communicators:
  //    - COMM_NULL (from MPI.java)
  //    - COMM_SELF (from MPI.java)
  //    - COMM_WORLD (from MPI.java via Intracomm(int code)
  Comm(int code) {
    GetComm(code);

    if (code != NULL) {
      long groupHandle = group();
      this.group = new Group(groupHandle);
    }
  }

  // Comm.java: create(int[] ids)
  // Comm.java: split(int color, int key)
  // Comm.java: Comm create(Group ngroup)
  // Comm.java: dup() and clone()
  Comm(long handle) throws MPIException {

    setHandle(handle);
    if (handle != MPI_COMM_NULL_VAL) {
      long groupHandle = group();
      this.group = new Group(groupHandle);
    }
  }

  // Comm.java: createIntercomm
  public Comm(long handle, Group localgroup_) throws MPIException {

    setHandle(handle);
    long groupHandle = group();
    this.group = new Group(groupHandle);
    this.localgroup = localgroup_;
  } 

  /**
   * Size of group of this communicator.
   *
   * @return number of processors in the group of this communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_SIZE}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int getSize() throws MPIException {
    return size(handle);
  }

  /**
   * Rank of this process in group of this communicator.
   *
   * @return rank of the calling process in the group of this communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_RANK}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int getRank() throws MPIException {
    return rank(handle);
  }

  protected long getHandle() {
    return this.handle;
  }

  private void setHandle(long handle) throws MPIException {
    this.handle = handle;
  }

  public final boolean isNull() {
    return this.handle == MPI_COMM_NULL_VAL;
  }

  // mpjdev method
  public Comm create(int[] ids) throws MPIException {

    long newComm = create(handle, ids); // this create is native

    // It means this process will be outside the group..
    if (newComm == 0) return null;
    else return new Comm(newComm);
  }

  /**
   * Destroy this communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_FREE}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void free() throws MPIException {
    free(handle);
  }

  /**
   * Return group associated with a communicator.
   *
   * @return group corresponding to this communicator group.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_GROUP}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Group getGroup() throws MPIException {
    return this.group;
  }

  /**
   * Test if this communicator is an inter-communicator.
   *
   * @return {@code true} if this is an inter-communicator, {@code false}
   *           otherwise.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_TEST_INTER}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public boolean isInter() throws MPIException {
    return Test_inter();
  }

  /**
   * Create an inter-communicator.
   *
   * @param local_comm local intra-communicator.
   * @param local_leader rank of local group leader in {@code localComm}.
   * @param remote_leader rank of remote group leader in this communicator.
   * @param tag ``tag'' tag.
   * @return new inter-communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_INTERCOMM_CREATE}.
   *
   * <p>(This operation is defined as a method on the ``peer communicator'', making it analogous to
   * a {@code send} or {@code recv} communication with the remote group leader.)
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Intercomm createIntercomm(Comm local_comm, int local_leader, int remote_leader, int tag)
      throws MPIException {

    long newIntercommValue =
        nativeCreateIntercomm(handle, local_comm.getHandle(), local_leader, remote_leader, tag);

    return new Intercomm(newIntercommValue);
  }

  /**
   * Retrieves attribute value by key.
   *
   * @param keyval one of the key values predefined by the implementation.
   * @return attribute value.
   *
   * <p>Java binding of the MPI operation {@code MPI_ATTR_GET}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Object getAttr(int keyval) throws MPIException {
    Object value = attrTable.get(new Integer(keyval));

    if (value != null) {
      return value;
    } else {
      throw new MPIException(
          " Error in attrGet: No corresponding value " + "to key <" + keyval + ">");
    }
  }

  /**
   * deletes attribute value by key.
   *
   * @param keyval one of the key values predefined by the implementation.
   *   
   * <p>Java binding of the MPI operation {@code MPI_ATTR_DELETE}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void deleteAttr(int keyval) throws MPIException {
    try {
      attrTable.remove(new Integer(keyval));
    } catch (Exception e) {
      throw new MPIException(e);
    }
  }

  /**
   * Attribute put method which can be used to add attributed to this communicator
   *
   * @param keyval one of the key values predefined by the implementation.
   * @param value the key's value.
   *
   * <p>Java binding of the MPI operaiton {@code MPI_COMM_SET_ATTR}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  void setAttr(int keyval, Object value) throws MPIException {
    try {
      attrTable.put(new Integer(keyval), value);
    } catch (Exception e) {
      throw new MPIException(e);
    }
  }

  /**
   * Blocking send operation.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in send buffer.
   * @param dest rank of destination.
   * @param tag message tag.
   *
   * <p>Java binding of the MPI operation {@code MPI_SEND}.
   *
   * <p>The actual argument associated with {@code buf} must be one-dimensional array.
   *
   * <p>If the {@code datatype} argument represents an MPI basic type, its value must agree with
   * the element type of {@code buf}---either a primitive type or a reference (object) type. If the
   * {@code datatype} argument represents an MPI derived type, its <em>base type</em> must agree
   * with the element type of {@code buf}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void send(Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    if (buf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      long commHandle = this.getHandle();
      nativeByteBufferSend(
        commHandle, (java.nio.ByteBuffer) buf, dest, count, tag);
      return;
    }

    try {
      arraysSend(buf, count, datatype, dest, tag);
    } catch (MPIException mpie) {
      throw mpie;
    }
  }

  private void arraysSend(
      Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    Packer packer = datatype.getPacker();

    mpjbuf.Buffer wBuffer = datatype.createWriteBuffer(count);

    try {
      packer.pack(wBuffer, buf, 0, count);
      wBuffer.commit();
      mv2j_send(wBuffer, datatype, dest, tag);
      wBuffer.clear();
      wBuffer.free(); // gadget memory leak fix?
      BufferFactory.destroy(wBuffer.getStaticBuffer());
    } catch (Exception e) {
      throw new MPIException(e);
    }
  }

  /**
   * Send in synchronous mode.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of items to send.
   * @param dest rank of destination.
   * @param tag message tag.
   *
   * <p>Java binding of the MPI operation {@code MPI_SSEND}.
   *
   * <p>Further comments as for {@code Send}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void sSend(Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    if (buf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      } 
      long commHandle = this.getHandle();
      nativeByteBufferSsend(commHandle, (java.nio.ByteBuffer) buf, dest, count, tag);
      return;
    }

    try {
      arraysSsend(buf, count, datatype, dest, tag);
    } catch (MPIException mpie) {
      throw mpie;
    }
  }

  private void arraysSsend(
      Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    Packer packer = datatype.getPacker();
    
    mpjbuf.Buffer wBuffer = datatype.createWriteBuffer(count);

    try {
      packer.pack(wBuffer, buf, 0, count);
      wBuffer.commit();
      mv2j_ssend(wBuffer, datatype, dest, tag);
      wBuffer.clear(); // bug-fix
      wBuffer.free(); // gadget memory leak fix?
      BufferFactory.destroy(wBuffer.getStaticBuffer());
    } catch (Exception e) {
      throw new MPIException(e);
    }
  }

  /**
   * Send in buffered mode.
   *
   * <p style="color:red"> Not supported in the current release 
   * as it relies on {@code MPI.attachBuffer} and {@code MPI.detachBuffer}</p>.
   *
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in send buffer.
   * @param dest rank of destination.
   * @param tag message tag.
   *
   * <p>Java binding of the MPI operation {@code MPI_BSEND}.
   *
   * <p>This operation copies message into a buffer (mpi.Buffer) specified by MPI.Buffer_attach
   * operation, and then sends using the standard mode of communication. Further comments as for
   * {@code Send}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void bSend(Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {
   
    if (buf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      long commHandle = this.getHandle();
      nativeByteBufferBsend(commHandle, (java.nio.ByteBuffer) buf, dest, count, tag);
      return;
    }

    try {
      arraysBsend(buf, count, datatype, dest, tag);
    } catch (MPIException mpie) {
      throw mpie;
    }
  }

  private void arraysBsend(
      Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    Packer packer = datatype.getPacker();

    mpjbuf.Buffer wBuffer = datatype.createWriteBuffer(count);

    try {
      packer.pack(wBuffer, buf, 0, count);
      wBuffer.commit();
      mv2j_bsend(wBuffer, dest, tag);
      wBuffer.clear();
      wBuffer.free(); // gadget memory leak fix?
      BufferFactory.destroy(wBuffer.getStaticBuffer());
    } catch (Exception e) {
      throw new MPIException(e);
    }
  }

  /**
   * Send in ready mode.
   * 
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   * 
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in send buffer.
   * @param dest rank of destination.
   * @param tag message tag.
   *
   * <p>Java binding of the MPI operation {@code MPI_RSEND}.
   *
   * <p>Further comments as for {@code Send}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void rSend(Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    if (buf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      long commHandle = this.getHandle();
      nativeByteBufferRsend(
        commHandle, (java.nio.ByteBuffer) buf, dest, count, tag);
      return;
    }

    try {
      arraysRsend(buf, count, datatype, dest, tag);
    } catch (MPIException mpie) {
      throw mpie;
    }
  }

  private void arraysRsend(
      Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    Packer packer = datatype.getPacker();

    mpjbuf.Buffer wBuffer = datatype.createWriteBuffer(count);

    try {
      packer.pack(wBuffer, buf, 0, count);
      wBuffer.commit();
      mv2j_rsend(wBuffer, dest, tag);
      wBuffer.clear();
      wBuffer.free(); // gadget memory leak fix?
      BufferFactory.destroy(wBuffer.getStaticBuffer());
    } catch (Exception e) {
      throw new MPIException(e);
    }
  }

  /**
   * Start a standard mode, nonblocking send.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in send buffer.
   * @param dest rank of destination
   * @param tag message tag
   * @return communication request
   *
   * <p>Java binding of the MPI operation {@code MPI_ISEND}.
   *
   * <p>Further comments as for {@code Send}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Request iSend(Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    if (buf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return null;
      }
      long commHandle = this.getHandle();
      NativeRequest sendRequest = new NativeRequest(true);
      nativeByteBufferIsend(
        commHandle, (java.nio.ByteBuffer) buf, dest, count, tag, sendRequest);
      return sendRequest;
    }

    Request request = new Request();
    return arraysIsend(buf, count, datatype, dest, tag, request);
  }

  protected Request arraysIsend(Object buf, int count, Datatype datatype, int dest, int tag, Request request)
      throws MPIException {

    Packer packer = datatype.getPacker();
    final mpjbuf.Buffer wBuffer = datatype.createWriteBuffer(count);

    try {
      packer.pack(wBuffer, buf, 0, count);
      wBuffer.commit();
      mv2j_isend(wBuffer, dest, tag, request);
    } catch (Exception e) {
      throw new MPIException(e);
    }

    request.addCompletionHandler(
        new CompletionHandler() {
          public void handleCompletion(Status status) {

            try {
              wBuffer.clear();
              wBuffer.free();
              BufferFactory.destroy(wBuffer.getStaticBuffer());
            } catch (Exception e) {
              throw new MPIException(e);
            }
          }
        });

    return request;
  }

  private Request mv2j_isend(Buffer buf, int dest, int tag, Request req)
      throws MPIException {

    int staticBufferLength = buf.getSize();

    long commHandle = this.getHandle();
    nativeArraysIsend(
        commHandle,
        buf,
        dest,
        tag,
        staticBufferLength + MPI.SEND_OVERHEAD,
        req);

    return req;
  }

  private Request mv2j_ibsend(Buffer buf, int dest, int tag)
      throws MPIException {

    Request req = new Request();

    int staticBufferLength = buf.getSize();

    long commHandle = this.getHandle();
    nativeArraysIbsend(
        commHandle,
        buf,
        dest,
        tag,
        staticBufferLength + MPI.SEND_OVERHEAD,
        req);

    return req;
  }

  private Request mv2j_irsend(Buffer buf, int dest, int tag, Request req)
      throws MPIException {

    int staticBufferLength = buf.getSize();

    long commHandle = this.getHandle();
    nativeArraysIrsend(
        commHandle,
        buf,
        dest,
        tag,
        staticBufferLength + MPI.SEND_OVERHEAD,
        req);

    return req;
  }

  private void mv2j_ssend(Buffer buf, Datatype datatype, int dest, int tag)
      throws MPIException {

    int staticBufferLength = buf.getSize();

    long commHandle = this.getHandle();
    nativeArraysSsend(
        commHandle,
        buf,
        dest,
        tag,
        staticBufferLength + MPI.SEND_OVERHEAD);
  }

  private Request mv2j_issend(Buffer buf, Datatype datatype, int dest, int tag, Request req)
      throws MPIException {

    int staticBufferLength = buf.getSize();

    long commHandle = this.getHandle();
    nativeArraysIssend(
        commHandle,
        buf,
        dest,
        tag,
        staticBufferLength + MPI.SEND_OVERHEAD,
        req);

    return req;
  }

  private void mv2j_send(Buffer buf, Datatype datatype, int dest, int tag) {

    int staticBufferLength = buf.getSize();

    long commHandle = this.getHandle();
    nativeArraysSend(
        commHandle,
        buf,
        dest,
        tag,
        staticBufferLength + MPI.SEND_OVERHEAD);
  }

  private void mv2j_sendRecv(
      Buffer sendbuf,
      int dest,
      int sendtag,
      Buffer recvbuf,
      int source,
      int recvtag,
      Status status) {
      
    int staticSendBufferLength = sendbuf.getSize();
    int staticRecvBufferLength = recvbuf.getCapacity();
    long commHandle = this.getHandle();
    
    nativeArraysSendRecv(
        commHandle,
        sendbuf,
        staticSendBufferLength + MPI.SEND_OVERHEAD,
        dest,
        sendtag,
        recvbuf,
        staticRecvBufferLength,
        source,
        recvtag,
        status);
  }

  private void mv2j_sendRecvReplace(
      Buffer sendbuf,
      int dest,
      int sendtag,
      int source,
      int recvtag,
      Status status) {
      
    int staticSendBufferLength = sendbuf.getSize();
    long commHandle = this.getHandle();
    
    nativeArraysSendRecvReplace(
        commHandle,
        sendbuf,
        staticSendBufferLength + MPI.SEND_OVERHEAD,
        dest,
        sendtag,
        source,
        recvtag,
        status);
  }

  private void mv2j_bsend(Buffer buf, int dest, int tag) {

    int staticBufferLength = buf.getSize();

    long commHandle = this.getHandle();
    nativeArraysBsend(
        commHandle,
        buf,
        dest,
        tag,
        staticBufferLength + MPI.SEND_OVERHEAD);
  }

  private void mv2j_rsend(Buffer buf, int dest, int tag) {

    int staticBufferLength = buf.getSize();

    long commHandle = this.getHandle();
    nativeArraysRsend(
        commHandle,
        buf,
        dest,
        tag,
        staticBufferLength + MPI.SEND_OVERHEAD);
  }

  private Status mv2j_recv(Buffer buf, int src, int tag, Status status)
      throws MPIException {

    int staticBufferLength = buf.getCapacity();
    long commHandle = this.getHandle();
    nativeArraysRecv(commHandle, buf, staticBufferLength, src, tag, status);

    return status;
  } 

  /**
   * Start a buffered mode, nonblocking send.
   * 
   * <p style="color:red"> Not supported in the current release 
   * as it relies on {@code MPI.attachBuffer} and {@code MPI.detachBuffer}</p>.
   *
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in send buffer.
   * @param dest rank of destination.
   * @param tag message tag.
   * @return communication request.
   *
   * <p>Java binding of the MPI operation {@code MPI_IBSEND}.
   *
   * <p>Further comments as for {@code Send}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Request ibSend(Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {
    
    if (buf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return null;
      }
      long commHandle = this.getHandle();
      NativeRequest bsendRequest = new NativeRequest(true);
      nativeByteBufferIbsend(
        commHandle, (java.nio.ByteBuffer) buf, dest, count, tag, bsendRequest);
      return bsendRequest;
    }

    try {
      Request request = new Request();
      return arraysIbsend(buf, count, datatype, dest, tag, request);
    } catch (MPIException mpie) {
      throw mpie;
    }
  }

  Request arraysIbsend(Object buf, int count, Datatype datatype, int dest, int tag, Request request)
      throws MPIException {

    Packer packer = datatype.getPacker();
    final mpjbuf.Buffer wBuffer = datatype.createWriteBuffer(count);

    try {
      packer.pack(wBuffer, buf, 0, count);
      wBuffer.commit();
      request = mv2j_ibsend(wBuffer, dest, tag); 
    } catch (Exception e) {
      throw new MPIException(e);
    }

    request.addCompletionHandler(
        new CompletionHandler() {
          public void handleCompletion(Status status) {

            try {
              wBuffer.clear();
              wBuffer.free();
              BufferFactory.destroy(wBuffer.getStaticBuffer());
            } catch (Exception e) {
              throw new MPIException(e);
            }
          }
        });

    return request;
  }

  /**
   * Start a synchronous mode, nonblocking send.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in send buffer.
   * @param dest rank of destination.
   * @param tag message tag.
   * @return communication request.
   *
   * <p>Java binding of the MPI operation {@code MPI_ISSEND}.
   *
   * <p>Further comments as for {@code Send}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Request isSend(Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    if (buf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return null;
      }
      long commHandle = this.getHandle();
      NativeRequest sendRequest = new NativeRequest(true);
      nativeByteBufferIssend(
        commHandle, (java.nio.ByteBuffer) buf, dest, count, tag, sendRequest);
      return sendRequest;
    }

    try {
      return arraysIssend(buf, count, datatype, dest, tag);
    } catch (MPIException mpie) {
      throw mpie;
    }
  }

  private Request arraysIssend(
      Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    Packer packer = datatype.getPacker();

    final mpjbuf.Buffer wBuffer = datatype.createWriteBuffer(count);

    Request request = new Request();

    try {
      packer.pack(wBuffer, buf, 0, count);
      wBuffer.commit();
      mv2j_issend(wBuffer, datatype, dest, tag, request);
    } catch (Exception e) {
      throw new MPIException(e);
    }

    request.addCompletionHandler(
        new CompletionHandler() {
          public void handleCompletion(Status status) {
            try {
              wBuffer.clear();
              wBuffer.free();
              BufferFactory.destroy(wBuffer.getStaticBuffer());
            } catch (Exception e) {
              throw new MPIException(e);
            }
          }
        });

    return request;
  }

  /**
   * Start a ready mode, nonblocking send.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in send buffer. 
   * @param dest rank of destination.
   * @param tag message tag.
   * @return communication request.
   *
   * <p>Java binding of the MPI operation {@code MPI_IRSEND}.
   *
   * <p>Further comments as for {@code Send}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Request irSend(Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    if (buf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return null;
      }
      long commHandle = this.getHandle();
      NativeRequest rsendRequest = new NativeRequest(true);
      nativeByteBufferIrsend(
        commHandle, (java.nio.ByteBuffer) buf, dest, count, tag, rsendRequest);
      return rsendRequest;
    }

    try {
      Request request = new Request();
      return arraysIrsend(buf, count, datatype, dest, tag, request);
    } catch (MPIException mpie) {
      throw mpie;
    }
  }

  Request arraysIrsend(Object buf, int count, Datatype datatype, int dest, int tag, Request request)
      throws MPIException {

    Packer packer = datatype.getPacker();
    final mpjbuf.Buffer wBuffer = datatype.createWriteBuffer(count);

    try {
      packer.pack(wBuffer, buf, 0, count);
      wBuffer.commit();
      mv2j_irsend(wBuffer, dest, tag, request);
    } catch (Exception e) {
      throw new MPIException(e);
    }

    request.addCompletionHandler(
        new CompletionHandler() {
          public void handleCompletion(Status status) {

            try {
              wBuffer.clear();
              wBuffer.free();
              BufferFactory.destroy(wBuffer.getStaticBuffer());
            } catch (Exception e) {
              throw new MPIException(e);
            }
          }
        });

    return request;
  }

  /**
   * Blocking receive operation.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf receive buffer array.
   * @param count number of items in receive buffer.
   * @param datatype datatypde of each item in receive buffer.
   * @param source rank of source.
   * @param tag message tag.
   * @return status object.
   *
   * <p>Java binding of the MPI operation {@code MPI_RECV}.
   *
   * <p>The actual argument associated with {@code buf} must be one-dimensional array.
   *
   * <p>If the {@code datatype} argument represents an MPI basic type, its value must agree with
   * the element type of {@code buf}---either a primitive type or a reference (object) type. If the
   * {@code datatype} argument represents an MPI derived type, its <em>base type</em> must agree
   * with the element type of {@code buf}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Status recv(Object buf, int count, Datatype datatype, int source, int tag)
      throws MPIException {

    if (buf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return null;
      }
      long commHandle = this.getHandle();
      Status status = new Status(true);
      nativeByteBufferRecv(
        commHandle, (java.nio.ByteBuffer) buf, source, tag, status);
      return status;
    }

    try {
      return arraysRecv(buf, count, datatype, source, tag);
    } catch (MPIException mpie) {
      throw new MPIException(mpie);
    }
  }

  protected Status arraysRecv(
      Object buf, int count, Datatype datatype, int source, int tag)
      throws MPIException {

    mpjbuf.Buffer rBuffer = datatype.createReadBuffer(count);
    Packer packer = datatype.getPacker();
    Status status = new Status();

    try {
      mv2j_recv(rBuffer, source, tag, status);
      rBuffer.commit();

      status.type = rBuffer.getSectionHeader();
      status.numEls = rBuffer.getSectionSize();

      packer.unpack(rBuffer, status.numEls, buf, 0, count);

      rBuffer.clear();
      rBuffer.free();

      BufferFactory.destroy(rBuffer.getStaticBuffer());
    } catch (Exception e) {
      e.printStackTrace();
      throw new MPIException(e);
    }

    if (datatype.size != 0) {
      int temp = status.numEls / datatype.size;
      if (temp * datatype.size == status.numEls) {
        status.count = temp;
      } else {
        status.count = MPI.UNDEFINED;
      }
    } else {
      status.count = count;
      status.numEls = MPI.UNDEFINED; // according to MPI specs, this should be
      // zero, but we are trusting our test-cases
      // which are passed by LAM, MPICH, and mpiJava
    }

    return status;
  }

  /**
   * Start a nonblocking receive.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf receive buffer array.
   * @param count number of items in receive buffer.
   * @param datatype datatype of each item in receive buffer.
   * @param src rank of source.
   * @param tag message tag.
   * @return communication request.
   *
   * <p>Java binding of the MPI operation {@code MPI_IRECV}.
   *
   * <p>Further comments as for {@code Recv}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Request iRecv(Object buf, int count, Datatype datatype, int src, int tag)
      throws MPIException {

    if (buf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return null;
      }
      long commHandle = this.getHandle();
      NativeRequest recvRequest = new NativeRequest(true);
      nativeByteBufferIrecv(
        commHandle, (java.nio.ByteBuffer) buf, src, tag, recvRequest);
      return recvRequest;
    }

    try {
      Request request = new Request();
      return arraysIrecv(buf, count, datatype, src, tag, request);
    } catch (MPIException mpie) {
      throw mpie;
    }
  }

  protected Request arraysIrecv(
      Object buf,
      int count,
      Datatype datatype,
      int src,
      int tag,
      Request request)
      throws MPIException {

    final mpjbuf.Buffer rBuffer = datatype.createReadBuffer(count);
    final Packer packer = datatype.getPacker();
    final int c = count;
    final Object buf1 = buf;
    final int source = src;
    final int t = tag;
    final Datatype dtype = datatype;
    Status status = null;

    try {
      request = mv2j_irecv(rBuffer, src, tag, status);
    } catch (Exception e) {
      throw new MPIException(e);
    }

    request.addCompletionHandler(
        new CompletionHandler() {
          public void handleCompletion(Status status) {

            int count_received = 0;

            try {

              rBuffer.setSize(status.numEls);
              rBuffer.commit();
              Type recType = rBuffer.getSectionHeader();
              count_received = rBuffer.getSectionSize();

              packer.unpack(rBuffer, count_received, buf1, 0, c);
              rBuffer.clear();
              rBuffer.free();

              BufferFactory.destroy(rBuffer.getStaticBuffer());
            } catch (Exception e) {
              e.printStackTrace();
              throw new MPIException(e);
            }

            status.numEls = count_received;
          }
        });

    request.datatype = datatype;
    return request;
  }

  private Request mv2j_irecv(
      Buffer buffer, int src, int tag, Status status)
      throws MPIException {

    long commHandle = this.getHandle();
    Request req = new Request();
    nativeArraysIrecv(commHandle, buffer, src, tag, status, req);
    return req;
  }

  /**
   * Check if there is an incoming message matching the pattern specified.
   *
   * @param source rank of source.
   * @param tag message tag.
   * @return status object or null handle.
   *
   * <p>Java binding of the MPI operation {@code MPI_IPROBE}.
   *
   * <p>If such a message is currently available, a status object similar to the return value of a
   * matching {@code Recv} operation is returned. Otherwise a null handle is returned.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Status iProbe(int source, int tag) throws MPIException {

    Status status = new Status(-1, -1, -1);

    nativeIprobe(handle, source, tag, status);

    if (status.source == -1) return null;
    else return status;
  }

  /**
   * Wait until there is an incoming message matching the pattern specified. 
   *
   * @param source rank of source.
   * @param tag message tag.
   * @return status object.
   *
   * <p>Java binding of the MPI operation {@code MPI_PROBE}.
   *
   * <p>Returns a status object similar to the return value of a matching {@code Recv} operation.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Status probe(int source, int tag) throws MPIException {

    Status status = new Status(source, tag, -1);
    nativeProbe(handle, source, tag, status);

    return status;
  }

  /**
   * Returns an upper bound on the increment of {@code position} effected by {@code pack}. 
   *
   * @param incount number of items in input buffer.
   * @param datatype datatype of each item in input buffer.
   * @return upper bound on size of packed message.
   *
   * <p>Java binding of the MPI operation {@code MPI_PACK_SIZE}.
   *
   * <p><em>It is an error to call this function if the base type of {@code datatype} is
   * {@code MPI.OBJECT}</em>.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int packSize(int incount, Datatype datatype) throws MPIException {
    return datatype.packedSize(incount);
  }

  /**
   * Packs message in send buffer {@code inbuf} into space specified in {@code outbuf}.
   *
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * @param inbuf input buffer array.
   * @param incount number of items in input buffer.
   * @param datatype datatype of each item in input buffer.
   * @param outbuf output buffer.
   * @param position initial position in output buffer.
   * @return final position in output buffer.
   *
   * <p>Java binding of the MPI operation {@code MPI_PACK}.
   *
   * <p>The return value is the output value of {@code position} - the inital value incremented by
   * the number of bytes written.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int pack(
      Object inbuf, int incount, Datatype datatype, mpjbuf.Buffer outbuf, int position)
      throws MPIException {

    outbuf.setSize(position);
    Packer packer = datatype.getPacker();
    int messageSize = datatype.packedSize(incount);

    try {
      outbuf.putSectionHeader(datatype.bufferType);
      packer.pack(outbuf, inbuf, 0, incount);
    } catch (Exception e) {
      throw new MPIException(e);
    }

    return outbuf.getSize();
  }

  /**
   * Unpacks message in receive buffer <tt>outbuf</tt> into space specified in <tt>inbuf</tt>.
   *
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * @param inbuf input buffer.
   * @param position initial position in input buffer.
   * @param outbuf output buffer.
   * @param outcount number of items in output buffer.
   * @param datatype datatype of each item in output buffer.
   * @return final position in input buffer.
   *
   * <p>Java binding of the MPI operation {@code MPI_UNPACK}.
   *
   * <p>The return value is the output value of {@code position} - the inital value incremented by
   * the number of bytes read.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int unpack(
      mpjbuf.Buffer inbuf, int position, Object outbuf, int outcount, Datatype datatype)
      throws MPIException {

    int expectedSize = MPI.COMM_WORLD.packSize(outcount, datatype);
    inbuf.setSize(expectedSize + position);

    try {
      Packer packer = datatype.getPacker();
      Type recType = inbuf.getSectionHeader();
      int count_received = inbuf.getSectionSize();
      packer.unpack(inbuf, count_received, outbuf, 0, outcount);
    } catch (Exception e) {
      throw new MPIException(e);
    }

    return inbuf.getSize();
  }

  /**
   * Creates a persistent communication request for a buffered mode send.
   *
   * <p style="color:red"> Not supported in the current release 
   * as it relies on {@code MPI.attachBuffer} and {@code MPI.detachBuffer}</p>.
   * 
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in send buffer.
   * @param dest rank of destination.
   * @param tag message tag.
   * @return persistent communication request.
   *
   * <p>Java binding of the MPI operation {@code MPI_BSEND_INIT}.
   *
   * <p>Further comments as for {@code Send}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Prequest bSendInit(
      Object buf, int count, Datatype datatype, int dest, int tag) throws MPIException {

    Prequest preq = new Prequest(buf, count, datatype, dest, tag, MPI.OP_BSEND, this);
    return preq;
  }

  /**
   * Creates a persistent communication request for a standard mode send.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in send buffer.
   * @param dest rank of destination.
   * @param tag message tag.
   * @return persistent communication request.
   *
   * <p>Java binding of the MPI operation {@code MPI_SEND_INIT}.
   *
   * <p>Further comments as for {@code Send}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Prequest sendInit(Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    Prequest preq = new Prequest(buf, count, datatype, dest, tag, MPI.OP_SEND, this);
    return preq;
  }

  /**
   * Creates a persistent communication request for a synchronous mode send.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in send buffer.
   * @param dest rank of destination.
   * @param tag message tag.
   * @return persistent communication request.
   *
   * <p>Java binding of the MPI operation {@code MPI_SSEND_INIT}.
   *
   * <p>Further comments as for {@code Send}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Prequest sSendInit(
      Object buf, int count, Datatype datatype, int dest, int tag) throws MPIException {
    Prequest preq = new Prequest(buf, count, datatype, dest, tag, MPI.OP_SSEND, this);
    return preq;
  }

  /**
   * Creates a persistent communication request for a ready mode send.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf send buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in send buffer.
   * @param dest rank of destination.
   * @param tag message tag.
   * @return persistent communication request.
   *
   * <p>Java binding of the MPI operation {@code MPI_RSEND_INIT}.
   *
   * <p>Further comments as for {@code Send}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Prequest rSendInit(
      Object buf, int count, Datatype datatype, int dest, int tag) throws MPIException {

    Prequest preq = new Prequest(buf, count, datatype, dest, tag, MPI.OP_RSEND, this);
    return preq;
  }

  /**
   * Creates a persistent communication request for a receive operation.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf receive buffer array.
   * @param count number of items in receive buffer.
   * @param datatype datatype of each item in receive buffer.
   * @param dest rank of destination.
   * @param tag message tag.
   * @return communication request.
   *
   * <p>Java binding of the MPI operation {@code MPI_RECV_INIT}.
   *
   * <p>Further comments as for {@code Recv}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Prequest recvInit(Object buf, int count, Datatype datatype, int dest, int tag)
      throws MPIException {

    Prequest preq = new Prequest(buf, count, datatype, dest, tag, MPI.OP_RECV, this);
    return preq;
  }

  /**
   * Execute a blocking send and receive operation.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param sendcount number of items to send.
   * @param sendtype datatype of each item in send buffer.
   * @param dest rank of destination.
   * @param sendtag send tag.
   * @param recvbuf receive buffer array.
   * @param recvcount number of items in receive buffer. 
   * @param recvtype datatype of each item in receive buffer.
   * @param source rank of source.
   * @param recvtag receive tag.
   * @return status object.
   *
   * <p>Java binding of the MPI operation {@code MPI_SENDRECV}.
   *
   * <p>Further comments as for {@code Send} and {@code Recv}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Status sendRecv(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      int dest,
      int sendtag,
      Object recvbuf,
      int recvcount,
      Datatype recvtype,
      int source,
      int recvtag)
      throws MPIException {

    if (sendbuf instanceof java.nio.Buffer) {
      if (sendtype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return null;
      }
      long commHandle = this.getHandle();
      Status status = new Status();
      nativeByteBufferSendRecv(
                 commHandle,
                 (java.nio.ByteBuffer) sendbuf,
                 sendcount,
                 dest,
                 sendtag,
                 (java.nio.ByteBuffer) recvbuf,
                 recvcount,
                 source,
                 recvtag,
                 status);
      return status;
    }

    try {
      return arraysSendRecv(
                 sendbuf,
                 sendcount,
                 sendtype,
                 dest,
                 sendtag,
                 recvbuf,
                 recvcount,
                 recvtype,
                 source,
                 recvtag);
    } catch (MPIException mpie) {
      throw mpie;
    }
  }

  private Status arraysSendRecv(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      int dest,
      int sendtag,
      Object recvbuf,
      int recvcount,
      Datatype recvtype,
      int source,
      int recvtag) {

    Packer packer = sendtype.getPacker();
    mpjbuf.Buffer wBuffer = sendtype.createWriteBuffer(sendcount);
    mpjbuf.Buffer rBuffer = recvtype.createReadBuffer(recvcount);
    Status status = new Status();

    try {
      packer.pack(wBuffer, sendbuf, 0, sendcount);
      wBuffer.commit();
      mv2j_sendRecv(
          wBuffer,
          dest,
          sendtag,
          rBuffer,
          source,
          recvtag,
          status);

      rBuffer.commit();
      status.type = rBuffer.getSectionHeader();
      status.numEls = rBuffer.getSectionSize();
      
      packer.unpack(rBuffer, status.numEls, recvbuf, 0, recvcount);

      rBuffer.clear();
      rBuffer.free();

      wBuffer.clear();
      wBuffer.free(); // gadget memory leak fix?
      
      BufferFactory.destroy(wBuffer.getStaticBuffer());
      BufferFactory.destroy(rBuffer.getStaticBuffer());
    } catch (Exception e) {
      throw new MPIException(e);
    }

    return status;
  }

  /**
   * Execute a blocking send and receive operation, receiving message into send buffer.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf buffer array.
   * @param count number of items to send.
   * @param datatype datatype of each item in buffer.
   * @param dest rank of destination.
   * @param sendtag send tag.
   * @param source rank of source.
   * @param recvtag receive tag.
   * @return status object.
   *
   * <p>Java binding of the MPI operation {@code MPI_SENDRECV_REPLACE}.
   * <p>Further comments as for {@code Send} and {@code Recv}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Status sendRecvReplace(
      Object buf, int count, Datatype datatype, int dest, int sendtag, int source, int recvtag)
      throws MPIException {

    if (buf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return null;
      }
      long commHandle = this.getHandle();
      Status status = new Status();
      nativeByteBufferSendRecvReplace(
                 commHandle,
                 (java.nio.ByteBuffer) buf,
                 count,
                 dest,
                 sendtag,
                 source,
                 recvtag,
                 status);
      return status;
    }

    try {
      return arraysSendRecvReplace(
                 buf,
                 count,
                 datatype,
                 dest,
                 sendtag,
                 source,
                 recvtag);
    } catch (MPIException mpie) {
      throw mpie;
    }
  } 

  private Status arraysSendRecvReplace(
      Object buf,
      int count,
      Datatype datatype,
      int dest,
      int sendtag,
      int source,
      int recvtag) {

    Packer packer = datatype.getPacker();
    mpjbuf.Buffer wBuffer = datatype.createWriteBuffer(count);
    Status status = new Status();

    try {
      packer.pack(wBuffer, buf, 0, count);
      wBuffer.commit();
      mv2j_sendRecvReplace(
          wBuffer,
          dest,
          sendtag,
          source,
          recvtag,
          status);

      status.type = wBuffer.getSectionHeader();
      status.numEls = wBuffer.getSectionSize();

      packer.unpack(wBuffer, status.numEls, buf, 0, count);

      wBuffer.clear();
      wBuffer.free(); // gadget memory leak fix?
      
      BufferFactory.destroy(wBuffer.getStaticBuffer());
    } catch (Exception e) {
      throw new MPIException(e);
    }

    return status;
  }

  /**
   * Returns the type of topology associated with the communicator.
   *
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * @return topology type of communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_TOPO_TEST}.
   *
   * <p>The return value will be one of {@code MPI.GRAPH}, {@code MPI.CART} or
   * {@code MPI.UNDEFINED}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int getTopology() throws MPIException {
    // -1 represents no topology ...
    return -1;
  }

  /**
   * Abort MPI.
   *
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * @param errorcode error code for Unix or POSIX environments.
   *
   * <p>Java binding of the MPI operation {@code MPI_ABORT}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void Abort(int errorcode) throws MPIException {}

  /**
   * Duplicate this communicator.
   *
   * @return copy of this communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_DUP}.
   *
   * <p>The new communicator is ``congruent'' to the old one, but has a different context.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Comm clone() throws MPIException {
    return new Comm(dup(getHandle()));
  } 

  /**
   * Spawn an MPI process.
   *
   * @param command name of program to be spawned.
   * @param argv arguments for the command.
   * @param maxprocs maximum number of processes to launch.
   * @param info a set of key-value pairs telling the runtime system where and how to
   *             start the processes.
   * @param root rank of root process.
   * @param errCodes array where error codes are populated per process after
   *                 launch. 
   * @return an intercommunicator and if any errors occur a populated {@code errCodes}.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_SPAWN}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Intercomm spawn(
      String command, String[] argv, int maxprocs, Info info, int root, int[] errCodes) 
      throws MPIException {

    return new Intercomm(nativeSpawn(handle, command, argv, maxprocs, info.handle, root, errCodes));
  }

  /**
   * Spawn an multiple MPI processes with different executable specifications.
   *
   * @param cmds programs to be executed.
   * @param argsArray arguments for commands.
   * @param maxprocs maximum number of processes to start for each command.
   * @param info info objects telling the runtime system where and how to start
   *             processes.
   * @param root rank of root process.
   * @param errcodes array for error codes.
   * @return an intercommunicator and if any errors occur a populated {@code errCodes}.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_SPAWN_MULTIPLE}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Intercomm spawnMultiple(
      String[] cmds, String[][] argsArray, int[] maxprocs, Info[] info, int root, int[] errcodes)
      throws MPIException {

    long infoHandles[] = new long[info.length];
    for (int i = 0; i < info.length; i++) {
      infoHandles[i] = info[i].handle;
    }

    return new Intercomm(
        nativeSpawnMultiple(handle, cmds, argsArray, maxprocs, infoHandles, root, errcodes));
  }

  /**
   * Compare two communicators.
   *
   * @param comm1 first communicator.
   * @param comm2 second communicator.
   * @return result.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_COMPARE}.
   *
   * <p>{@code MPI.IDENT}(0) results if the {@code comm1} and {@code comm2} are references to the
   * same object (ie, if {@code comm1 == comm2}). {@code MPI.CONGRUENT}(3) results if the
   * underlying groups are identical but the communicators differ by context.
   * {@code MPI.SIMILAR}(1) results if the underlying groups are similar but the communicators
   * differ by context. {@code MPI.UNEQUAL}(2) results otherwise.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static native int compare(Comm comm1, Comm comm2) throws MPIException;

  static native int getTagUB();
    
  private native boolean Test_inter() throws MPIException;

  native long dup(long commHandle) throws MPIException;

  private native int rank(long commHandle) throws MPIException;

  private native int size(long commHandle) throws MPIException;

  private native long create(long commHandle, int[] ids) throws MPIException;

  private native long nativeCreateIntercomm(
      long peerCommHandle, long localCommHandle, int localleader, int remoteleader, int tag)
      throws MPIException;

  private native void free(long commHandle) throws MPIException;

  private native long group() throws MPIException;

  private native void GetComm(int Type) throws MPIException;

  private native void nativeProbe(long handle, int src, int tag, Status status) throws MPIException;

  private native void nativeIprobe(long handle, int src, int tag, Status status)
      throws MPIException;

  /* native standard iSend()/send() methods for arrays */
  private native void nativeArraysIsend(
      long handle,
      Buffer buffer,
      int dest,
      int tag,
      int staticBufferLength,
      Request req)
      throws MPIException;

  private native void nativeArraysSend(
      long handle,
      Buffer buf,
      int dest,
      int tag,
      int slen)
      throws MPIException;

  /* native standard ibSend()/Bsend() methods for arrays */
  private native void nativeArraysIbsend(
      long handle,
      Buffer buffer,
      int dest,
      int tag,
      int staticBufferLength,
      Request req)
      throws MPIException;

  private native void nativeArraysBsend(
      long handle, Buffer buf, int dest, int tag, int slen)
      throws MPIException;

  /* native standard irSend()/rSend() methods for arrays */
  private native void nativeArraysIrsend(
      long handle,
      Buffer buffer,
      int dest,
      int tag,
      int staticBufferLength,
      Request req)
      throws MPIException;

  private native void nativeArraysRsend(
      long handle,
      Buffer buf,
      int dest,
      int tag,
      int slen)
      throws MPIException;

  /* native synchronous isSend()/sSend() methods for arrays */
  private native void nativeArraysIssend(
      long handle,
      Buffer buffer,
      int dest,
      int tag,
      int staticBufferLength,
      Request req)
      throws MPIException;

  private native void nativeArraysSsend(
      long handle,
      Buffer buf,
      int dest,
      int tag,
      int slen)
      throws MPIException;

  /* native sendRecv()/sendRecvReplace() methods for arrays */
  private native Status nativeArraysSendRecv(
      long handle,
      Buffer sendbuf,
      int staticSendBufferLength,
      int dest,
      int sendtag,
      Buffer recvBuf,
      int staticRecvBufferLength,
      int source,
      int recvtag,
      Status status)
      throws MPIException; 

  private native Status nativeArraysSendRecvReplace(
      long handle,
      Buffer sendbuf,
      int staticSendBufferLength,
      int dest,
      int sendtag,
      int source,
      int recvtag,
      Status status)
      throws MPIException;

  /* native iRecv()/recv() methods for arrays */
  private native void nativeArraysIrecv(
      long handle,
      Buffer buffer,
      int src,
      int tag,
      Status status,
      Request req)
      throws MPIException;

  private native void nativeArraysRecv(
      long handle,
      Buffer buffer,
      int size,
      int src,
      int tag,
      Status status)
      throws MPIException;

  /* native standard iSend/send() method for ByteBuffers */
  private native void nativeByteBufferIsend(
      long handle,
      ByteBuffer buffer,
      int dest,
      int count,
      int tag,
      Request request)
      throws MPIException;

  private native void nativeByteBufferSend(
      long handle,
      ByteBuffer buffer,
      int dest,
      int count,
      int tag)
      throws MPIException;

  /* native standard ibSend/bSend() method for ByteBuffers */
  private native void nativeByteBufferIbsend(
      long handle, ByteBuffer buffer, int dest, int count, int tag, Request request)
      throws MPIException;

  private native void nativeByteBufferBsend(
      long handle, ByteBuffer buffer, int dest, int count, int tag) throws MPIException;

  /* native standard irSend/rSend() method for ByteBuffers */
  private native void nativeByteBufferIrsend(
      long handle,
      ByteBuffer buffer,
      int dest,
      int count,
      int tag,
      Request request)
      throws MPIException;

  private native void nativeByteBufferRsend(
      long handle,
      ByteBuffer buffer,
      int dest,
      int count,
      int tag)
      throws MPIException;

  /* native synchronous isSend/sSend() method for ByteBuffers */
  private native void nativeByteBufferIssend(
      long handle,
      ByteBuffer buffer,
      int dest,
      int count,
      int tag,
      Request request)
      throws MPIException;

  private native void nativeByteBufferSsend(
      long handle,
      ByteBuffer buffer,
      int dest,
      int count,
      int tag)
      throws MPIException;

  /* native sendRecv()/sendRecvReplace() method for ByteBuffers */ 
  private native void nativeByteBufferSendRecv(
      long handle,
      ByteBuffer sendBuffer,
      int sendcount,
      int dest,
      int sendtag,
      ByteBuffer recvBuffer,
      int recvcount,
      int source,
      int recvtag,
      Status status)
      throws MPIException;

  private native void nativeByteBufferSendRecvReplace(
      long handle,
      ByteBuffer buffer,
      int count,
      int dest,
      int sendtag,
      int source,
      int recvtag,
      Status status)
      throws MPIException;

  /* native iRecv/recv() method for ByteBuffers */
  private native void nativeByteBufferIrecv(
      long handle,
      ByteBuffer buffer,
      int src, 
      int tag,
      Request request)
      throws MPIException;

  private native void nativeByteBufferRecv(
      long handle,
      ByteBuffer buffer,
      int src,
      int tag,
      Status status)
      throws MPIException; 

  private native long nativeSpawn(
      long handle,
      String command,
      String[] argv,
      int maxprocs,
      long info,
      int root,
      int[] errCodes)
      throws MPIException;

  private native long nativeSpawnMultiple(
      long comm,
      String[] cmds,
      String[][] argsArray,
      int[] maxrprocs,
      long[] infoArr,
      int root,
      int[] errcodes)
      throws MPIException;
}
