/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
 The MIT License

 Copyright (c) 2005 - 2008
   1. Distributed Systems Group, University of Portsmouth (2005)
   2. Community Grids Laboratory, Indiana University (2005)
   3. Aamir Shafi (2005 - 2008)
   4. Bryan Carpenter (2005 - 2008)
   5. Mark Baker (2005 - 2008)

 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the
 "Software"), to deal in the Software without restriction, including
 without limitation the rights to use, copy, modify, merge, publish,
 distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to
 the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * File         : Datatype.java
 * Author       : Sang Lim, Sung-Hoon Ko, Xinying Li, Bryan Carpenter
 *                Aamir Shafi
 * Created      : Thu Apr  9 12:22:15 BST 1998
 * Revision     : $Revision: 1.15 $
 * Updated      : $Date: 2014/03/11 13:12:07 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

package mpi;

import java.nio.ByteBuffer;

/**
 * The {@code Datatype} class represents {@code MPI_DATATYPE} handles.
 */
public abstract class Datatype implements Freeable {

  // The following magic numbers for different datatypes are coming
  // from mpi.h in the MVAPICH2 software. These are hard-coded
  // currently but will be good to read these at runtime during
  // initialization from the native library. Things will break
  // because of the hard-coding if these Java bindings are tried with
  // other MPI libraries.
  static final int UNDEFINED = -1;
  static final int NULL = 0;
  static final int BYTE = 0x4c00010d;
  static final int CHAR = 0x4c000101;
  static final int SHORT = 0x4c000203;
  static final int BOOLEAN = 4;
  static final int INT = 0x4c000405;
  static final int LONG = 0x4c000807;
  static final int FLOAT = 0x4c00040a;
  static final int DOUBLE = 0x4c00080b;
  static final int PACKED = 0x4c00010f;
  static final int LB = 0x4c000010;
  static final int UB = 0x4c000011;
  static final int OBJECT = 12;
  static final int SHORT2 = 3;
  static final int INT2 = 5;
  static final int LONG2 = 6;
  static final int FLOAT2 = 7;
  static final int DOUBLE2 = 8;

  public int baseType = UNDEFINED;

  mpjbuf.Type bufferType = mpjbuf.Type.UNDEFINED;

  int numEls;
  int size;
  int byteSize;
  int lb, ub, extent;
  boolean ubSet, lbSet;
  // Flags set if MPI.UB, MPI.LB respectively appears as a component type.
  Packer packer;

  /**
   * Construct new datatype representing replication of old datatype into contiguous locations.
   *
   * @param count replication count.
   * @param oldtype datatype.
   * @return new datatype.
   *
   * <p>Java binding of the MPI operation {@code MPI_TYPE_CONTIGUOUS}.
   *
   * <p>The base type of the new datatype is the same as the base type of {@code oldtype}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Datatype createContiguous(int count, Datatype oldtype) throws MPIException {
    return new Contiguous(count, oldtype);
  }

  /**
   * Construct new datatype representing replication of old datatype into locations that consist of
   * equally spaced blocks.
   *
   * @param count number of blocks.
   * @param blocklength number of elements in each block.
   * @param stride number of elements between start of each block.
   * @param oldtype old datatype.
   * @return new datatype.
   *
   * <p>Java binding of the MPI operation {@code MPI_TYPE_VECTOR}.
   *
   * <p>The base type of the new datatype is the same as the base type of {@code oldtype}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Datatype createVector(int count, int blocklength, int stride, Datatype oldtype)
      throws MPIException {

    return new Vector(count, blocklength, stride, oldtype, true);
  }

  /**
   * Identical to <tt>vector</tt> except that the stride is expressed directly in terms of the
   * buffer index, rather than the units of the old type.
   *
   * @param count number of blocks.
   * @param blocklength number of elements in each block.
   * @param stride number of elements between start of each block.
   * @param oldtype old datatype.
   * @return new datatype.
   *
   *
   * <p>Java binding of the MPI operation {@code MPI_TYPE_CREATE_HVECTOR}.
   *
   * <p><em>Unlike other language bindings</em>, the value of {@code stride} is <em>not</em>
   * measured in bytes.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Datatype createHVector(int count, int blocklength, int stride, Datatype oldtype)
      throws MPIException {

    return new Vector(count, blocklength, stride, oldtype, false);
  }

  /**
   * Construct new datatype representing replication of old datatype into a sequence of blocks where
   * each block can contain a different number of copies and have a different displacement.
   *
   * @param array_of_blocklengths number of elements per block.
   * @param array_of_displacements displacement of each block in units of old
   *                               type.
   * @param oldtype old datatype.
   * @return new datatype.
   *
   * <p>Java binding of the MPI operation {@code MPI_TYPE_INDEXED}.
   *
   * <p>The number of blocks is taken to be size of the {@code array_of_blocklengths} argument. The
   * second argument, {@code array_of_displacements}, should be the same size. The base type of the
   * new datatype is the same as the base type of {@code oldtype}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Datatype createIndexed(
      int[] array_of_blocklengths, int[] array_of_displacements, Datatype oldtype)
      throws MPIException {

    return new Indexed(array_of_blocklengths, array_of_displacements, oldtype, true);
  }

  /**
   * Identical to {@code indexed} except that the displacements are expressed directly in terms of
   * the buffer index, rather than the units of the old type.
   *
   * @param array_of_blocklengths number of elements per block.
   * @param array_of_displacements displacement in buffer for each block.
   * @param oldtype old datatype.
   * @return new datatype.
   *
   * <p>Java binding of the MPI operation {@code MPI_TYPE_HINDEXED}.
   *
   * <p><em>Unlike other language bindings</em>, the values in {@code array_of_displacements} are
   * <em>not</em> measured in bytes.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Datatype hIndexed(
      int[] array_of_blocklengths, int[] array_of_displacements, Datatype oldtype)
      throws MPIException {
    return new Indexed(array_of_blocklengths, array_of_displacements, oldtype, false);
  }

  /**
   * The most general type constructor.
   *
   * @param array_of_blocklengths number of elements per block.
   * @param array_of_displacements displacement in buffer for each block.
   * @param array_of_types type of elements in each block.
   * @return new datatype.
   *
   * <p>Java binding of the MPI operation {@code MPI_TYPE_STRUCT}.
   *
   * <p>The number of blocks is taken to be size of the {@code array_of_blocklengths} argument. The
   * second and third arguments, {@code array_of_displacements}, and {@code array_of_types},
   * should be the same size. <em>Unlike other language bindings</em>, the values in
   * {@code array_of_displacements} are <em>not</em> measured in bytes. All elements of
   * {@code array_of_types} with definite base types <em>must have the same base
   * type</em>: this will be the base type of new datatype.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Datatype struct(
      int[] array_of_blocklengths, int[] array_of_displacements, Datatype[] array_of_types)
      throws MPIException {
    return new Struct(array_of_blocklengths, array_of_displacements, array_of_types);
  }

  Packer getPacker() {

    if (packer == null) setPacker();

    return packer;
  }

  // added for native MPI support
  public int getType() {
    return this.baseType;
  }

  // added for native MPI support
  public int getByteSize() {
    return this.byteSize;
  }

  abstract void setPacker();

  /*
   * offset will be 'MPI.SEND_OVERHEAD' size will be 'packedSize(count) +
   * MPI.SEND_OVERHEAD '
   */
  abstract mpjbuf.Buffer createWriteBuffer(int count) throws MPIException;

  /*
   * messageSize will be 'packedSize(count)+MPI.BSEND_OVERHEAD' offset will be
   * set in this method to 'MPI.BSEND_OVERHEAD '
   */
  abstract mpjbuf.Buffer createWriteBuffer(ByteBuffer slicedBuffer, int messageSize);

  /*
   * size is 'packedSize(count)+MPI.RECV_OVERHEAD' ; offset is
   * 'MPI.RECV_OVERHEAD'
   */
  abstract mpjbuf.Buffer createReadBuffer(int count);

  abstract int packedSize(int count);

  /**
   * Returns the extent of a datatype - the difference between upper and lower bound.
   *
   * @return datatype extent.
   *
   * <p>Java binding of the MPI operation {@code MPI_TYPE_EXTENT}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int extent() throws MPIException {
    return extent;
  }

  /**
   * Returns the total size of a datatype - the number of buffer elements it represents.
   *
   * @return datatype size.
   *
   * <p>Java binding of the MPI operation {@code MPI_TYPE_SIZE}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int size() throws MPIException {
    return size;
  }

  /**
   * Find the lower bound of a datatype - the least value in its displacement sequence.
   *
   * @return displacement of lower bound from origin.
   *
   * <p>Java binding of the MPI operation {@code MPI_TYPE_LB}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int lb() throws MPIException {
    return lb;
  }

  /**
   * Find the upper bound of a datatype - the greatest value in its displacement sequence.
   *
   * @return displacement of upper bound from origin.
   *
   * <p>Java binding of the MPI operation {@code MPI_TYPE_UB}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int ub() throws MPIException {
    return ub;
  }

  /** Commit a derived datatype. 
   * 
   * <p>Java binding of the MPI operation {@code MPI_TYPE_COMMIT}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void commit() throws MPIException {}

  /**
   * Frees the datatype.
   *
   * <p>Java binding of the MPI operation {@code MPI_TYPE_FREE}. 
   *
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void free() throws MPIException {}
}
