/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
The MIT License

Copyright (c) 2005 - 2008
  1. Distributed Systems Group, University of Portsmouth (2005)
  2. Community Grids Laboratory, Indiana University (2005)
  3. Aamir Shafi (2005 - 2008)
  4. Bryan Carpenter (2005 - 2008)
  5. Mark Baker (2005 - 2008)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/*
 * File         : Graphcomm.java
 * Author       : Aamir Shafi, Bryan Carpenter
 * Created      : Fri Sep 10 12:22:15 BST 2004
 * Revision     : $Revision: 1.8 $
 * Updated      : $Date: 2005/09/28 13:33:59 $
 */

package mpi;

/**
 * The class represents a communicator with a graph structure.
 */
public class Graphcomm extends Intracomm {
  GraphParms graphParms = null;

  Graphcomm(int[] index, int[] edges, boolean reorder, mpi.Comm mpjdevComm, Group group)
      throws MPIException {

    super(mpjdevComm, group);

    int rank = group.getRank();
    graphParms = new GraphParms();

    graphParms.index = new int[index.length];
    System.arraycopy(index, 0, graphParms.index, 0, index.length);

    graphParms.edges = new int[edges.length];
    System.arraycopy(edges, 0, graphParms.edges, 0, edges.length);
  }

  /** */
  public Graphcomm clone() throws MPIException {
    return null;
  }

  /**
   * Returns graph topology information.
   *
   * @return object defining node degrees and edges of graph.
   *
   * <p>Java binding of the MPI operation {@code MPI_GRAPHDIMS_GET}.
   *
   * <p>The number of nodes and number of edges can be extracted from the sizes of the
   * {@code index} and {@code edges} fields of the returned object.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public GraphParms getDims() throws MPIException {
    return graphParms;
  }

  /**
   * Provides adjacency information for general graph topology.
   *
   * @param rank rank of a process in the group of this communicator.
   * @return array of ranks of neighbouring processes to one specified.
   *
   * <p>Java binding of the MPI operations {@code MPI_GRAPH_NEIGHBOURS_COUNT} and
   * {@code MPI_GRAPH_NEIGHBOURS}.
   *
   * <p>The number of neighbours can be extracted from the size of the result.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int[] getNeighbours(int rank) throws MPIException {
    int nnbrs = graphParms.index[rank];
    int i = 0;
    int p = 0; // index to graphParms edges

    if (rank > 0) {
      i = graphParms.index[rank - 1];
      nnbrs -= i;
      p = i;
    }

    int[] nbrs = new int[nnbrs];

    for (i = 0; i < nnbrs; ++i, ++p) {
      nbrs[i] = graphParms.edges[p];
    }

    return nbrs;
  }

  /**
   * Compute an optimal placement.
   *
   * @param index node degrees.
   * @param edges graph edges.
   * @return reordered rank of calling process.
   *
   * <p>Java binding of the MPI operation {@code MPI_GRAPH_MAP}.
   *
   * <p>The number of nodes is taken to be size of the {@code index} argument.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int map(int[] index, int[] edges) throws MPIException {
    int myrank = this.group.getRank();
    // -1 in the next line is MPI.UNDEFINED ...
    return ((myrank < 0) || (myrank >= index.length)) ? -1 : myrank;
  }
  
  /**
   * Returns the type of topology associated with the communicator.
   *
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * @return topology type of communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_TOPO_TEST}.
   *
   * <p>The return value will be one of {@code MPI.GRAPH}, {@code MPI.CART} or
   * {@code MPI.UNDEFINED}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int getTopology() throws MPIException {
    return MPI.GRAPH; // this means no topology ....
  }
}
