/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
The MIT License

Copyright (c) 2005 - 2008
  1. Distributed Systems Group, University of Portsmouth (2005)
  2. Community Grids Laboratory, Indiana University (2005)
  3. Aamir Shafi (2005 - 2008)
  4. Bryan Carpenter (2005 - 2008)
  5. Mark Baker (2005 - 2008)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/*
 * File         : Group.java
 * Author       : Aamir Shafi, Bryan Carpenter
 * Created      : Fri Sep 10 12:22:15 BST 2004
 * Revision     : $Revision: 1.14 $
 * Updated      : $Date: 2014/03/11 13:24:14 $
 */

package mpi;

/**
 * This class represents {@code MPI_Group}.
 */
public class Group implements Freeable {

  static final int EMPTY = 1;
  static final int NULL = 2;

  static {
    init();
  }

  private long handleOfGroup;

  Group() {}

  Group(long _handle) {
    handleOfGroup = _handle;
  }

  public Group(int type) {
    GetGroup(type);
  }

  /**
   * Translate ranks within one group to ranks within another.
   *
   * @param group1 first group.
   * @param ranks1 array of valid ranks in {@code group1}.
   * @param group2 second group.
   * @return array of corresponding ranks in {@code group2}
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_TRANSLATE_RANKS}.
   *
   * <p>Result elements are {@code MPI.UNDEFINED} where no correspondence exists.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static int[] translateRanks(Group group1, int[] ranks1, Group group2) throws MPIException {

    return nativetransRanks(group1.getHandle(), group2.getHandle(), ranks1);
  }

  /**
   * Compare two groups.
   *
   * @param group1 first group.
   * @param group2 second group.
   * @return result.
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_COMPARE}.
   *
   * <p>{@code MPI.IDENT} results if the group members and group order are exactly the same in both
   * groups. {@code MPI.SIMILAR} results if the group members are the same but the order is
   * different. {@code MPI.UNEQUAL} results otherwise.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static int compare(Group group1, Group group2) throws MPIException {

    return nativeCompare(group1.getHandle(), group2.getHandle());
  }

  /**
   * Set union of two groups.
   *
   * @param group1 first group.
   * @param group2 second group.
   * @return union group.
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_UNION}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Group union(Group group1, Group group2) throws MPIException {

    return new Group(nativeUnion(group1.getHandle(), group2.getHandle()));
  }

  /**
   * Set intersection of two groups.
   *
   * @param group1 first group.
   * @param group2 second group.
   * @return intersection group.
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_INTERSECTION}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Group intersection(Group group1, Group group2) throws MPIException {

    return new Group(nativeIntersection(group1.getHandle(), group2.getHandle()));
  }

  /**
   * Result contains all elements of the first group that are not in the second group.
   *
   * @param group1 first group.
   * @param group2 second group.
   * @return difference group.
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_DIFFERENCE}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Group difference(Group group1, Group group2) throws MPIException {

    return new Group(nativeDifference(group1.getHandle(), group2.getHandle()));
  }

  private static native void init();

  private static native int[] nativetransRanks(long handleGroup1, long handleGroup2, int[] ranks1);

  private static native int nativeCompare(long handleGroup1, long handleGroup2);

  private static native long nativeUnion(long handleGroup1, long handleGroup2);

  private static native long nativeIntersection(long handleGroup1, long handleGroup2);

  private static native long nativeDifference(long handleGroup1, long handleGroup2);

  long getHandle() {
    return handleOfGroup;
  }

  /**
   * Size of group.
   *
   * @return number of processors in the group.
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_SIZE}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int getSize() throws MPIException {
    return nativeSize();
  }

  /**
   * Rank of this process in group.
   *
   * @return rank of the calling process in the group.
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_RANK}.
   *
   * <p>Result value is {@code MPI.UNDEFINED} if this process is not a member of the group.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int getRank() throws MPIException {
    return nativeRank();
  }

  /**
   * Create a subset group including specified processes.
   *
   * @param ranks ranks from this group to appear in new group.
   * @return new group.
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_INCL}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Group incl(int[] ranks) throws MPIException {

    if (ranks == null) {
      throw new MPIException("Group.Incl does not accept a null integer " + " array ");
    }

    if (ranks.length == 0) {
      return MPI.GROUP_EMPTY;
    }

    return new Group(nativeIncl(ranks));
  }

  /**
   * Create a subset group excluding specified processes.
   *
   * @param ranks ranks from thsi group<em>not</em> to appear in new group.
   * @return new group.
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_EXCL}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Group excl(int[] ranks) throws MPIException {
    if (ranks == null) {
      throw new MPIException("Group.Excl does not accept a null integer " + " array ");
    }

    if (ranks.length == 0) {
      return this;
    }

    return new Group(nativeExcl(ranks));
  }

  /**
   * Create a subset group including processes specified by strided intervals of ranks.
   *
   * @param ranges array of integer triplets.
   * @return new group.
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_RANGE_INCL}.
   *
   * <p>The triplets are of the form (first rank, last rank, stride) indicating ranks in this group
   * to be included in the new group. The size of the first dimension of {@code ranges} is the
   * number of triplets. The size of the second dimension is 3.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Group rangeIncl(int[][] ranges) throws MPIException {
    if (ranges == null) {
      throw new MPIException(
          "Group.Range_incl does not accept null integer " + "array as argument");
    }

    // convert the 2D ranges into 1D and pass it to the nativeRangeIncl
    //
    // as we know that the dimension is 3 so ranges[len][3]
    // here find len with ranges.length

    int len = ranges.length;
    int oneDimRanges[] = new int[len * 3];

    for (int i = 0; i < len; i++) {
      oneDimRanges[(i * 3) + 0] = ranges[i][0];
      oneDimRanges[(i * 3) + 1] = ranges[i][1];
      oneDimRanges[(i * 3) + 2] = ranges[i][2];
    }

    return new Group(nativeRangeIncl(len, oneDimRanges));
  }

  /**
   * Create a subset group excluding processes specified by strided intervals of ranks.
   *
   * @param ranges array of integer triplets.
   * @return new group.
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_RANGE_EXCL}.
   *
   * <p>Triplet array is defined as for {@code Range_incl}, the ranges indicating ranks in this
   * group to be excluded from the new group.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Group rangeExcl(int[][] ranges) throws MPIException {
    if (ranges == null) {
      throw new MPIException(
          "Group.Range_incl does not accept null integer " + "array as argument");
    }

    // convert the 2D ranges into 1D and pass it to the nativeRangeIncl
    //
    // as we know that the dimension is 3 so ranges[len][3]
    // here find len with ranges.length
    int len = ranges.length;
    int oneDimRanges[] = new int[len * 3];

    for (int i = 0; i < len; i++) {
      oneDimRanges[(i * 3) + 0] = ranges[i][0];
      oneDimRanges[(i * 3) + 1] = ranges[i][1];
      oneDimRanges[(i * 3) + 2] = ranges[i][2];
    }

    return new Group(nativeRangeExcl(len, oneDimRanges));
  }

  /**
   * Destructor.
   *
   * <p>Java binding of the MPI operation {@code MPI_GROUP_FREE}.
   */
  public void free() {
    nativeFree();
  }

  private native void GetGroup(int Type);

  private native void nativeFree();

  private native int nativeSize();

  private native int nativeRank();

  private native long nativeIncl(int[] ranks);

  private native long nativeExcl(int[] ranks);

  private native long nativeRangeIncl(int n, int[] ranges);

  private native long nativeRangeExcl(int n, int[] ranges);
}
