/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
The MIT License

Copyright (c) 2005 - 2008
  1. Distributed Systems Group, University of Portsmouth (2005)
  2. Community Grids Laboratory, Indiana University (2005)
  3. Aamir Shafi (2005 - 2008)
  4. Bryan Carpenter (2005 - 2008)
  5. Mark Baker (2005 - 2008)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/*
 * File         : Intercomm.java
 * Author       : Aamir Shafi, Bryan Carpenter
 * Created      : Fri Sep 10 12:22:15 BST 2004
 * Revision     : $Revision: 1.10 $
 * Updated      : $Date: 2005/07/29 14:03:09 $
 */

package mpi;

/**
 * This class represents intercommunicators.
 */
public class Intercomm extends Comm {

  Intercomm(long handle) {
    super(handle);
  }

  /**
   * Get the parent of this communicator.
   *
   * @return parent of this communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_GET_PARENT}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Intercomm getParent() throws MPIException {
    return new Intercomm(nativeGetParent());
  }

  private static native long nativeGetParent() throws MPIException;

  /**
   * Size of remote group.
   *
   * @return number of process in remote group of this communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_REMOTE_SIZE}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int getRemoteSize() throws MPIException {
    return Remote_size();
  }

  private native int Remote_size() throws MPIException;

  /**
   * Return the remote group.
   *
   * @return remote group of this communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_REMOTE_GROUP}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Group getRemoteGroup() throws MPIException {
    return new Group(remote_group());
  }

  private native long remote_group();

  /**
   * Create an inter-communicator.
   *
   * @param high {@code true} if the local group has higher ranks in combined
   *               group, otherwise {@code false}.
   * @return new intra-communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_INTERCOMM_MERGE}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Intracomm merge(boolean high) throws MPIException {
    return new Intracomm(nativeMerge(high));
  }

  /**
   * Duplicate this communicator.
   *
   * @return copy of this communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_COMM_DUP}.
   *
   * <p>The new communicator is ``congruent'' to the old one, but has a different context.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Intercomm clone() throws MPIException {
    return new Intercomm(super.dup(getHandle()));
  }

  private native long nativeMerge(boolean high);
}
