/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
The MIT License

Copyright (c) 2005 - 2008
  1. Distributed Systems Group, University of Portsmouth (2005)
  2. Community Grids Laboratory, Indiana University (2005)
  3. Aamir Shafi (2005 - 2008)
  4. Bryan Carpenter (2005 - 2008)
  5. Mark Baker (2005 - 2008)
  6. Bibrak Qamar (2013 - 2014)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/*
 * File         : Intracomm.java
 * Author       : Aamir Shafi, Bryan Carpenter
 * Created      : Fri Sep 10 12:22:15 BST 2004
 * Revision     : $Revision: 1.30 $
 * Updated      : $Date: 2014/03/11 13:13:10 PKT $
 */

package mpi;

import mpjbuf.*;
import java.nio.*;

/**
 * This class represents intracommunicators.
 */
public class Intracomm extends Comm {

  Intracomm() {}

  // MPI.java COMM_WORLD is made using this
  Intracomm(int code) throws MPIException {
    super(code);
  }

  Intracomm(long handle) throws MPIException {
    super(handle);
  }

  Intracomm(Comm mpjdevComm, Group group_) throws MPIException {
    this.mpjdevComm = mpjdevComm;
    this.group = group_;
  }

  /**
   * Partition the group associated with this communicator and create a new communicator within each
   * subgroup.
   *
   * @param color control of subset assignment.
   * @param key control of rank assignment.
   * @return new communicator.
   *
   * Java binding of the MPI operation {@code MPI_COMM_SPLIT}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Intracomm split(int color, int key) throws MPIException {

    long newCommHandle = nativeSplit(this.handle, color, key);

    if (newCommHandle == MPI.COMM_NULL.getHandle()) return null;
    else return new Intracomm(newCommHandle);
  }

  /**
   * Duplicate this communicator.
   *
   * @return copy of this communicator.
   *
   * Java binding of the MPI operation {@code MPI_COMM_DUP}.
   *
   * The new communicator is ``congruent'' to the old one, but has a different context.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Intracomm clone() throws MPIException {
    return new Intracomm(dup(getHandle()));
  }

  /**
   * Create a new communicator.
   *
   * @param group group which is a subset of the group of this communicator.
   * @return new communicator.
   *
   * Java binding of the MPI operation {@code MPI_COMM_CREATE}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Intracomm create(Group group) throws MPIException {

    long newCommHandle = nativeCreate(group.getHandle());

    if (newCommHandle == MPI.COMM_NULL.getHandle()) return null;
    else return new Intracomm(newCommHandle);
  }

  /**
   * A call to {@code Barrier} blocks the caller until all process in the group have called it.
   *
   * Java binding of the MPI operation {@code MPI_BARRIER}.
   */
  public void barrier() throws MPIException {
    nativeBarrier(this.getHandle());
  }

  /**
   * Broadcast a message from the process with rank <tt>root</tt> to all processes of the group.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param buf buffer array.
   * @param count number of items in buffer.
   * @param type datatype of each item in buffer.
   * @param root rank of broadcast root.
   *
   * <p>Java binding of the MPI operation {@code MPI_BCAST}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void bcast(Object buf, int count, Datatype type, int root) throws MPIException {

    if (buf instanceof ByteBuffer) {
      if (type != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      nativeBcast(this.getHandle(), (ByteBuffer) buf, count, root);
      return;
    }

    arraysBcast(buf, count, type, root);
  }

  private void arraysBcast(Object buf, int count, Datatype type, int root) {

    int index = this.getRank();

    int numBytes = count * type.getByteSize();
    ByteBuffer wBuffer = ByteBuffer.allocateDirect(numBytes);

    try {
      if (index == root) {
        byteBufferSetData(buf, wBuffer, count, type.getType());
        wBuffer.flip();
      }
      wBuffer.limit(numBytes);

      nativeBcast(this.getHandle(), wBuffer, numBytes, root);

      byteBufferGetData(wBuffer, buf, count, type.getType());
      wBuffer.clear();

    } catch (Exception e) {
      e.printStackTrace();
    }
    return;
  } 

  /**
   * Each process sends the contents of its send buffer to the root process.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are  {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param sendcount number of items to send.
   * @param sendtype datatype of each item in send buffer.
   * @param recvbuf receive buffer array.
   * @param recvcount number of items to receive.
   * @param recvtype datatype of each item in receive buffer.
   * @param root rank of receiving process.
   *
   * <p>Java binding of the MPI operation {@code MPI_GATHER}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void gather(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int recvcount,
      Datatype recvtype,
      int root)
      throws MPIException {

    if (sendbuf instanceof java.nio.Buffer) {
      if (sendtype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      nativeGather(
          this.getHandle(),
          (java.nio.Buffer) sendbuf,
          sendcount,
          (java.nio.Buffer) recvbuf,
          recvcount,
          root);
      return;
    }

    arraysGather(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root);
  }

  private void arraysGather(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int recvcount,
      Datatype recvtype,
      int root) {

    int index = this.getRank(); // Get the rank of the current process
    int size = this.getSize();

    int numSendBytes = sendcount * sendtype.getByteSize();
    ByteBuffer wBuffer = ByteBuffer.allocateDirect(numSendBytes);

    ByteBuffer rBuffer = null;
    int numRecvBytes = -1;

    if (index == root) {
      numRecvBytes = recvcount * recvtype.getByteSize() * size;
      rBuffer = ByteBuffer.allocateDirect(numRecvBytes);
    }

    try {
      byteBufferSetData(sendbuf, wBuffer, sendcount, sendtype.getType());
      wBuffer.flip();
      wBuffer.limit(numSendBytes);

      nativeGather(
          this.getHandle(),
          wBuffer,
          numSendBytes,
          rBuffer,
          numSendBytes,
          root);

      if (index == root) { // root is the source point of gather
        rBuffer.limit(numRecvBytes);
        byteBufferGetData(rBuffer, recvbuf, recvcount * size, recvtype.getType());
        rBuffer.clear();
      } // ends if root

    } catch (Exception e) {
      e.printStackTrace();
    }
    return;
  }

  /**
   * Extends functionality of <tt>Gather</tt> by allowing varying counts of data from each process.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}
   * {@code MPI.SHORT}, {@code MPI.INT}, {@code MPI.LONG}, {@code MPI.BOOLEAN}
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are  {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param sendcount number of items to send.
   * @param sendtype datatype of each item in send buffer.
   * @param recvbuf receive buffer array.
   * @param recvcount number of elements received from each process.
   * @param displs displacements at which to place incoming data.
   * @param recvtype datatype of each item in receive buffer.
   * @param root rank of receiving process.
   *
   * <p>Java binding of the MPI operation {@code MPI_GATHERV}.
   *
   * <p>The sizes of arrays {@code recvcount} and {@code displs} should be the size of the group.
   * <p>Entry {@code i} of {@code displs} specifies the displacement relative to element
   * {@code recvoffset} of {@code recvbuf} at which to place incoming data.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void gatherv(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int[] recvcount,
      int[] displs,
      Datatype recvtype,
      int root)
      throws MPIException {

    int sendoffset = 0;
    int recvoffset = 0;

    if (sendbuf instanceof java.nio.Buffer) {
      if (sendtype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      nativeGatherv(
          this.getHandle(),
          (java.nio.Buffer) sendbuf,
          sendcount,
          sendtype.getType(),
          (java.nio.Buffer) recvbuf,
          recvcount,
          recvtype.getType(),
          displs,
          root);
      return;
    }

    arraysGatherv(
        sendbuf,
        sendcount,
        sendtype,
        recvbuf,
        recvcount,
        displs,
        recvtype,
        root);
  }

  private void arraysGatherv(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int[] recvcount,
      int[] displs,
      Datatype recvtype,
      int root)
      throws MPIException {

    int rank = this.getRank();
    int size = this.getSize();
    int recvCountTotal = 0;

    int[] recvcountBytes = null; 
    int[] displsBytes = null;

    if (rank == root) {
      recvCountTotal = displs[size - 1] + recvcount[size - 1];

      recvcountBytes = new int[size];
      displsBytes = new int[size];


      for (int i = 0; i < size; i++) {
        recvcountBytes[i] = recvcount[i] * recvtype.getByteSize();
        displsBytes[i] = displs[i] * recvtype.getByteSize();
      }

    }

    ByteBuffer wBuffer = null;
    ByteBuffer rBuffer = null;

    int numSendBytes = sendcount * sendtype.getByteSize();
    wBuffer = ByteBuffer.allocateDirect(numSendBytes);
    wBuffer.order(ByteOrder.LITTLE_ENDIAN);

    int numRecvBytes = -1;

    if (rank == root) {
      numRecvBytes = recvCountTotal * recvtype.getByteSize();
      rBuffer = ByteBuffer.allocateDirect(numRecvBytes);
      rBuffer.order(ByteOrder.LITTLE_ENDIAN);
    }

    try {

      byteBufferSetData(sendbuf, wBuffer, sendcount, sendtype.getType());
      wBuffer.flip();
      wBuffer.limit(numSendBytes);

      nativeGatherv(
          this.getHandle(),
          wBuffer,
          numSendBytes,
          sendtype.getType(),
          rBuffer,
          recvcountBytes,
          recvtype.getType(),
          displsBytes,
          root);

      if (rank == root) {
        rBuffer.limit(numRecvBytes);
        byteBufferGetData(rBuffer, recvbuf, recvCountTotal, recvtype.getType());
        rBuffer.clear();
      }

    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * Inverse of the operation {@code Gather}.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param sendcount number of items to send.
   * @param sendtype datatype of each item in send buffer.
   * @param recvbuf receive buffer array.
   * @param recvcount number of items to receive.
   * @param recvtype datatype of each item in receive buffer.
   * @param root rank of sending process.
   *
   * <p>Java binding of the MPI operation {@code MPI_SCATTER}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void scatter(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int recvcount,
      Datatype recvtype,
      int root)
      throws MPIException {

    if (sendbuf instanceof java.nio.Buffer) {
      if (sendtype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      nativeScatter(
          this.getHandle(),
          (java.nio.Buffer) sendbuf,
          sendcount,
          (java.nio.Buffer) recvbuf,
          recvcount,
          root);
      return;
    }

    arraysScatter(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root);
  }

  private void arraysScatter(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int recvcount,
      Datatype recvtype,
      int root) {

    int index = this.getRank();
    int size = this.getSize();

    ByteBuffer wBuffer = null;
    int numSendBytes = -1;

    if (index == root) {
      numSendBytes = sendcount * sendtype.getByteSize() * size;
      wBuffer = ByteBuffer.allocateDirect(numSendBytes);
    }

    ByteBuffer rBuffer = null;
    int numRecvBytes = -1;

    numRecvBytes = recvcount * recvtype.getByteSize();
    rBuffer = ByteBuffer.allocateDirect(numRecvBytes);

    try {
      if (index == root) {
        byteBufferSetData(sendbuf, wBuffer, sendcount * size, sendtype.getType());
        wBuffer.flip();
        wBuffer.limit(numSendBytes);
      }

      nativeScatter(
          this.getHandle(),
          wBuffer,
          numRecvBytes,
          rBuffer,
          numRecvBytes,
          root);

      rBuffer.limit(numRecvBytes);
      byteBufferGetData(rBuffer, recvbuf, recvcount, recvtype.getType());
      rBuffer.clear();
    } catch (Exception e) {
      e.printStackTrace();
    }
    return;
  }

  /**
   * Inverse of the operation {@code Gatherv}.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.INT}, {@code MPI.LONG}, {@code MPI.BOOLEAN},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param sendcount number of items to send.
   * @param displs displacements from which to take outgoing data.
   * @param sendtype datatype of each item in send buffer.
   * @param recvbuf receive buffer array.
   * @param recvcount number of items to receive.
   * @param recvtype datatype of each item in receive buffer.
   * @param root rank of sending process.
   *
   * <p>Java binding of the MPI operation {@code MPI_SCATTERV}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void scatterv(
      Object sendbuf,
      int[] sendcount,
      int[] displs,
      Datatype sendtype,
      Object recvbuf,
      int recvcount,
      Datatype recvtype,
      int root)
      throws MPIException {

    if (sendbuf instanceof java.nio.Buffer) {
      if (sendtype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      nativeScatterv(
          this.getHandle(),
          (java.nio.Buffer) sendbuf,
          sendcount,
          sendtype.getType(),
          displs,
          (java.nio.Buffer) recvbuf,
          recvcount,
          recvtype.getType(),
          root);
      return;
    }

    arraysScatterv(
        sendbuf,
        sendcount,
        displs,
        sendtype,
        recvbuf,
        recvcount,
        recvtype,
        root);
  }

  private void arraysScatterv(
      Object sendbuf,
      int[] sendcount,
      int[] displs,
      Datatype sendtype,
      Object recvbuf,
      int recvcount,
      Datatype recvtype,
      int root)
      throws MPIException {

    int rank = this.getRank();
    int size = this.getSize();
    int sendCountTotal = 0;

    if (rank == root) {
      sendCountTotal = displs[size - 1] + sendcount[size - 1];
    }

    ByteBuffer wBuffer = null;
    ByteBuffer rBuffer = null;

    int numSendBytes = -1;

    int[] sendcountBytes = null;
    int[] displsBytes = null;

    if (rank == root) {
      numSendBytes = sendCountTotal * sendtype.getByteSize();
      wBuffer = ByteBuffer.allocateDirect(numSendBytes);
      wBuffer.order(ByteOrder.LITTLE_ENDIAN);
      
      sendcountBytes = new int[size];
      displsBytes = new int[size];

      for (int i = 0; i < size; i++) {
        sendcountBytes[i] = sendcount[i] * sendtype.getByteSize();
        displsBytes[i] = displs[i] * sendtype.getByteSize();
      }  
    }

    int numRecvBytes = recvcount * recvtype.getByteSize();
    rBuffer = ByteBuffer.allocateDirect(numRecvBytes);
    rBuffer.order(ByteOrder.LITTLE_ENDIAN);

    try {

      if (rank == root) {
        byteBufferSetData(sendbuf, wBuffer, sendCountTotal, sendtype.getType());
        wBuffer.flip();
        wBuffer.limit(numSendBytes);
      }

      nativeScatterv(
          this.getHandle(),
          wBuffer,
          sendcountBytes,
          sendtype.getType(),
          displsBytes,
          rBuffer,
          numRecvBytes,
          recvtype.getType(),
          root);

      rBuffer.limit(numRecvBytes);
      byteBufferGetData(rBuffer, recvbuf, recvcount, recvtype.getType());
      rBuffer.clear();

    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * Similar to {@code Gather}, but all processes receive the result.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}, 
   * {@code MPI.SHORT}, {@code MPI.BOOLEAN}, {@code MPI.INT}, {@code MPI.LONG},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param sendcount number of items to send.
   * @param sendtype datatype of each item in send buffer.
   * @param recvbuf receive buffer array.
   * @param recvcount number of items to receive.
   * @param recvtype datatype of each item in receive buffer.
   *
   * <p>Java binding of the MPI operation {@code MPI_ALLGATHER}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void allGather(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int recvcount,
      Datatype recvtype)
      throws MPIException {

    if (sendbuf instanceof java.nio.Buffer) {
      if (sendtype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      nativeAllgather(
          this.getHandle(),
          (java.nio.Buffer) sendbuf,
          sendcount,
          (java.nio.Buffer) recvbuf,
          recvcount);
      return;
    }

    arraysAllGather(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype);
  }

  private void arraysAllGather(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int recvcount,
      Datatype recvtype) {

    int index = this.getRank();
    int size = this.getSize();

    ByteBuffer wBuffer = null;
    int numSendBytes = -1;

    numSendBytes = sendcount * sendtype.getByteSize();
    wBuffer = ByteBuffer.allocateDirect(numSendBytes);

    ByteBuffer rBuffer = null;
    int numRecvBytes = -1;

    numRecvBytes = recvcount * recvtype.getByteSize() * size;
    rBuffer = ByteBuffer.allocateDirect(numRecvBytes);

    try {

      byteBufferSetData(sendbuf, wBuffer, sendcount, sendtype.getType());
      wBuffer.flip();
      wBuffer.limit(numSendBytes);

      nativeAllgather(
          this.getHandle(),
          (java.nio.Buffer) wBuffer,
          numSendBytes,
          (java.nio.Buffer) rBuffer,
          numSendBytes);

      rBuffer.limit(numRecvBytes);
      byteBufferGetData(rBuffer, recvbuf, recvcount * size, recvtype.getType());
      rBuffer.clear();

    } catch (Exception e) {
      e.printStackTrace();
    }
    return;
  }

  /**
   * Similar to {@code Gatherv}, but all processes receive the result.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR},
   * {@code MPI.SHORT}, {@code MPI.INT}, {@code MPI.LONG}, {@code MPI.BOOLEAN},
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are  {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param sendcount number of items to send.
   * @param sendtype datatype of each item in send buffer.
   * @param recvbuf receive buffer array.
   * @param recvcount number of items to receive.
   * @param displs displacements at which to place incoming data.
   * @param recvtype datatype of each item in receive buffer.
   *
   * <p>Java binding of the MPI operation {@code MPI_ALLGATHERV}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void allGatherv(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int[] recvcount,
      int[] displs,
      Datatype recvtype)
      throws MPIException {

    if (sendbuf instanceof java.nio.Buffer) {
      if (sendtype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      nativeAllgatherv(
          this.getHandle(),
          (java.nio.Buffer) sendbuf,
          sendcount,
          sendtype.getType(),
          (java.nio.Buffer) recvbuf,
          recvcount,
          recvtype.getType(),
          displs);
      return;
    }

    arraysAllGatherv(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, displs, recvtype);
  }

  private void arraysAllGatherv(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int[] recvcount,
      int[] displs,
      Datatype recvtype)
      throws MPIException {

    int rank = this.getRank();
    int size = this.getSize();
    int recvCountTotal = 0;
    int displsTotal = 0;

    recvCountTotal = displs[size - 1] + recvcount[size - 1];

    ByteBuffer wBuffer = null;
    ByteBuffer rBuffer = null;

    int[] recvcountBytes = new int[size];
    int[] displsBytes = new int[size];

    for (int i = 0; i < size; i++) {
      recvcountBytes[i] = recvcount[i] * recvtype.getByteSize();
      displsBytes[i] = displs[i] * recvtype.getByteSize();  
    }

    int numSendBytes = sendcount * sendtype.getByteSize();
    wBuffer = ByteBuffer.allocateDirect(numSendBytes);
    wBuffer.order(ByteOrder.LITTLE_ENDIAN);

    int numRecvBytes = -1;

    numRecvBytes = recvCountTotal * recvtype.getByteSize();
    rBuffer = ByteBuffer.allocateDirect(numRecvBytes);
    rBuffer.order(ByteOrder.LITTLE_ENDIAN);

    try {

      byteBufferSetData(sendbuf, wBuffer, sendcount, sendtype.getType());
      wBuffer.flip();
      wBuffer.limit(numSendBytes);

      nativeAllgatherv(
          this.getHandle(),
          wBuffer,
          numSendBytes,
          sendtype.getType(),
          rBuffer,
          recvcountBytes,
          recvtype.getType(),
          displsBytes);

      rBuffer.limit(numRecvBytes);
      byteBufferGetData(rBuffer, recvbuf, recvCountTotal, recvtype.getType());
      rBuffer.clear();

    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * Extension of {@code Allgather} to the case where each process sends distinct data to each of
   * the receivers.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.BOOLEAN}
   * {@code MPI.SHORT}, {@code MPI.INT}, {@code MPI.LONG}, {@code MPI.CHAR}
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are  {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param sendcount number of items to send.
   * @param sendtype datatype of each item in send buffer.
   * @param recvbuf receive buffer array.
   * @param recvcount number of items to receive.
   * @param recvtype datatype to receive buffer items.
   *
   * <p>Java binding of the MPI operation {@code MPI_ALLTOALL}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void allToAll(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int recvcount,
      Datatype recvtype)
      throws MPIException {

    if (sendbuf instanceof java.nio.Buffer) {
      if (sendtype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      nativeAlltoall(
          this.getHandle(),
          (java.nio.Buffer) sendbuf,
          sendcount,
          (java.nio.Buffer) recvbuf,
          recvcount);
      return;
    }

    arraysAllToAll(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype);
  }

  private void arraysAllToAll(
      Object sendbuf,
      int sendcount,
      Datatype sendtype,
      Object recvbuf,
      int recvcount,
      Datatype recvtype) {

    int me = this.getRank();

    int size = this.getSize();

    int numSendBytes = sendcount * sendtype.getByteSize() * size;
    ByteBuffer wBuffer = ByteBuffer.allocateDirect(numSendBytes);

    int numRecvBytes = recvcount * recvtype.getByteSize() * size;
    ByteBuffer rBuffer = ByteBuffer.allocateDirect(numRecvBytes);

    try {

      byteBufferSetData(sendbuf, wBuffer, sendcount * size, sendtype.getType());
      wBuffer.flip();
      wBuffer.limit(numSendBytes);

      nativeAlltoall(
          this.getHandle(),
          wBuffer,
          numSendBytes / size,
          rBuffer,
          numRecvBytes / size);

      rBuffer.limit(numRecvBytes);
      byteBufferGetData(rBuffer, recvbuf, recvcount * size, recvtype.getType());
      rBuffer.clear();

    } catch (Exception e) {
      e.printStackTrace();
    }

    wBuffer = null;
    rBuffer = null;

    return;
  }

  /**
   * Adds flexibility to {@code Alltoall}: location of data for send is specified by
   * {@code sdispls} and location to place data on receive side is specified by {@code rdispls}.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.CHAR}
   * {@code MPI.SHORT}, {@code MPI.INT}, {@code MPI.LONG}, {@code MPI.BOOLEAN}
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are  {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param sendcount number of items to send.
   * @param sdispls displacements from which to take outgoing data.
   * @param sendtype datatype of each item in send buffer.
   * @param recvbuf receive buffer array.
   * @param recvcount number of items to receive.
   * @param rdispls displacements at which to place incoming data.
   * @param recvtype datatype to receive buffer items.
   *
   * <p>Java binding of the MPI operation {@code MPI_ALLTOALLV}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void allToAllv(
      Object sendbuf,
      int[] sendcount,
      int[] sdispls,
      Datatype sendtype,
      Object recvbuf,
      int[] recvcount,
      int[] rdispls,
      Datatype recvtype)
      throws MPIException {

    if (sendbuf instanceof java.nio.Buffer) {
      if (sendtype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      nativeAlltoallv(
          this.getHandle(),
          (java.nio.Buffer) sendbuf,
          sendcount,
          sendtype.getType(),
          sdispls,
          (java.nio.Buffer) recvbuf,
          recvcount,
          recvtype.getType(),
          rdispls);
      return;
    }

    arraysAllToAllv(
        sendbuf,
        sendcount,
        sdispls,
        sendtype,
        recvbuf,
        recvcount,
        rdispls,
        recvtype);
  }

  private void arraysAllToAllv(
      Object sendbuf,
      int[] sendcount,
      int[] sdispls,
      Datatype sendtype,
      Object recvbuf,
      int[] recvcount,
      int[] rdispls,
      Datatype recvtype)
      throws MPIException {

    int rank = this.getRank();
    int size = this.getSize();
    int sendCountTotal = sdispls[size - 1] + sendcount[size - 1];
    int recvCountTotal = rdispls[size - 1] + recvcount[size - 1];

    ByteBuffer wBuffer = null;
    ByteBuffer rBuffer = null;

    int[] sendcountBytes = new int[size];
    int[] recvcountBytes = new int[size];
    int[] sdisplsBytes = new int[size];
    int[] rdisplsBytes = new int[size];

    for (int i = 0; i < size; i++) {
        sendcountBytes[i] = sendcount[i] * sendtype.getByteSize();
        recvcountBytes[i] = recvcount[i] * recvtype.getByteSize();
        sdisplsBytes[i] = sdispls[i] * sendtype.getByteSize();
        rdisplsBytes[i] = rdispls[i] * recvtype.getByteSize();
    }

    int numSendBytes = sendCountTotal * sendtype.getByteSize();
    wBuffer = ByteBuffer.allocateDirect(numSendBytes);
    wBuffer.order(ByteOrder.LITTLE_ENDIAN);

    int numRecvBytes = recvCountTotal * recvtype.getByteSize();
    rBuffer = ByteBuffer.allocateDirect(numRecvBytes);
    rBuffer.order(ByteOrder.LITTLE_ENDIAN);

    try {

      byteBufferSetData(sendbuf, wBuffer, sendCountTotal, sendtype.getType());
      wBuffer.flip();
      wBuffer.limit(numSendBytes);

      nativeAlltoallv(
          this.getHandle(),
          wBuffer,
          sendcountBytes,
          sendtype.getType(),
          sdisplsBytes,
          rBuffer,
          recvcountBytes,
          recvtype.getType(),
          rdisplsBytes);

      rBuffer.limit(numRecvBytes);
      byteBufferGetData(rBuffer, recvbuf, recvCountTotal, recvtype.getType());
      rBuffer.clear();

    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * Combine elements in input buffer of each process using the reduce operation, and return the
   * combined value in the output buffer of the root process.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.BOOLEAN},
   * {@code MPI.SHORT}, {@code MPI.INT}, {@code MPI.LONG}, {@code MPI.CHAR}
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are  {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param recvbuf receive buffer array.
   * @param count number of items in send buffer.
   * @param datatype datatype of each item in send buffer.
   * @param op reduce operation.
   * @param root rank of root process.
   *
   * <p>Java binding of the MPI operation {@code MPI_REDUCE}.
   *
   * <p>The predefined operations are available in Java as {@code MPI.MAX}, {@code MPI.MIN},
   * {@code MPI.SUM}, {@code MPI.PROD}, {@code MPI.LAND}, {@code MPI.BAND}, {@code MPI.LOR},
   * {@code MPI.BOR}, {@code MPI.LXOR}, {@code MPI.BXOR}, {@code MPI.MINLOC} and
   * {@code MPI.MAXLOC}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void reduce(Object sendbuf, Object recvbuf, int count, Datatype datatype, Op op, int root)
      throws MPIException {

    if (sendbuf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      nativeReduce(
          this.getHandle(),
          (java.nio.Buffer) sendbuf,
          (java.nio.Buffer) recvbuf,
          op.opCode,
          count,
          datatype.getType(),
          root);
      return;
    }

    try {
      arraysReduce(sendbuf, recvbuf, count, datatype, op, root);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  private void arraysReduce(
      Object sendbuf, Object recvbuf, int count, Datatype datatype, Op op, int root)
      throws MPIException, BufferException {

    int index = this.getRank();
    int size = this.getSize();

    int numSendBytes = count * datatype.getByteSize();
    ByteBuffer wBuffer = ByteBuffer.allocateDirect(numSendBytes);
    wBuffer.order(ByteOrder.LITTLE_ENDIAN);

    ByteBuffer rBuffer = null;
    int numRecvBytes = -1;

    if (index == root) {
      numRecvBytes = count * datatype.getByteSize();
      rBuffer = ByteBuffer.allocateDirect(numRecvBytes);
      rBuffer.order(ByteOrder.LITTLE_ENDIAN);
    }

    try {
      byteBufferSetData(sendbuf, wBuffer, count, datatype.getType());
      wBuffer.flip();
      wBuffer.limit(numSendBytes);

      nativeReduce(this.getHandle(), wBuffer, rBuffer, op.opCode, count, datatype.getType(), root);

      if (index == root) {
        rBuffer.limit(numRecvBytes);
        byteBufferGetData(rBuffer, recvbuf, count, datatype.getType());
        rBuffer.clear();
      }

    } catch (Exception e) {
      e.printStackTrace();
    }
    return;
  }

  /**
   * Same as {@code reduce} except that the result appears in receive buffer of all process in the
   * group.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.BYTE}, {@code MPI.BOOLEAN},
   * {@code MPI.SHORT}, {@code MPI.INT}, {@code MPI.LONG}, {@code MPI.CHAR}
   * {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported. For ByteBuffers, 
   * the datatypes we currently support are  {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param recvbuf receive buffer array.
   * @param count number of items in send buffer.
   * @param datatype datatype of each item in send buffer.
   * @param op reduce operation.
   *
   * <p>Java binding of the MPI operation {@code MPI_ALLREDUCE}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void allReduce(Object sendbuf, Object recvbuf, int count, Datatype datatype, Op op)
      throws MPIException {

    if (sendbuf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
        System.out.println("ERROR: Datatype is not supported!");
        return;
      }
      nativeAllReduce(
          getHandle(),
          (java.nio.Buffer) sendbuf,
          (java.nio.Buffer) recvbuf,
          op.opCode,
          datatype.getType(),
          count);
      return;
    }

    arraysAllReduce(sendbuf, recvbuf, count, datatype, op);
  }

  private void arraysAllReduce(
      Object sendbuf, Object recvbuf, int count, Datatype datatype, Op op) {

    int numSendBytes = count * datatype.getByteSize();
    ByteBuffer wBuffer = ByteBuffer.allocateDirect(numSendBytes);
    wBuffer.order(ByteOrder.LITTLE_ENDIAN);

    ByteBuffer rBuffer = null;
    int numRecvBytes = count * datatype.getByteSize();
    rBuffer = ByteBuffer.allocateDirect(numRecvBytes);
    rBuffer.order(ByteOrder.LITTLE_ENDIAN);

    try {
      byteBufferSetData(sendbuf, wBuffer, count, datatype.getType());
      wBuffer.flip();
      wBuffer.limit(numSendBytes);

      nativeAllReduce(getHandle(), wBuffer, rBuffer, op.opCode, datatype.getType(), count);

      rBuffer.limit(numRecvBytes);
      byteBufferGetData(rBuffer, recvbuf, count, datatype.getType());
      rBuffer.clear();

    } catch (Exception e) {
      e.printStackTrace();
    }

    return;
  }

  /**
   * Combine elements in input buffer of each process using the reduce operation, and scatter the
   * combined values over the output buffers of the processes.
   * 
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.SHORT}, {@code MPI.INT},
   * {@code MPI.BOOLEAN}, {@code MPI.CHAR}, {@code MPI.BYTE},
   * {@code MPI.LONG}, {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported.
   * For ByteBuffers, the datatypes we currently support are  {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param recvbuf receive buffer array.
   * @param recvcounts numbers of result elements distributed to each process.
   * @param datatype datatype of each item in send buffer.
   * @param op reduce operation.
   *
   * <p>Java binding of the MPI operation {@code MPI_REDUCE_SCATTER}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void reduceScatter(
      Object sendbuf, Object recvbuf, int[] recvcounts, Datatype datatype, Op op)
      throws MPIException {

    if (sendbuf instanceof java.nio.Buffer) {
      if (datatype != MPI.BYTE) {
          System.out.println("ERROR: Datatype is not supported!");
          return;
      }
      nativeReduceScatter(
          this.getHandle(),
          (java.nio.Buffer) sendbuf,
          (java.nio.Buffer) recvbuf,
          recvcounts,
          datatype.getType(),
          op.opCode);
      return;
    }

    arraysReduceScatter(sendbuf, recvbuf, recvcounts, datatype, op);
  }

  private void arraysReduceScatter(
      Object sendbuf, Object recvbuf, int[] recvcounts, Datatype datatype, Op op) {

    int rank = this.getRank();
    int size = this.getSize();
    int root = 0;
    int sendcount = 0;

    for (int k = 0; k < recvcounts.length; k++) {
      sendcount += recvcounts[k];
    }

    ByteBuffer wBuffer = null;
    ByteBuffer rBuffer = null;

    int numSendBytes = sendcount * datatype.getByteSize();
    wBuffer = ByteBuffer.allocateDirect(numSendBytes);
    wBuffer.order(ByteOrder.LITTLE_ENDIAN);

    int numRecvBytes = recvcounts[rank] * datatype.getByteSize();
    rBuffer = ByteBuffer.allocateDirect(numRecvBytes);
    rBuffer.order(ByteOrder.LITTLE_ENDIAN);

    try {

      byteBufferSetData(sendbuf, wBuffer, sendcount, datatype.getType());
      wBuffer.flip();
      wBuffer.limit(numSendBytes);

      nativeReduceScatter(
          this.getHandle(), wBuffer, rBuffer, recvcounts, datatype.getType(), op.opCode);

      rBuffer.limit(numRecvBytes);
      byteBufferGetData(rBuffer, recvbuf, recvcounts[rank], datatype.getType());
      rBuffer.clear();

    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * Perform a prefix reduction on data distributed across the group.
   *
   * <p>The method supports communication of both {@code ByteBuffers} and Java
   * arrays. For arrays, the datatypes {@code MPI.SHORT}, {@code MPI.INT},
   * {@code MPI.BOOLEAN}, {@code MPI.CHAR}, {@code MPI.BYTE},
   * {@code MPI.LONG}, {@code MPI.FLOAT}, and {@code MPI.DOUBLE} are supported.
   * For ByteBuffers, the datatypes we currently support are  {@code MPI.BYTE}.
   *
   * @param sendbuf send buffer array.
   * @param recvbuf receive buffer array.
   * @param count number of items in input buffer.
   * @param datatype datatype of each item in send buffer.
   * @param op reduce operations.
   *
   * <p>Java binding of the MPI operation {@code MPI_SCAN}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void scan(Object sendbuf, Object recvbuf, int count, Datatype datatype, Op op)
      throws MPIException {
     
    if (sendbuf instanceof ByteBuffer) {
      if (datatype != MPI.BYTE) {
        System.out.println("ERROR: Datatype is not supported!");
        return;
      }
      nativeScan(
              this.getHandle(),
              (ByteBuffer)sendbuf,
              (ByteBuffer)recvbuf,
              count,
              datatype.getType(),
              op.opCode);
      return;
    }

    arraysScan(
            sendbuf, recvbuf, count, datatype, op);
  }

  private void arraysScan(Object sendbuf, Object recvbuf, int count, Datatype datatype, Op op) {
    
    int numSendBytes = count * datatype.getByteSize();
    ByteBuffer wBuffer = ByteBuffer.allocateDirect(numSendBytes);
    wBuffer.order(ByteOrder.LITTLE_ENDIAN);

    ByteBuffer rBuffer = null;
    int numRecvBytes = count * datatype.getByteSize();
    rBuffer = ByteBuffer.allocateDirect(numRecvBytes);
    rBuffer.order(ByteOrder.LITTLE_ENDIAN);

    try {
      byteBufferSetData(sendbuf, wBuffer, count, datatype.getType());
      wBuffer.flip();
      wBuffer.limit(numSendBytes);

      nativeScan(getHandle(), wBuffer, rBuffer, count, datatype.getType(), op.opCode);

      rBuffer.limit(numRecvBytes);
      byteBufferGetData(rBuffer, recvbuf, count, datatype.getType());
      rBuffer.clear();

    } catch (Exception e) {
      e.printStackTrace();
    }

    return;
  }

  /**
   * Create a Cartesian topology communicator whose group is a subset of the group of this
   * communicator.
   *
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * @param dims the number of processes in each dimension.
   * @param periods {@code true} if grid is periodic in each dimension,
   *                  otherwise {@code false}.
   * @param reorder {@code true} if ranking may be reordered, {@code false}, {@code false} if not.
   * @return new Cartesian topology communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_CART_CREATE}.
   *
   * <p>The number of dimensions of the Cartesian grid is taken to be the size of the {@code dims}
   * argument. The array {@code periods} must be the same size.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Cartcomm createCart(int[] dims, boolean[] periods, boolean reorder) throws MPIException {
    return null;
  }

  /**
   * Create a graph topology communicator whose group is a subset of the group of this communicator.
   *
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * @param index node degrees.
   * @param edges graph edges.
   * @param reorder {@code true} if ranking may be reordered, {@code false}, {@code false} if not.
   * @return new graph topology communicator.
   *
   * <p>Java binding of the MPI operation {@code MPI_GRAPH_CREATE}.
   *
   * <p>The number of nodes in the graph, {@code nnodes}, is taken to be size of the {@code index}
   * argument. The size of array {@code edges} must be {@code index [nnodes} - 1]}.
   */
  public Graphcomm createGraph(int[] index, int[] edges, boolean reorder) throws MPIException {

    return null;
  }

  // utility methods - set data
  private void byteBufferSetData(
      Object src, ByteBuffer dest, int count, int typeCode) {

    if (typeCode == Datatype.BYTE) {
     
      dest.put((byte[]) src, 0, count);

    } else if (typeCode == Datatype.CHAR) {

      CharBuffer CharBuffer = dest.asCharBuffer();
      CharBuffer.put((char[]) src, 0, count);

    } else if (typeCode == Datatype.SHORT) {

      ShortBuffer ShortBuffer = dest.asShortBuffer();
      ShortBuffer.put((short[]) src, 0, count);

    } else if (typeCode == Datatype.BOOLEAN) {
     
      boolean srcB[] = (boolean[]) src;
      for (int i = 0; i < count; i++) {
        dest.put((byte) (srcB[i + 0] ? 1 : 0)); // boolean
      }

    } else if (typeCode == Datatype.INT) {
     
      IntBuffer IntBuffer = dest.asIntBuffer();
      IntBuffer.put((int[]) src, 0, count);

    } else if (typeCode == Datatype.LONG) {

      LongBuffer LongBuffer = dest.asLongBuffer();
      LongBuffer.put((long[]) src, 0, count);

    } else if (typeCode == Datatype.FLOAT) {

      FloatBuffer FloatBuffer = dest.asFloatBuffer();
      FloatBuffer.put((float[]) src, 0, count);

    } else if (typeCode == Datatype.DOUBLE) {

      DoubleBuffer DoubleBuffer = dest.asDoubleBuffer();
      DoubleBuffer.put((double[]) src, 0, count);
    }
  }

  // utility methods - get data
  private void byteBufferGetData(
      ByteBuffer src, Object dest, int count, int typeCode) {

    if (typeCode == Datatype.BYTE) {

      src.get((byte[]) dest, 0, count);

    } else if (typeCode == Datatype.CHAR) {

      CharBuffer CharBuffer = src.asCharBuffer();
      CharBuffer.get((char[]) dest, 0, count);

    } else if (typeCode == Datatype.SHORT) {

      ShortBuffer ShortBuffer = src.asShortBuffer();
      ShortBuffer.get((short[]) dest, 0, count);

    } else if (typeCode == Datatype.BOOLEAN) {

      boolean destB[] = (boolean[]) dest;

      for (int i = 0; i < count; i++) {
        destB[i + 0] = (src.get() == 1); // boolean
      }

    } else if (typeCode == Datatype.INT) {

      IntBuffer IntBuffer = src.asIntBuffer();
      IntBuffer.get((int[]) dest, 0, count);

    } else if (typeCode == Datatype.LONG) {

      LongBuffer LongBuffer = src.asLongBuffer();
      LongBuffer.get((long[]) dest, 0, count);

    } else if (typeCode == Datatype.FLOAT) {

      FloatBuffer FloatBuffer = src.asFloatBuffer();
      FloatBuffer.get((float[]) dest, 0, count);

    } else if (typeCode == Datatype.DOUBLE) {

      DoubleBuffer DoubleBuffer = src.asDoubleBuffer();
      DoubleBuffer.get((double[]) dest, 0, count);
    }
  }
  
  /**
   * Create a new intracommunicator for the given group.
   *
   * @param group group which is a subset of the group of this communicator
   * @param tag an integer tag
   * 
   * <p>Java binding of the MPI operation {@code MPI_COMM_CREATE_GROUP}.
   *
   * @return new communicator
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Intracomm createGroup(Group group, int tag) throws MPIException {
    return new Intracomm(nativeCreateGroup(this.handle, group.getHandle(), tag));
  }

  private native long nativeCreateGroup(long handle, long group, int tag) throws MPIException;

  private native long nativeCreate(long ngroup) throws MPIException;

  private native long nativeSplit(long handle, int color, int key) throws MPIException;

  private native void nativeBarrier(long commHandle);

  private native void nativeReduce(
      long commHandle,
      java.nio.Buffer sendbuf,
      java.nio.Buffer recvbuf,
      int op,
      int count,
      int datatype,
      int root);

  private native void nativeAllReduce(
      long commHandle,
      java.nio.Buffer sendbuf,
      java.nio.Buffer recvbuf,
      int op,
      int datatype,
      int count);

  private native void nativeGather(
      long commHandle,
      java.nio.Buffer sendbuf,
      int sendcount,
      java.nio.Buffer recvbuf,
      int recvcount,
      int root);

  public native void nativeGatherv(
      long commHandle,
      java.nio.Buffer sendbuf,
      int count,
      int datatype,
      java.nio.Buffer recvbuf,
      int[] recvcount,
      int recvtype,
      int[] displs,
      int root);

  private native void nativeAllgather(
      long commHandle,
      java.nio.Buffer sendbuf,
      int sendcount,
      java.nio.Buffer recvbuf,
      int recvcount);

  private native void nativeAllgatherv(
      long commHandle,
      java.nio.Buffer sendbuf,
      int sendcount,
      int sendtype,
      java.nio.Buffer recvbuf,
      int[] recvcount,
      int recvtype,
      int[] displs);

  private native void nativeScatter(
      long commHandle,
      java.nio.Buffer sendbuf,
      int sendcount,
      java.nio.Buffer recvbuf,
      int recvcount,
      int root);

  private native void nativeScatterv(
      long commHandle,
      java.nio.Buffer sendbuf,
      int[] sendcount,
      int sendtype,
      int[] displs,
      java.nio.Buffer recvbuf,
      int recvcount,
      int recvtype,
      int root);

  private native void nativeAlltoall(
      long commHandle,
      java.nio.Buffer sendbuf,
      int sendcount,
      java.nio.Buffer recvbuf,
      int recvcount);

  private native void nativeAlltoallv(
      long commHandle,
      java.nio.Buffer sendbuf,
      int[] sendcount,
      int sendtype,
      int[] sdispls,
      java.nio.Buffer recvbuf,
      int[] recvcount,
      int recvtype,
      int[] rdispls);

  private native void nativeReduceScatter(
      long commHandle,
      java.nio.Buffer sendbuf,
      java.nio.Buffer recvbuf,
      int[] recvcounts,
      int datatype,
      int op);

  private native void nativeScan(
      long commHandle,
      java.nio.Buffer sendbuf,
      java.nio.Buffer recvbuf,
      int count,
      int datatype,
      int op);
  
   private native void nativeBcast(
      long commHandle, ByteBuffer buffer, int count, int root);
}
