/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
The MIT License

Copyright (c) 2005 - 2008
1. Distributed Systems Group, University of Portsmouth (2005)
2. Community Grids Laboratory, Indiana University (2005)
3. Aamir Shafi (2005 - 2008)
4. Bryan Carpenter (2005 - 2008)
5. Mark Baker (2005 - 2008)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/*
 * File         : MPI.java
 * Author       : Aamir Shafi, Bryan Carpenter
 * Created      : Fri Sep 10 12:22:15 BST 2004
 * Revision     : $Revision: 1.5 $
 * Updated      : $Date: 2014/07/11 13:26:15 PKT $
 */

package mpi;

import java.nio.ByteBuffer;

/**
 * This class represents the MPI environment.
 */
public class MPI {

  static {
    System.loadLibrary("mvapich2-j");
  }

  public static final int NUM_OF_PROCESSORS = Runtime.getRuntime().availableProcessors();
  
  public static final Info INFO_NULL = new Info(Info.NULL);
  
  static final int OP_BSEND = 1;
  static final int OP_RSEND = 2;
  static final int OP_SEND = 3;
  static final int OP_RECV = 4;
  static final int OP_SSEND = 5;
  
  public static Datatype NULL = new BasicType(0);
  public static Datatype BYTE = new BasicType(Datatype.BYTE);
  public static Datatype CHAR = new BasicType(Datatype.CHAR);
  public static Datatype SHORT = new BasicType(Datatype.SHORT);
  public static Datatype BOOLEAN = new BasicType(Datatype.BOOLEAN);
  public static Datatype INT = new BasicType(Datatype.INT);
  public static Datatype LONG = new BasicType(Datatype.LONG);
  public static Datatype FLOAT = new BasicType(Datatype.FLOAT);
  public static Datatype DOUBLE = new BasicType(Datatype.DOUBLE);

  /**
   * <p style="color:red"> Not supported in the current release 
   */
  public static Datatype PACKED = new BasicType(Datatype.PACKED);
  public static Datatype LB = new BasicType(Datatype.LB);
  public static Datatype UB = new BasicType(Datatype.UB);
  public static Datatype OBJECT = new BasicType(12);
  
  public static int THREAD_SINGLE = 0;
  public static int THREAD_FUNNELED = 1;
  public static int THREAD_SERIALIZED = 2;
  public static int THREAD_MULTIPLE = 3;
  
  /**
   * <p style="color:red"> Not supported in the current release 
   */
  public static Datatype SHORT2 = Datatype.createContiguous(2, MPI.SHORT);
  public static Datatype INT2 = Datatype.createContiguous(2, MPI.INT);
  public static Datatype LONG2 = Datatype.createContiguous(2, MPI.LONG);
  public static Datatype FLOAT2 = Datatype.createContiguous(2, MPI.FLOAT);
  public static Datatype DOUBLE2 = Datatype.createContiguous(2, MPI.DOUBLE);
  
  public static Op MAX = new Max();
  public static Op MIN = new Min();
  public static Op SUM = new Sum();
  public static Op PROD = new Prod();
  public static Op LAND = new Land();
  public static Op BAND = new Band();
  public static Op LOR = new Lor();
  public static Op BOR = new Bor();
  public static Op LXOR = new Lxor();
  public static Op BXOR = new Bxor();
  public static Op MAXLOC = new Op(new Maxloc(), true, Constants.MAXLOC_CODE);
  public static Op MINLOC = new Op(new Minloc(), true, Constants.MINLOC_CODE);
  
  // values are set in mpi_MPI.c
  public static int ANY_SOURCE;
  public static int ANY_TAG;
  public static Status EMPTY_STATUS = new Status(MPI.ANY_SOURCE, MPI.ANY_TAG, 0, 0);
  public static int PROC_NULL;
  public static int GRAPH, CART;
  public static int BSEND_OVERHEAD;
  public static int UNDEFINED;
  public static int IDENT, CONGRUENT, SIMILAR, UNEQUAL;
  public static int TAG_UB, HOST, IO;
  
  // needed for mpjbuf.Buffers
  public static int SEND_OVERHEAD = 8;
  public static int RECV_OVERHEAD = 8;
  public static Group GROUP_EMPTY, GROUP_NULL;
  
  // all initialized here in this file
  public static Intracomm COMM_WORLD;
  public static Comm COMM_SELF, COMM_NULL;
  public static Request REQUEST_NULL = new Request(Request.NULL);
  private static boolean initialized = false; 

  /**
   * Provides to MPI a buffer in user's memory to be used for buffering outgoing messages.
   *
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * @param userBuffer the user buffer.
   * 
   * <p>Java binding of the MPI operation {@code MPI_BUFFER_ATTACH}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static void attachBuffer(ByteBuffer userBuffer) throws MPIException {
  }

  /**
   * Detach the buffer currently associated with MPI. 
   * 
   * <p style="color:red"> Not Implemented in the current release </p>.
   *
   * Java binding of the MPI operation {@code MPI_BUFFER_DETACH}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static void detachBuffer() throws MPIException {
  }

  /** 
   * Used to initialize MPI with certain level of threadedness ...
   *
   * @param args arguments to {@code main} method.
   * @param required Level of desired thread support.
   * @return level of provided thread support.
   * 
   * <p> Java binding of the MPI operation {@code MPI_INIT_THREAD}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static int InitThread(String[] args, int required) throws MPIException {
    int provided;

    try {
      provided = nativeInitThread(args, required);
      COMM_WORLD = new Intracomm(Comm.WORLD);
    } catch (Exception e) {
      throw new MPIException(e);
    }

    COMM_SELF = new Comm(Comm.SELF);
    GROUP_EMPTY = new Group(Group.EMPTY);

    // for errors
    GROUP_NULL = new Group(Group.NULL);
    COMM_NULL = new Comm(Comm.NULL);

    SetConstant();

    COMM_WORLD.setAttr(MPI.TAG_UB, COMM_WORLD.getTagUB());
    COMM_WORLD.setAttr(MPI.HOST, PROC_NULL);
    COMM_WORLD.setAttr(MPI.IO, COMM_WORLD.getRank());

    initialized = true;

    return provided;
  }

  /**
   * Returns true if this thread initialized MPI
   *
   * @return {@code true} for thread which initialized MPI.
   */
  public static boolean isThreadMain() {
    return true;
  }

  /**
   * Returns the level of thread support provided by the MPI library and the underlying device
   * selected
   *
   * @return {@code MPI.THREAD_MULTIPLE}.
   */
  public static int queryThread() {
    return MPI.THREAD_MULTIPLE;
  }

  /**
   * Initialize MPI.
   *
   * @param argv arguments to {@code main} method.
   * @return arguments passed
   *
   * <p>Java binding of the MPI operation {@code MPI_INIT}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static String[] Init(String[] argv) throws MPIException {

    try {

      nativeInit(argv);

      COMM_WORLD = new Intracomm(Comm.WORLD);
    } catch (Exception e) {
      throw new MPIException(e);
    }

    COMM_SELF = new Comm(Comm.SELF);

    GROUP_EMPTY = new Group(Group.EMPTY);

    // for errors
    GROUP_NULL = new Group(Group.NULL);
    COMM_NULL = new Comm(Comm.NULL);

    SetConstant();

    COMM_WORLD.setAttr(MPI.TAG_UB, COMM_WORLD.getTagUB());
    COMM_WORLD.setAttr(MPI.HOST, PROC_NULL);
    COMM_WORLD.setAttr(MPI.IO, COMM_WORLD.getRank());

    initialized = true;

    return argv;
  }

  /**
   * Finalize MPI.
   *
   * <p>Java binding of the MPI operation {@code MPI_FINALIZE}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static void Finalize() throws MPIException {

    MPI.detachBuffer();

    try {
      nativeFinish();
    } catch (Exception e) {
      throw new MPIException(e);
    }

    initialized = false;
  }

  /**
   * Returns the name of the processor on which it is called.
   *
   * @return a unique specifier for the actual node.
   *
   * <p>Java binding of the MPI operation {@code MPI_GET_PROCESSOR_NAME}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static String getProcessorName() throws MPIException {
    try {
      java.net.InetAddress addr = java.net.InetAddress.getLocalHost();
      String hostname = addr.getHostName();
      return hostname;
    } catch (java.net.UnknownHostException e) {
      throw new MPIException(e);
    }
  }

  /** 
   * Set Error Handler <p style="color:red"> Not Implemented in the current release </p>
   *
   * @param errhandler error handler.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static void Errorhandler_set(Errhandler errhandler) throws MPIException {}

  /** 
   * Gets the error handler <p style="color:red"> Not Implemented in the current release </p>
   * 
   * @return error handler
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Errhandler Errorhandler_get() throws MPIException {
    return null;
  }

  /**
   * Returns wallclock time.
   *
   * @return elapsed wallclock time in seconds since sometime in the past.
   *
   * <p>Java binding of the MPI operation {@code MPI_WTIME}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static double wtime() throws MPIException {
    return nativeWtime();
  }

  /**
   * Returns resolution of timer.
   *
   * @return resolution of {@code wtime} in seconds.
   *
   * <p>Java binding of the MPI operation {@code MPI_WTICK}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static double wtick() throws MPIException {
    double time = 0L;
    double newTime;
    double smaller = Double.MAX_VALUE;

    for (int i = 0; i < 100000; i++) {
      time = wtime();
      newTime = wtime();
      smaller = Math.min(smaller, (newTime - time));
    }

    return smaller;
  }

  /**
   * Test if MPI has been initialized.
   *
   * @return {@code true} if {@code init} has been called, {@code false}
   *         otherwise.
   *
   * <p>Java binding of the MPI operation {@code MPI_INITIALIZED}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static boolean isInitialized() throws MPIException {
    return initialized;
  }

  private static native void SetConstant();

  private static native void nativeInit(String[] args);

  private static native int nativeInitThread(String[] args, int required);

  private static native void nativeFinish();

  private static native double nativeWtime();
}
