/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
The MIT License

Copyright (c) 2005 - 2008
  1. Distributed Systems Group, University of Portsmouth (2005)
  2. Community Grids Laboratory, Indiana University (2005)
  3. Aamir Shafi (2005 - 2008)
  4. Bryan Carpenter (2005 - 2008)
  5. Mark Baker (2005 - 2008)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/*
 * File         : Request.java
 * Author       : Aamir Shafi, Bryan Carpenter
 * Created      : Fri Sep 10 12:22:15 BST 2004
 * Revision     : $Revision: 1.26 $
 * Updated      : $Date: 2014/03/11 13:26:15 PKT $
 */

package mpi;

import java.util.Iterator;
import java.util.LinkedList;

/**
 * Request object representing {@code MPI_REQUEST}.
 */
public class Request {

  static final int NULL = 0;
  public long handle;
  public boolean isBufferComms = false;
  protected long commHandle;
  protected mpjbuf.Buffer bufferHandle;
  Datatype datatype = null;
  volatile boolean completed = false;
  Status status = null;
  LinkedList list = new LinkedList();

  Request() {}

  Request(boolean isBufferComms) {
    this.isBufferComms = isBufferComms;
  }

  public Request(long commHandle_) {
    this.commHandle = commHandle_;
  }

  public Request(long commHandle_, boolean isBufferComms) {
    this.commHandle = commHandle_;
    this.isBufferComms = isBufferComms;
  }

  Request(int type) {
    GetReq(type);
  }

  /**
   * Tests for completion of either one or none of the operations associated with active requests.
   *
   * @param r array of requests.
   * @return status object or null reference.
   *
   * <p>Java binding of the MPI operation {@code MPI_TESTANY}.
   *
   * <p>If some request completed, the index in {@code array_of_requests} for that request can be
   * obtained from the returned status object through the. The corresponding element of
   * {@code array_of_requests} becomes inactive. If no request completed, {@code Testany} returns
   * a null reference.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Status testAnyStatus(Request[] r) throws MPIException {
    return nativeTestAny(r);
  }

  /**
   * Blocks until all of the operations associated with the active requests in the array have
   * completed.
   *
   * @param r array of requests.
   *
   * <p>Java binding of the MPI operation {@code MPI_WAITALL}.
   *
   * <p>The result array will be the same size as {@code array_of_requests}. On exit, requests
   * become inactive. If the <em>input</em> value of {@code arrayOfRequests} contains inactive
   * requests, corresponding elements of the result array will contain null status references.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static void waitAll(Request[] r) throws MPIException {

    Status s[] = nativeWaitAll(r);

    for (int i = 0; i < s.length; i++) {
      s[i].free();
    }
  }

  /**
   * Blocks until all of the operations associated with the active requests in the array have
   * completed.
   *
   * @param r array of requests.
   * @return array of status objects.
   *
   * <p>Java binding of the MPI operation {@code MPI_WAITALL}.
   *
   * <p>The result array will be the same size as {@code array_of_requests}. On exit, requests
   * become inactive. If the <em>input</em> value of {@code arrayOfRequests} contains inactive
   * requests, corresponding elements of the result array will contain null status references.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Status[] waitAllStatus(Request[] r) throws MPIException {
    return nativeWaitAll(r);
  }

  private static native Status[] nativeWaitAll(Request[] r);

  private static native Status[] nativeTestAll(Request[] r);

  private static native Status nativeWaitAny(Request[] r);

  private static native Status nativeTestAny(Request[] r);

  /**
   * Blocks until one of the operations associated with the active requests in the array has
   * completed.
   *
   * @param r array of requests.
   * @return status object.
   *
   * <p>Java binding of the MPI operation {@code MPI_WAITANY}.
   *
   * <p>The index in {@code array_of_requests} for the request that completed can be obtained from
   * the returned status object through the {@code Status.index} field. The corresponding element
   * of {@code array_of_requests} becomes inactive.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Status waitAnyStatus(Request[] r) throws MPIException {
    return nativeWaitAny(r);
  }

  /**
   * Tests for completion of <em>all</em> of the operations associated with active requests.
   *
   * @param r array of requests.
   * @return array of status objects.
   *
   * <p>Java binding of the MPI operation {@code MPI_TESTALL}.
   *
   * <p>If all operations have completed, the exit value of the argument array and the result array
   * are as for {@code Waitall}. If any operation has not completed, the result value is null and
   * no element of the argument array is modified.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static boolean testAll(Request[] r) throws MPIException {

    boolean completed = true;
    Status s[] = nativeTestAll(r);

    if (s != null) {
      for (int i = 0; i < s.length; i++) {
        if (s[i] != null) {
          s[i].free();
        } else {
          completed = false;
        }
      }
    } else {
      completed = false;
    }

    return completed;
  }

  /**
   * Tests for completion of <em>all</em> of the operations associated with active requests.
   *
   * @param r array of requests.
   * @return array of status objects.
   *
   * <p>Java binding of the MPI operation {@code MPI_TESTALL}.
   *
   * <p>If all operations have completed, the exit value of the argument array and the result array
   * are as for {@code Waitall}. If any operation has not completed, the result value is null and
   * no element of the argument array is modified.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Status[] testAllStatus(Request[] r) throws MPIException {
    return nativeTestAll(r);
  }

  /**
   * Blocks until at least one of the operations associated with the active requests in the array
   * has completed.
   *
   * @param r array of requests.
   * @return array of status objects.
   *
   * <p>Java binding of the MPI operation {@code MPI_WAITSOME}.
   *
   * <p>The size of the result array will be the number of operations that completed. The index in
   * {@code array_of_requests} for each request that completed can be obtained from the returned
   * status objects through the {@code Status.index} field. The corresponding element in
   * {@code array_of_requests} becomes inactive.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Status[] waitSomeStatus(Request[] r) throws MPIException {
    return nativeWaitAll(r);
  }

  /**
   * Behaves like {@code Waitsome}, except that it returns immediately.
   *
   * @param r array of requests.
   * @return array of status objects.
   *
   * <p>Java binding of the MPI operation {@code MPI_TESTSOME}.
   *
   * <p>If no operation has completed, {@code TestSome} returns an array of length zero and
   * elements of {@code array_of_requests} are unchanged. Otherwise, arguments and return value are
   * as for {@code Waitsome}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public static Status[] testSomeStatus(Request[] r) throws MPIException {
    return nativeTestAll(r);
  }

  /**
   * Blocks until the operation identified by the request is complete.
   *
   * @return status object.
   *
   * <p>Java binding of the MPI operation {@code MPI_WAIT}.
   *
   * <p>After the call returns, the request object becomes inactive.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public boolean waitFor() throws MPIException {

    if (completed) {
      return true;
    }

    Status status = new Status(isBufferComms);
    Wait(status);
    complete(status);
    this.completed = true;
    this.status = status;

    status.free();
    return true;
  }

  /**
   * Blocks until the operation identified by the request is complete.
   *
   * @return status object.
   *
   * <p>Java binding of the MPI operation {@code MPI_WAIT}.
   *
   * <p>After the call returns, the request object becomes inactive.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Status waitStatus() throws MPIException {

    if (completed) {
      return this.status;
    }

    Status result = new Status(isBufferComms);
    Wait(result);
    complete(result);
    this.completed = true;
    this.status = result;

    return result;
  }

  private native Status Wait(Status stat);

  private native void GetReq(int type);

  /**
   * Returns a status object if the operation identified by the request is complete, or a null
   * reference otherwise.
   *
   * @return status object or null reference.
   *
   * <p>Java binding of the MPI operation {@code MPI_TEST}.
   *
   * <p>After the call, if the operation is complete (ie, if the return value is non-null), the
   * request object becomes inactive.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public boolean test() throws MPIException {
    if (this.completed) {
      return true;
    }

    Status devStatus = new Status(isBufferComms);

    if (Test(devStatus) == null) return false;
    else {
      complete(devStatus);
      this.completed = true;
      this.status = devStatus;
    }

    if (devStatus == null) {
      return false;
    }

    devStatus.free();
    return true;
  }

  /**
   * Returns a status object if the operation identified by the request is complete, or a null
   * reference otherwise.
   *
   * @return status object or null reference.
   *
   * <p>Java binding of the MPI operation {@code MPI_TEST}.
   *
   * <p>After the call, if the operation is complete (ie, if the return value is non-null), the
   * request object becomes inactive.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public Status testStatus() throws MPIException {

    if (this.completed) {
      return null;
    }

    Status devStatus = new Status(isBufferComms);

    if (Test(devStatus) == null) return null;
    else {
      complete(devStatus);
      this.completed = true;
      this.status = devStatus;
    }

    if (devStatus == null) {
      return null;
    }

    return devStatus;
  }

  private native Status Test(Status stat);

  public void finalize() throws MPIException {}

  /**
   * Test if request object is void.
   *
   * @return {@code true} if the request object is void, 
   *         otherwise {@code false}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public boolean isNull() throws MPIException {

    if (this.handle == MPI.REQUEST_NULL.handle) return true;
    else return false;
  }

  /**
   * Mark a pending nonblocking communication for cancellation. 
   * 
   * <p>Java binding of the MPI operation {@code MPI_CANCEL}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public void cancel() throws MPIException {}

  public boolean mpjdev_cancel() {
    return false;
  }

  public boolean mpjdev_isnull() {
    return false;
  }

  public void mpjdev_free() {}

  // invoke all completion handlers in order they were defined
  protected void complete(mpi.Status status) {
    Iterator iter = list.iterator();
    CompletionHandler handler = null;

    while (iter.hasNext()) {

      handler = (CompletionHandler) iter.next();

      handler.handleCompletion(status);
    }

    list.clear();
  }

  // add handler (logically) to the end of a list
  public void addCompletionHandler(CompletionHandler handler) {
    list.add(handler);
  }
}
