/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
The MIT License

Copyright (c) 2005 - 2008
  1. Distributed Systems Group, University of Portsmouth (2005)
  2. Community Grids Laboratory, Indiana University (2005)
  3. Aamir Shafi (2005 - 2008)
  4. Bryan Carpenter (2005 - 2008)
  5. Mark Baker (2005 - 2008)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*
 * File         : Status.java
 * Author       : Aamir Shafi, Bryan Carpenter
 * Created      : Fri Sep 10 12:22:15 BST 2004
 * Revision     : $Revision: 1.14 $
 * Updated      : $Date: 2005/11/27 19:37:22 $
 */

package mpi;

/**
 * This class represents {@code MPI_STATUS}.
 */
public class Status implements Freeable {

  /** For a receive operation, the source of the message. */
  public int source;
  /** For a receive operation, the tag in the message. */
  public int tag;
  /**
   * For a `waitany()' operation in class `Request'. 
   *
   * <p>This field defines which communication in the `reqs' array was selected.
   */
  public int index;

  public int count;
  public int numEls;
  public mpjbuf.Type type;
  public boolean isBufferComms = false;
  long handle;
  private boolean countAlreadyCalled = false;

  // this.source = -1; is introduced here so as to provide functionality for
  // completionHandler for nativeDevice .. see /mpjdev/nativ/Comm irecv()
  public Status() {
    alloc();
    this.source = -1;
  }

  public Status(boolean isBufferComms) {
    alloc();
    this.source = -1;
    this.isBufferComms = isBufferComms;
  }

  public Status(int source, int tag, int index) {
    alloc();
    this.source = source;
    this.tag = tag;
    this.index = index;
  }

  // used in MPI.java to initialize EMPTY_STATUS
  Status(int s, int t, int numEls, int count) {
    alloc();
    this.source = s;
    this.tag = t;
    this.numEls = numEls;
    this.count = count;
  }

  private native void alloc();

  private native void dealloc();

  protected void finalize() throws Throwable {
    free();
    try {
      super.finalize();
    } catch (Throwable th) {
      th.printStackTrace();
    }
  }

  /**
   * Get the number of received entries.
   *
   * @param datatype datatype of each item in receive buffer.
   * @return number of received entries.
   *
   * <p>Java binding of the MPI operation {@code MPI_GET_COUNT}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int getCount(Datatype datatype) throws MPIException {
    int count;

    if (datatype == MPI.BOOLEAN) count = get_count(MPI.BYTE.getType());
    else count = get_count(datatype.getType());

    if (count == 0) return count;

    if (!isBufferComms) {

      switch (datatype.getType()) {
        case Datatype.BYTE:
          count -= 16;
          break;

        case Datatype.CHAR:
          count -= 16;
          count /= 2; // MPI.CHAR is 2 bytes long
          break;

        case Datatype.SHORT:
          count -= 8;
          break;

        case Datatype.BOOLEAN:
          count -= 16;
          break;

        case Datatype.INT:
          count -= 4;
          break;

        case Datatype.LONG:
          count -= 2;
          break;

        case Datatype.FLOAT:
          count -= 4;
          break;

        case Datatype.DOUBLE:
          count -= 2;
          break;

        case Datatype.NULL:
        case Datatype.UNDEFINED:
        case Datatype.PACKED:
        case Datatype.OBJECT:
          System.out.println("Unsupported datatype (" + datatype.getType() + ") in getCount()");
          break;

        default:
          System.out.println("Unsupported datatype in getCount()");
          break;
      }
    }

    return count;
  }

  private native int get_count(int datatype);

  public int getSource() {
    return nativeGetSource();
  }

  private native int nativeGetSource();

  public int getTag() {
    return nativeGetTag();
  }

  private native int nativeGetTag();

  public int getIndex() {
    return nativeGetIndex();
  }

  private native int nativeGetIndex();

  /**
   * Test if communication was cancelled.
   *
   * @return {@code true} if the operation was successfully cancelled,
   *         otherwise {@code false}.
   *
   * <p>Java binding of the MPI operation {@code MPI_TEST_CANCELLED}.
   * 
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public boolean isCancelled() throws MPIException {
    System.out.println("Cancel functionality is not implemented.");
    return false;
  }

  /** */
  public void free() {
    dealloc();
  }

  /**
   * Retrieve number of basic elements from status.
   *
   * @param datatype datatype used by receive operation.
   * @return number of received basic elements.
   *
   * <p>Java binding of the MPI operation {@code MPI_GET_ELEMENTS}.
   *
   * @throws MPIException Points out that an MPI exception has occured.
   */
  public int getElements(Datatype datatype) throws MPIException {

    int elements;

    if (datatype == MPI.BOOLEAN) elements = get_elements(MPI.BYTE.getType());
    else elements = get_elements(datatype.getType());

    if (elements == 0) return elements;

    if (!isBufferComms) {

      switch (datatype.getType()) {
        case Datatype.BYTE:
          elements -= 16;
          break;

        case Datatype.CHAR:
          elements -= 16;
          elements /= 2; // MPI.CHAR is 2 bytes long
          break;

        case Datatype.SHORT:
          elements -= 8;
          break;

        case Datatype.BOOLEAN:
          elements -= 16;
          break;

        case Datatype.INT:
          elements -= 4;
          break;

        case Datatype.LONG:
          elements -= 2;
          break;

        case Datatype.FLOAT:
          elements -= 4;
          break;

        case Datatype.DOUBLE:
          elements -= 2;
          break;

        case Datatype.NULL:
        case Datatype.UNDEFINED:
        case Datatype.PACKED:
        case Datatype.OBJECT:
          System.out.println("Unsupported datatype (" + datatype.getType() + ") in getCount()");
          break;

        default:
          System.out.println("Unsupported datatype in getCount()");
          break;
      }
    }

    return elements;
  }

  private native int get_elements(int datatype);
}
