/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/*
The MIT License

 Copyright (c) 2005 - 2008
   1. Distributed Systems Group, University of Portsmouth (2005)
   2. Community Grids Laboratory, Indiana University (2005)
   3. Aamir Shafi (2005 - 2008)
   4. Bryan Carpenter (2005 - 2008)
   5. Mark Baker (2005 - 2008)

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/*
 * File         : Buddy2BufferFactory.java
 * Author       : Aamir Shafi, Bryan Carpenter
 * Created      : Sat July 9 12:22:15 BST 2005
 * Revision     : $Revision: 1.6 $
 * Updated      : $Date: 2005/09/11 00:14:35 $
 *
 */
package mpjbuf;

import java.util.ArrayList;

public class Buddy2BufferFactory extends BufferFactory {

  private ArrayList<Buddy2Region> regionLists = new ArrayList<Buddy2Region>();

  public synchronized void init() {
    Buddy2Region region = new Buddy2Region(BufferConstants.DEFAULT_REGION_SIZE);
    regionLists.add(region);
  }

  public synchronized void finalixe() {

    mpjbuf.Buddy2Buffer current = null;

    for (Buddy2Region reg : regionLists) {

      int numLevels = reg.freeLists.size();

      for (int l = 0; l < numLevels; l++) {

        Buddy2FreeList fList = reg.freeLists.get(l);

        if (!fList.isEmpty()) {

          current = (mpjbuf.Buddy2Buffer) fList.head;

          do {
            current.free();
            current = (mpjbuf.Buddy2Buffer) current.prev;
          } while (current != (mpjbuf.Buddy2Buffer) fList.head);
        }

        fList = null;
      }

      reg.buffer = null;

      reg = null;
    }

    regionLists = null;
  }

  /** Return a free buffer of level greater than or equal to `level'. */
  mpjbuf.Buddy2Buffer findInAllRegions(int level) {
    mpjbuf.Buddy2Buffer freeBuffer = null;
    foundIt:
    for (Buddy2Region reg : regionLists) {

      int numLevels = reg.freeLists.size();

      if (level < numLevels) {
        for (int l = level; l < numLevels; l++) {

          Buddy2FreeList freeList = reg.freeLists.get(l);

          if (!freeList.isEmpty()) {

            if (freeList.head == null) {
              continue;
            } else {
              mpjbuf.Buddy2Buffer current = (mpjbuf.Buddy2Buffer) freeList.head;
              do {
                if (current.free) {
                  freeBuffer = current;
                  break foundIt;
                } 
                current = (mpjbuf.Buddy2Buffer) current.prev;
              } while (current != (mpjbuf.Buddy2Buffer) freeList.head);
            } // end else.
          } // end if.
        } // end for.
      } // end if.
    } // end for iterating the regions ..

    return freeBuffer;
  }

  mpjbuf.Buddy2Buffer find(Buddy2Region inRegion, int level) {
    mpjbuf.Buddy2Buffer freeBuffer = null;

    int numLevels = inRegion.freeLists.size();

    if (level < numLevels) {
      foundIt:
      for (int l = level; l < numLevels; l++) {
        Buddy2FreeList freeList = inRegion.freeLists.get(l);

        if (!freeList.isEmpty()) {

          if (freeList.head == null) {
            continue;
          } else {
            mpjbuf.Buddy2Buffer current = (mpjbuf.Buddy2Buffer) freeList.head;

            do {
              if (current.free) {
                freeBuffer = current;
                break foundIt;
              } 
              current = (mpjbuf.Buddy2Buffer) current.prev;
            } while (current != (mpjbuf.Buddy2Buffer) freeList.tail);
          } // end else.
        } // end if.
      } // end for.
    } // end if.

    return freeBuffer;
  }

  public synchronized mpjbuf.RawBuffer createBuffer(int size) {

    if (size == 0) {

      // do something weird.
      return null;

    } else {

      int numChunks = (size + BufferConstants.CHUNK_SIZE - 1) / BufferConstants.CHUNK_SIZE;
      int level = CustomMath.widthInBits(numChunks - 1);

      mpjbuf.Buddy2Buffer freeNode = findInAllRegions(level);

      if (freeNode == null) {
        int newRegionSize = -1;

        if (size > BufferConstants.DEFAULT_REGION_SIZE) {
          for (int h = 2; ; h = h * 2) {
            if (size < BufferConstants.DEFAULT_REGION_SIZE * h) {
              newRegionSize = BufferConstants.DEFAULT_REGION_SIZE * h;
              break;
            }
          }

        } else {
          newRegionSize = BufferConstants.DEFAULT_REGION_SIZE;
        }

        Buddy2Region reg = new Buddy2Region(newRegionSize);
        regionLists.add(reg);

        freeNode = find(reg, level);
      }

      mpjbuf.Buddy2Buffer output = block(freeNode, freeNode.region, level);

      return output;
    }
  }

  /**
   * return a buffer of level 'level' from freeBuffer and return remainder of freeBuffer (if any) to
   * the freeList
   */
  private mpjbuf.Buddy2Buffer block(
      mpjbuf.Buddy2Buffer freeBuffer, Buddy2Region region, int level) {

    int l = freeBuffer.level;

    mpjbuf.Buddy2Buffer buddy = null;

    /* luckily we found a freeBuffer that is at the level we wanted */
    if (level == l) {

      freeBuffer.free = false;

      return freeBuffer;
    } else {

      region.freeLists.get(l).remove(freeBuffer);

      for (l = l - 1; l >= level; l--) {
        int childCap = freeBuffer.capacity / 2;

        region.buffer.position(freeBuffer.realOffset);
        region.buffer.limit(freeBuffer.realOffset + childCap);
        freeBuffer =
            new mpjbuf.Buddy2Buffer(
                region.buffer.slice(), 0, childCap, l, region, freeBuffer.realOffset);
        region.buffer.position(freeBuffer.realOffset + childCap);
        region.buffer.limit(freeBuffer.realOffset + (2 * childCap));

        buddy =
            new mpjbuf.Buddy2Buffer(
                region.buffer.slice(), 0, childCap, l, region, (freeBuffer.realOffset + childCap));
        region.buffer.clear(); // dodgy ...
        region.freeLists.get(l).add(buddy);

        if (l == level) {

          region.freeLists.get(l).add(freeBuffer);
        }
      }

      freeBuffer.free = false;

      return freeBuffer;
    }
  }

  private mpjbuf.Buddy2Buffer calculateBuddy(mpjbuf.Buddy2Buffer buffer) {
    Buddy2FreeList fList = buffer.region.freeLists.get(buffer.level);
    
    int buddyOffset = buffer.realOffset ^ (BufferConstants.CHUNK_SIZE << buffer.level);
    
    mpjbuf.Buddy2Buffer current = (mpjbuf.Buddy2Buffer) fList.head;
    mpjbuf.Buddy2Buffer buddy = null;

    do {

      if (current.realOffset == buddyOffset) {
        if (current.free) {

          buddy = current;
          break;
        } else {

          // this means we've found the buddy, but its not free ..so there's
          // not much we can do except set buffer.free = true and exit ..
          break;
        }
      } 

      current = (mpjbuf.Buddy2Buffer) current.prev;

    } while (current != (mpjbuf.Buddy2Buffer) fList.head);

    return buddy;
  }

  public synchronized void destroyBuffer(mpjbuf.RawBuffer buffer) {
    mpjbuf.Buddy2Buffer buf = (mpjbuf.Buddy2Buffer) buffer;

    mpjbuf.Buddy2Buffer buddy = calculateBuddy(buf);

    if (buddy == null) {
      buf.free = true;
      try {
        ((NIOBuffer) buf).clear();
      } catch (Exception e) {
        e.printStackTrace();
      }
      return;
    }

    int numLevels = buf.region.freeLists.size();
    int level = buf.level;
    Buddy2FreeList fList, topLevelList;

    while (buddy != null) {
      fList = buf.region.freeLists.get(level);

      if (buf.region.freeLists.size() == (level + 1)) {
        topLevelList = new Buddy2FreeList();
        buf.region.freeLists.add(topLevelList);
      } else {
        topLevelList = buf.region.freeLists.get(level + 1);
      }

      int parentOffset = -1;

      if (buddy.realOffset > buf.realOffset) {
        parentOffset = buf.realOffset;
      } else {
        parentOffset = buddy.realOffset;
      }

      int parentbufoffset = 0;

      try {
        ((NIOBuffer) buf).clear();
        buf.free = false;
        ((NIOBuffer) buddy).clear();
        buddy.free = false;
      } catch (Exception e) {
        e.printStackTrace();
      }

      fList.remove(buddy); // we know buddy is free ...
      fList.remove(buf);
      buf.region.buffer.position(parentOffset);
      buf.region.buffer.limit(parentOffset + (buf.capacity * 2));
      buf =
          new mpjbuf.Buddy2Buffer(
              buf.region.buffer.slice(),
              parentbufoffset,
              buf.capacity * 2,
              (level + 1),
              buf.region,
              parentOffset);
      topLevelList.add(buf);
      buf.region.buffer.clear();
      buddy = calculateBuddy(buf);
      level++;
    }
  }
}
