/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

package mpi.ccl;

/****************************************************************************
 *
 * MESSAGE PASSING INTERFACE TEST CASE SUITE
 *
 * Copyright IBM Corp. 1995
 *
 * IBM Corp. hereby grants a non-exclusive license to use, copy, modify, and
 * distribute this software for any purpose and without fee provided that the
 * above copyright notice and the following paragraphs appear in all copies.
 *
 * IBM Corp. makes no representation that the test cases comprising this
 * suite are correct or are an accurate representation of any standard.
 *
 * In no event shall IBM be liable to any party for direct, indirect, special
 * incidental, or consequential damage arising out of the use of this software
 * even if IBM Corp. has been advised of the possibility of such damage.
 *
 * IBM CORP. SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS AND IBM
 * CORP. HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 ****************************************************************************
 *
 * These test cases reflect an interpretation of the MPI Standard.  They are
 * are, in most cases, unit tests of specific MPI behaviors.  If a user of any
 * test case from this set believes that the MPI Standard requires behavior
 * different than that implied by the test case we would appreciate feedback.
 *
 * Comments may be sent to:
 * Richard Treumann
 * treumann@kgn.ibm.com
 ****************************************************************************
 *
 * MPI-Java version :
 * Sung-Hoon Ko(shko@npac.syr.edu)
 * Northeast Parallel Architectures Center at Syracuse University
 * 03/22/98
 ****************************************************************************/

import mpi.*;

import java.nio.*;

public class ReduceScatterWithByteBuffer {
  public ReduceScatterWithByteBuffer() {}

  public ReduceScatterWithByteBuffer(String[] args) throws Exception {

    final int MAXLEN = 10000;

    ByteBuffer outInt = ByteBuffer.allocateDirect(MAXLEN * 100 * MPI.INT.getByteSize());
    ByteBuffer outDouble = ByteBuffer.allocateDirect(MAXLEN * 100 * MPI.DOUBLE.getByteSize());
    ByteBuffer outFloat = ByteBuffer.allocateDirect(MAXLEN * 100 * MPI.FLOAT.getByteSize());
    ByteBuffer outLong = ByteBuffer.allocateDirect(MAXLEN * 100 * MPI.LONG.getByteSize());
    ByteBuffer outShort = ByteBuffer.allocateDirect(MAXLEN * 100 * MPI.SHORT.getByteSize());

    ByteBuffer inInt = ByteBuffer.allocateDirect(MAXLEN * 100 * MPI.INT.getByteSize());
    ByteBuffer inDouble = ByteBuffer.allocateDirect(MAXLEN * 100 * MPI.DOUBLE.getByteSize());
    ByteBuffer inFloat = ByteBuffer.allocateDirect(MAXLEN * 100 * MPI.FLOAT.getByteSize());
    ByteBuffer inLong = ByteBuffer.allocateDirect(MAXLEN * 100 * MPI.LONG.getByteSize());
    ByteBuffer inShort = ByteBuffer.allocateDirect(MAXLEN * 100 * MPI.SHORT.getByteSize());

    inInt.order(ByteOrder.LITTLE_ENDIAN);
    inDouble.order(ByteOrder.LITTLE_ENDIAN);
    inShort.order(ByteOrder.LITTLE_ENDIAN);
    inLong.order(ByteOrder.LITTLE_ENDIAN);
    inFloat.order(ByteOrder.LITTLE_ENDIAN);

    outInt.order(ByteOrder.LITTLE_ENDIAN);
    outDouble.order(ByteOrder.LITTLE_ENDIAN);
    outShort.order(ByteOrder.LITTLE_ENDIAN);
    outLong.order(ByteOrder.LITTLE_ENDIAN);
    outFloat.order(ByteOrder.LITTLE_ENDIAN);

    int i, j, k;
    int myself, tasks;
    int recvcounts[] = new int[128];

    MPI.Init(args);
    myself = MPI.COMM_WORLD.getRank();
    tasks = MPI.COMM_WORLD.getSize();

    if (tasks > 8) {
      if (myself == 0) {
        System.out.println("ReduceScatterWithByteBuffer must run with 8 tasks!");
      }
      MPI.Finalize();
      return;
    }

    for (j = 1; j <= MAXLEN * tasks; j *= 10) {
      for (i = 0; i < tasks; i++) recvcounts[i] = j;

      for (i = 0; i < j * tasks; i++) {
        outInt.putInt(i);
        outDouble.putDouble((double)i);
        outFloat.putFloat((float)i);
        outShort.putShort((short)i);
        outLong.putLong((long)i);
      }

      MPI.COMM_WORLD.reduceScatter(outInt, inInt, recvcounts, MPI.INT, MPI.SUM);
      MPI.COMM_WORLD.reduceScatter(outDouble, inDouble, recvcounts, MPI.DOUBLE, MPI.SUM);
      MPI.COMM_WORLD.reduceScatter(outFloat, inFloat, recvcounts, MPI.FLOAT, MPI.SUM);
      MPI.COMM_WORLD.reduceScatter(outShort, inShort, recvcounts, MPI.SHORT, MPI.SUM);
      MPI.COMM_WORLD.reduceScatter(outLong, inLong, recvcounts, MPI.LONG, MPI.SUM);

      IntBuffer inIntBuffer = inInt.asIntBuffer();
      DoubleBuffer inDoubleBuffer = inDouble.asDoubleBuffer();
      FloatBuffer inFloatBuffer = inFloat.asFloatBuffer();
      ShortBuffer inShortBuffer = inShort.asShortBuffer();
      LongBuffer inLongBuffer = inLong.asLongBuffer();

      for (k = 0; k < j; k++) {
        int valInt = inIntBuffer.get(k);
        double valDouble = inDoubleBuffer.get(k);
        float valFloat = inFloatBuffer.get(k);
        long valLong = inLongBuffer.get(k);
        short valShort = inShortBuffer.get(k);

        if (valInt != tasks * (myself * j + k)) {
          System.out.println(
              "bad answer ("
                  + valInt
                  + ") at index "
                  + k
                  + " of "
                  + j
                  + "(should be "
                  + tasks * (myself * j + k)
                  + ")");
          break;
        }
        if (valDouble != (double)(tasks * (myself * j + k))) {
          System.out.println(
              "bad answer ("
                  + valDouble
                  + ") at index "
                  + k
                  + " of "
                  + j
                  + "(should be "
                  + (double)(tasks * (myself * j + k))
                  + ")");
          break;
        }
        if (valLong != (long)(tasks * (myself * j + k))) {
          System.out.println(
              "bad answer ("
                  + valLong
                  + ") at index "
                  + k
                  + " of "
                  + j
                  + "(should be "
                  + (long)(tasks * (myself * j + k))
                  + ")");
          break;
        }
        if (valShort != (short)(tasks * (myself * j + k))) {
          System.out.println(
              "bad answer ("
                  + valShort
                  + ") at index "
                  + k
                  + " of "
                  + j
                  + "(should be "
                  + (short)(tasks * (myself * j + k))
                  + ")");
          break;
        }
        if (valFloat != (float)(tasks * (myself * j + k))) {
          System.out.println(
              "bad answer ("
                  + valFloat
                  + ") at index "
                  + k
                  + " of "
                  + j
                  + "(should be "
                  + (float)(tasks * (myself * j + k))
                  + ")");
          break;
        }
      }

      inInt.clear();
      inDouble.clear();
      inShort.clear();
      inLong.clear();
      inFloat.clear();

      outInt.clear();
      outDouble.clear();
      outShort.clear();
      outLong.clear();
      outFloat.clear();
    }

    MPI.COMM_WORLD.barrier();
    if (myself == 0) System.out.println("ReduceScatterWithByteBuffer TEST COMPLETE");
    MPI.Finalize();
  }

  public static void main(String[] args) throws Exception {
    try {
      ReduceScatterWithByteBuffer c = new ReduceScatterWithByteBuffer(args);
    } catch (Exception e) {
    }
  }
}
