/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

/****************************************************************************
 *
 * MESSAGE PASSING INTERFACE TEST CASE SUITE
 *
 * Copyright IBM Corp. 1995
 *
 * IBM Corp. hereby grants a non-exclusive license to use, copy, modify, and
 * distribute this software for any purpose and without fee provided that the
 * above copyright notice and the following paragraphs appear in all copies.
 *
 * IBM Corp. makes no representation that the test cases comprising this
 * suite are correct or are an accurate representation of any standard.
 *
 * In no event shall IBM be liable to any party for direct, indirect, special
 * incidental, or consequential damage arising out of the use of this software
 * even if IBM Corp. has been advised of the possibility of such damage.
 *
 * IBM CORP. SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS AND IBM
 * CORP. HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 ****************************************************************************
 *
 * These test cases reflect an interpretation of the MPI Standard.  They are
 * are, in most cases, unit tests of specific MPI behaviors.  If a user of any
 * test case from this set believes that the MPI Standard requires behavior
 * different than that implied by the test case we would appreciate feedback.
 *
 * Comments may be sent to:
 * Richard Treumann
 * treumann@kgn.ibm.com
 ****************************************************************************
 *
 * MPI-Java version :
 * Sung-Hoon Ko(shko@npac.syr.edu)
 * Northeast Parallel Architectures Center at Syracuse University
 * 03/22/98
 ****************************************************************************/

package mpi.ccl;

import mpi.*;

import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import java.nio.FloatBuffer;

import java.nio.ByteOrder;

import java.lang.Math;

public class ReduceWithByteBuffer {
  public ReduceWithByteBuffer() {}

  public ReduceWithByteBuffer(String[] args) throws Exception {

    final int MAXLEN = 100000;

    int root, i, j, k;
    ByteBuffer out = ByteBuffer.allocateDirect(MAXLEN * MPI.INT.getByteSize());
    ByteBuffer outLong = ByteBuffer.allocateDirect(MAXLEN * MPI.LONG.getByteSize());

    ByteBuffer outFloat = ByteBuffer.allocateDirect(MAXLEN * MPI.FLOAT.getByteSize());
    ByteBuffer outDouble = ByteBuffer.allocateDirect(MAXLEN * MPI.DOUBLE.getByteSize());
    ByteBuffer outChar = ByteBuffer.allocateDirect(MAXLEN * MPI.CHAR.getByteSize());

    out.order(ByteOrder.LITTLE_ENDIAN);
    outLong.order(ByteOrder.LITTLE_ENDIAN);
    outFloat.order(ByteOrder.LITTLE_ENDIAN);
    outDouble.order(ByteOrder.LITTLE_ENDIAN);
    outChar.order(ByteOrder.LITTLE_ENDIAN);

    ByteBuffer in = ByteBuffer.allocateDirect(MAXLEN * MPI.INT.getByteSize());
    ByteBuffer inLong = ByteBuffer.allocateDirect(MAXLEN * MPI.LONG.getByteSize());
    ByteBuffer inFloat = ByteBuffer.allocateDirect(MAXLEN * MPI.FLOAT.getByteSize());
    ByteBuffer inDouble = ByteBuffer.allocateDirect(MAXLEN * MPI.DOUBLE.getByteSize());

    in.order(ByteOrder.LITTLE_ENDIAN);
    inLong.order(ByteOrder.LITTLE_ENDIAN);
    inFloat.order(ByteOrder.LITTLE_ENDIAN);
    inDouble.order(ByteOrder.LITTLE_ENDIAN);

    IntBuffer outIntBuffer = out.asIntBuffer();
    IntBuffer inIntBuffer = in.asIntBuffer();

    int myself, tasks;

    MPI.Init(args);
    myself = MPI.COMM_WORLD.getRank();
    tasks = MPI.COMM_WORLD.getSize();

    root = tasks / 2;

    for (j = 1; j <= MAXLEN; j *= 10) {

      for (i = 0; i < j; i++) {
        outIntBuffer.put(i);
        outLong.putLong(i * 1L);
        outFloat.putFloat(i * 1.0f);
        outDouble.putDouble(i * 1.0);
      }

      outIntBuffer.flip();
      out.flip();
      outLong.flip();
      outFloat.flip();
      outDouble.flip();

      MPI.COMM_WORLD.reduce(outIntBuffer, inIntBuffer, j, MPI.INT, MPI.SUM, root);
      MPI.COMM_WORLD.reduce(outLong, inLong, j, MPI.LONG, MPI.SUM, root);
      MPI.COMM_WORLD.reduce(outFloat, inFloat, j, MPI.FLOAT, MPI.SUM, root);
      MPI.COMM_WORLD.reduce(outDouble, inDouble, j, MPI.DOUBLE, MPI.SUM, root);

      if (myself == root) {

        for (k = 0; k < j; k++) {

          int val = in.getInt();
          if (val != k * tasks) {
            System.out.println(
                "bad answer ("
                    + (val)
                    + ") at index "
                    + k
                    + " of "
                    + j
                    + "(should be "
                    + (k * tasks)
                    + ")");
          }

          long valLong = inLong.getLong();
          if (valLong != k * tasks * 1L) {
            System.out.println(
                "bad answer ("
                    + (val)
                    + ") at index "
                    + k
                    + " of "
                    + j
                    + "(should be "
                    + (k * tasks * 1L)
                    + ")");
            break;
          }

          float valFloat = inFloat.getFloat();
          if (valFloat != k * tasks * 1.0f) {
            System.out.println(
                "bad answer ("
                    + (valFloat)
                    + ") at index "
                    + k
                    + " of "
                    + j
                    + "(should be "
                    + (k * tasks * 1.1)
                    + ")");
            break;
          }

          double valDouble = inDouble.getDouble();
          if (valDouble != k * tasks * 1.0) {
            System.out.println(
                "bad answer ("
                    + (valDouble)
                    + ") at index "
                    + k
                    + " of "
                    + j
                    + "(should be "
                    + (k * tasks * 1.1)
                    + ")");
            break;
          }
        }
      }

      out.clear();
      in.clear();

      outLong.clear();
      inLong.clear();

      outIntBuffer.clear();
      inIntBuffer.clear();

      outFloat.clear();
      inFloat.clear();

      outDouble.clear();
      inDouble.clear();
    }

    MPI.COMM_WORLD.barrier();
    if (myself == 0) System.out.println("Reduce TEST COMPLETE");
    MPI.Finalize();
  }

  public static void main(String[] args) throws Exception {
    try {
      ReduceWithByteBuffer c = new ReduceWithByteBuffer(args);
    } catch (Exception e) {
    }
  }
}
