/* Copyright (c) 2001-2022, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-J software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2-J directory.
 */

package mpi.pt2pt;

import mpi.*;

import java.util.Arrays;
import java.nio.ByteBuffer;

public class BlockingPt2ptWithByteBuffer {
  public BlockingPt2ptWithByteBuffer(String args[]) throws Exception {

    MPI.Init(args);

    int msgSize;

    if (args.length != 1) {
      msgSize = 1024; // default value
    } else {
      msgSize = Integer.parseInt(args[0]);
    }

    ByteBuffer sendBuffer = ByteBuffer.allocateDirect(msgSize);
    ByteBuffer recvBuffer = ByteBuffer.allocateDirect(msgSize);

    int tag = 10;

    Request request;
    Status s1;

    for (int i = 0; i < msgSize; i++) {
      sendBuffer.put((byte) 1);
      recvBuffer.put((byte) 0);
    }

    if (MPI.COMM_WORLD.getRank() == 0) {

      int sTag = tag;
      MPI.COMM_WORLD.send(sendBuffer, msgSize, MPI.BYTE, 1, sTag++);

    } else if (MPI.COMM_WORLD.getRank() == 1) {

      int rTag = tag;
      s1 = MPI.COMM_WORLD.recv(recvBuffer, msgSize, MPI.BYTE, 0, rTag++);

      boolean pass = true;

      if (s1.getSource() != 0
          || s1.getTag() != tag
          || s1.getElements(MPI.BYTE) != msgSize
          || s1.getCount(MPI.BYTE) != msgSize) {
        System.out.println(
            "Incorrect values for s1: s1.source=<"
                + s1.getSource()
                + ">, "
                + "s1.tag=<"
                + s1.getTag()
                + ">, s1.numEls=<"
                + s1.getElements(MPI.BYTE)
                + ">, s1.count=<"
                + s1.getCount(MPI.BYTE)
                + ">");
        System.out.println(
            "Correct values for s1: s1.source=<0>, "
                + "s1.tag=<"
                + tag
                + ">, s1.numEls=<"
                + msgSize
                + ">, s1.count=<"
                + msgSize
                + ">, s1.countInBytes=<"
                + msgSize
                + ">\n");
        pass = false;
      }

      recvBuffer.flip();
      sendBuffer.flip();

      byte srcByte, dstByte;

      for (int i = 0; i < msgSize; i++) {
        srcByte = sendBuffer.get();
        dstByte = recvBuffer.get();
        if (srcByte != dstByte) {
          pass = false;
          break;
        }
      }

      if (pass) {

        System.out.println("#################" + "\n <<<<PASSED>>>> " + "\n#################");

      } else {

        System.out.println("#################" + "\n <<<<FAILED>>>> " + "\n#################");
      }
    }

    MPI.COMM_WORLD.barrier();
    try {
      MPI.Finalize();
    } catch (Exception e) {
    }
  }

  public static void main(String args[]) throws Exception {
    BlockingPt2ptWithByteBuffer test = new BlockingPt2ptWithByteBuffer(args);
  }
}
